import 'package:collection/collection.dart';

import 'json.dart';
import 'logger.dart';
import 'path.dart';
import 'pointer.dart';
import 'push_pop.dart';
import 'runtime_object.dart';
import 'simple_json.dart';
import 'story.dart';
import 'value.dart';

final _logger = Logger('CallStack');

class CallStack {
  List<Thread> _threads = [];
  int _threadCounter = 0;
  late Pointer _startOfRoot;

  List<Element> get elements => _callStack;

  int get depth => elements.length;

  Element get currentElement {
    var thread = _threads.last;
    var cs = thread.callstack;
    return cs.last;
  }

  int get currentElementIndex => _callStack.length - 1;

  Thread get currentThread => _threads.last;

  set currentThread(value) {
    assert(_threads.length == 1,
        "Shouldn't be directly setting the current thread when we have a stack of them");
    _threads.clear();
    _threads.add(value);
  }

  // for json
  CallStack();

  CallStack.of({required Story storyContext}) {
    _startOfRoot = Pointer.startOf(storyContext.rootContentContainer);
    reset();
  }

  CallStack.copy(CallStack toCopy) {
    _threads = [];
    toCopy._threads.forEach((otherThread) {
      _threads.add(otherThread.copy());
    });
    _threadCounter = toCopy._threadCounter;
    _startOfRoot = toCopy._startOfRoot.copyWith();
  }

  void reset() {
    _threads = [];
    _threads.add(Thread());
    _threads.first.callstack.add(Element(PushPopType.Tunnel, _startOfRoot));
  }

  /// Unfortunately it's not possible to implement jsonToken since
  /// the setter needs to take a Story as a context in order to
  /// look up objects from paths for currentContainer within elements.
  void setJsonToken(Map<String, dynamic> jObject, Story storyContext) {
    _threads.clear();
    List<dynamic> jThreads = jObject['threads'] ?? [];

    for (var jThreadTok in jThreads) {
      Map<String, dynamic> jThreadObj = jThreadTok;
      var thread = Thread.from(jThreadObj, storyContext);
      _threads.add(thread);
    }

    _threadCounter = jObject['threadCounter'] ?? 0;
    _startOfRoot = Pointer.startOf(storyContext.rootContentContainer);
  }

  void pushThread() {
    var newThread = currentThread.copy();
    _threadCounter++;
    newThread.threadIndex = _threadCounter;
    _threads.add(newThread);
  }

  Thread forkThread() {
    var forkedThread = currentThread.copy();
    _threadCounter++;
    forkedThread.threadIndex = _threadCounter;
    return forkedThread;
  }

  void popThread() {
    _logger.finest('=== popThread() ===');
    if (canPopThread) {
      _logger.finest('Can pop thread');
      var removed = _threads.remove(currentThread);
      _logger.finest('removed : $removed, size : ${_threads.length}');
    } else {
      _logger.severe("Can't pop thread");
      throw Exception("Can't pop thread");
    }
  }

  bool get canPopThread => _threads.length > 1 && !elementIsEvaluateFromGame;

  bool get elementIsEvaluateFromGame =>
      currentElement.type == PushPopType.FunctionEvaluationFromGame;

  void push(PushPopType type,
      {int externalEvaluationStackHeight = 0,
      int outputStreamLengthWithPushed = 0}) {
    var element = Element(type, currentElement.currentPointer,
        inExpressionEvaluation: false);
    element.evaluationStackHeightWhenPushed = externalEvaluationStackHeight;
    element.functionStartInOutputStream = outputStreamLengthWithPushed;
    _callStack.add(element);
  }

  bool canPop([PushPopType? type]) {
    if (_callStack.length <= 1) {
      return false;
    }
    if (type == null) {
      return true;
    }
    return currentElement.type == type;
  }

  void pop([PushPopType? type]) {
    if (canPop(type)) {
      _callStack.removeAt(_callStack.length - 1);
      return;
    } else {
      throw Exception('Mismatched push/pop in Callstack');
    }
  }

  /// Get variable value, dereferencing a variable pointer if necessary
  //public Runtime.Object GetTemporaryVariableWithName(string name, int contextIndex = -1)
  RuntimeObject? getTemporaryVariableWithName(String? name,
      {int contextIndex = -1}) {
    _logger.finest('getTemporaryVariableWithName');
    if (contextIndex == -1) {
      contextIndex = currentElementIndex + 1;
    }

    var contextElement = _callStack[contextIndex - 1];

    return contextElement.temporaryVariables[name];
  }

  void setTemporaryVariable(String name, RuntimeObject value, bool declareNew,
      {int contextIndex = -1}) {
    if (contextIndex == -1) {
      contextIndex = currentElementIndex + 1;
    }
    var contextElement = _callStack[contextIndex - 1];

    if (!declareNew && !contextElement.temporaryVariables.containsKey(name)) {
      // TODO StoryException
      throw Exception('Could not find temporary variable to set: $name');
    }

    var oldValue = contextElement.temporaryVariables[name];
    if (oldValue != null) {
      ListValue.retainListOriginsForAssignment(oldValue, value);
    }
    contextElement.temporaryVariables[name] = value;
  }

  /// Find the most appropriate context for this variable.
  /// Are we referencing a temporary or global variable?
  /// Note that the compiler will have warned us about possible conflicts,
  /// so anything that happens here should be safe!
  int contextForVariableNamed(String name) {
    // Current temporary context?
    // (Shouldn't attempt to access contexts higher in the callstack.)
    if (currentElement.temporaryVariables.containsKey(name)) {
      return currentElementIndex + 1;
    }
    // Global
    else {
      return 0;
    }
  }

  Thread? threadWithIndex(int index) {
    return _threads.firstWhereOrNull((t) => t.threadIndex == index);
  }

  List<Element> get _callStack => currentThread.callstack;

  String get callStackTrace {
    var sb = StringBuffer();

    for (var t = 0; t < _threads.length; t++) {
      var thread = _threads[t];
      var isCurrent = (t == _threads.length - 1);
      sb.writeln(
          "=== THREAD ${(t + 1)}/${_threads.length} ${isCurrent ? '(current) ' : ''}===");

      for (var i = 0; i < thread.callstack.length; i++) {
        if (thread.callstack[i].type == PushPopType.Function) {
          sb.write('  [FUNCTION] ');
        } else {
          sb.write('  [TUNNEL] ');
        }

        var pointer = thread.callstack[i].currentPointer;
        if (!pointer.isNull) {
          sb.write('<SOMEWHERE IN ');
          sb.write(pointer.container!.path.toString());
          sb.write('>');
        }
      }
    }

    return sb.toString();
  }

  //factory CallStack.fromJson(Map<String, dynamic> json) {

  //};
  // Unfortunately it's not possible to implement jsonToken since
  // the setter needs to take a Story as a context in order to
  // look up objects from paths for currentContainer within elements.
  void SetJsonToken(Map<String, dynamic> jObject, Story storyContext) {
    _threads.clear();

    var jThreads = jObject["threads"] as List;

    for (var jThreadTok in jThreads) {
      var jThreadObj = jThreadTok as Map<String, dynamic>;
      var thread = Thread.from(jThreadObj, storyContext);
      _threads.add(thread);
    }

    _threadCounter = jObject["threadCounter"];
    _startOfRoot = Pointer.startOf(storyContext.rootContentContainer);
  }

  void WriteJson(SimpleJsonWriter w) {
    w.WriteObject((writer) {
      writer.WritePropertyStart("threads");
      {
        writer.WriteArrayStart();

        for (var thread in _threads) {
          thread.WriteJson(writer);
        }

        writer.WriteArrayEnd();
      }
      writer.WritePropertyEnd();

      writer.WritePropertyStart("threadCounter");
      {
        writer.Write(_threadCounter);
      }
      writer.WritePropertyEnd();
    });
  }
}

class Element {
  Pointer currentPointer;

  /// When this callstack element is actually a function evaluation called from the game,
  /// we need to keep track of the size of the evaluation stack when it was called
  /// so that we know whether there was any return value.
  bool inExpressionEvaluation = false;

  Map<String, RuntimeObject> temporaryVariables;
  PushPopType type;

  /// When functions are called, we trim whitespace from the start and end of what
  /// they generate, so we make sure know where the function's start and end are.
  int functionStartInOutputStream = 0;

  // When this callstack element is actually a function evaluation called from the game,
  // we need to keep track of the size of the evaluation stack when it was called
  // so that we know whether there was any return value.
  int evaluationStackHeightWhenPushed = 0;

  // When functions are called, we trim whitespace from the start and end of what
  // they generate, so we make sure know where the function's start and end are.
  int functionStartInOuputStream = 0;

  Element(this.type, this.currentPointer, {this.inExpressionEvaluation = false})
      : temporaryVariables = {};

  Element copy() {
    var copy = Element(type, currentPointer,
        inExpressionEvaluation: inExpressionEvaluation);
    copy.temporaryVariables = Map.of(temporaryVariables);
    copy.evaluationStackHeightWhenPushed = evaluationStackHeightWhenPushed;
    copy.functionStartInOuputStream = functionStartInOuputStream;

    return copy;
  }
}

String? previousPointerToJson(Pointer pointer) {
  return pointer.resolve()?.path.toString();
}

//@JsonSerializable(explicitToJson: true)
class Thread {
  List<Element> callstack = <Element>[];
  int threadIndex = 0;
  Pointer previousPointer = Pointer.Null;

  Thread() {
    callstack = <Element>[];
  }

  Thread.from(Map<String, dynamic> jThreadObj, Story storyContext) {
    callstack = <Element>[];

    threadIndex = jThreadObj['threadIndex'];

    List<dynamic> jThreadCallstack = jThreadObj['callstack'];

    for (var jElTok in jThreadCallstack) {
      Map<String, dynamic> jElementObj = jElTok;

      var pushPopType = PushPopType.values[jElementObj['type']];

      var pointer = Pointer.Null;

      String? currentContainerPathStr;
      dynamic currentContainerPathStrToken;
      currentContainerPathStrToken = jElementObj['cPath'];
      if (currentContainerPathStrToken != null) {
        currentContainerPathStr = currentContainerPathStrToken.toString();

        var threadPointerResult =
            storyContext.contentAtPath(Path(currentContainerPathStr));
        pointer = Pointer(
            container: threadPointerResult.container,
            index: jElementObj['idx'] ?? -1);

        if (threadPointerResult.obj == null) {
          throw Exception(
              "When loading state, internal story location couldn't be found: $currentContainerPathStr . Has the story changed since this save data was created?");
        } else if (threadPointerResult.approximate) {
          storyContext.warning(
              "When loading state, exact internal story location couldn't be found: '$currentContainerPathStr', so it was approximated to '" +
                  pointer.container!.path.toString() +
                  "' to recover. Has the story changed since this save data was created?");
        }
      }

      bool inExpressionEvaluation = jElementObj['exp'];
      var el = Element(pushPopType, pointer,
          inExpressionEvaluation: inExpressionEvaluation);

      if (jElementObj.containsKey('temp')) {
        el.temporaryVariables =
            JsonSerialization.JObjectToDictionaryRuntimeObjs(
                jElementObj['temp']);
      } else {
        el.temporaryVariables.clear();
      }
      callstack.add(el);
    }
    if (jThreadObj.containsKey('previousContentObject')) {
      var prevPath = Path(jThreadObj['previousContentObject']);
      previousPointer = storyContext.pointerAtPath(prevPath);
    }
  }

  Thread copy() {
    var copy = Thread();
    copy.threadIndex = threadIndex;
    for (var e in callstack) {
      copy.callstack.add(e.copy());
    }
    copy.previousPointer = previousPointer;
    return copy;
  }

  void WriteJson(SimpleJsonWriter writer) {
    writer.WriteObjectStart();

    // callstack
    writer.WritePropertyStart("callstack");
    writer.WriteArrayStart();
    for (Element el in callstack) {
      writer.WriteObjectStart();
      if (!el.currentPointer.isNull) {
        writer.WriteProperty(
            "cPath", el.currentPointer.container!.path.componentsString);
        writer.WriteProperty("idx", el.currentPointer.index);
      }

      writer.WriteProperty("exp", el.inExpressionEvaluation);
      writer.WriteProperty("type", el.type.index);

      if (el.temporaryVariables.isNotEmpty) {
        writer.WritePropertyStart("temp");
        JsonSerialization.WriteDictionaryRuntimeObjs(
            writer, el.temporaryVariables);
        writer.WritePropertyEnd();
      }

      writer.WriteObjectEnd();
    }
    writer.WriteArrayEnd();
    writer.WritePropertyEnd();

    // threadIndex
    writer.WriteProperty("threadIndex", threadIndex);

    if (!previousPointer.isNull) {
      writer.WriteProperty(
          "previousContentObject", previousPointer.resolve()!.path.toString());
    }

    writer.WriteObjectEnd();
  }
}
