import 'package:meta/meta.dart';
import 'package:quiver/strings.dart';

import 'extension.dart';
import 'logger.dart';
import 'named_content.dart';
import 'path.dart';
import 'runtime_object.dart';
import 'search_result.dart';

class CountFlags {
  static final int Visits = 1;
  static final int Turns = 2;
  static final int CountStartOnly = 4;
}

final _logger = Logger('Container');

class Container extends RuntimeObject with NamedContent {
  final List<RuntimeObject> _content = [];
  Map<String, NamedContent> namedContent = {};
  @override
  bool visitsShouldBeCounted = false;
  @override
  bool turnIndexShouldBeCounted = false;
  @override
  bool countingAtStartOnly = false;

  @override
  List<RuntimeObject> get content => _content;

  @override
  String name = '';

  Container();

  void addContents(Iterable<RuntimeObject> contentList) {
    for (var c in contentList) {
      addContent(c);
    }
  }

  void addContent(RuntimeObject contentObj) {
    content.add(contentObj);
    if (contentObj.parent != null) {
      throw Exception('content is already in ${contentObj.parent}');
    }
    contentObj.parent = this;
    tryAddNamedContent(contentObj);
  }

  void insertContent(RuntimeObject contentObj, int index) {
    content.insert(index, contentObj);
    if (contentObj.parent != null) {
      throw Exception('content is already in ${contentObj.parent}');
    }
    contentObj.parent = this;
    tryAddNamedContent(contentObj);
  }

  void tryAddNamedContent(RuntimeObject contentObj) {
    var namedContentObj = asOrNull<NamedContent>(contentObj);
    if (namedContentObj != null && namedContentObj.hasValidName) {
      addToNamedContentOnly(namedContentObj);
    }
  }

  @override
  SearchResult contentAtPath(Path path,
      {int partialPathStart = 0, int partialPathLength = -1}) {
    _logger.finest('contentAtPath ( path : $path )');

    if (partialPathLength == -1) {
      partialPathLength = path.length;
    }
    _logger.finest('partialPathLength ( $partialPathLength )');

    var result = SearchResult();
    result.approximate = false;

    Container? currentContainer = this;
    RuntimeObject currentObj = this;

    for (var i = partialPathStart; i < partialPathLength; ++i) {
      var comp = path.getComponent(i);

      _logger.finest('comp : $comp');

      // Path component was wrong type
      if (currentContainer == null) {
        result.approximate = true;
        break;
      }

      var foundObj = currentContainer.contentWithPathComponent(comp);

      // Couldn't resolve entire path?
      if (foundObj == null) {
        _logger.finest("Couldn't resolve entire path?");
        result.approximate = true;
        break;
      }

      currentObj = foundObj;
      currentContainer = asOrNull<Container>(foundObj);
    }

    result.obj = currentObj;

    return result;
  }

//protected Runtime.Object ContentWithPathComponent(Path.Component component)
  @protected
  RuntimeObject? contentWithPathComponent(Component component) {
    if (component.isIndex) {
      if (component.index >= 0 && component.index < content.length) {
        return content[component.index];
      }

      // When path is out of range, quietly return nil
      // (useful as we step/increment forwards through content)
      else {
        return null;
      }
    } else if (component.isParent) {
      return parent;
    } else {
      NamedContent? foundContent = namedContent[component.name];
      if (foundContent != null) {
        return foundContent as RuntimeObject;
      } else {
        return null;
      }
    }
  }

  @override
  bool get hasValidName => isNotBlank(name);

  void addToNamedContentOnly(NamedContent namedContentObj) {
    assert(namedContentObj is RuntimeObject);
    var runtimeObj = namedContentObj as RuntimeObject;
    runtimeObj.parent = this;

    namedContent[namedContentObj.name] = namedContentObj;
  }

  @override
  int get countFlags {
    var flags = 0;
    if (visitsShouldBeCounted) flags |= CountFlags.Visits;
    if (turnIndexShouldBeCounted) flags |= CountFlags.Turns;
    if (countingAtStartOnly) flags |= CountFlags.CountStartOnly;
    // If we're only storing CountStartOnly, it serves no purpose,
    // since it's dependent on the other two to be used at all.
    // (e.g. for setting the fact that *if* a gather or choice's
    // content is counted, then is should only be counter at the start)
    // So this is just an optimisation for storage.
    if (flags == CountFlags.CountStartOnly) {
      flags = 0;
    }
    return flags;
  }

  @override
  set countFlags(int value) {
    if ((value & CountFlags.Visits) > 0) visitsShouldBeCounted = true;
    if ((value & CountFlags.Turns) > 0) turnIndexShouldBeCounted = true;
    if ((value & CountFlags.CountStartOnly) > 0) countingAtStartOnly = true;
  }

  @override
  Map<String, RuntimeObject>? get namedOnlyContent {
    var namedOnlyContentDict = <String, RuntimeObject>{};
    namedContent.forEach((key, value) {
      namedOnlyContentDict[key] = value as RuntimeObject;
    });

    for (var c in content) {
      var named = asOrNull<NamedContent>(c);
      if (named != null && named.hasValidName) {
        namedOnlyContentDict.remove(named.name);
      }
    }

    if (namedOnlyContentDict.isEmpty) {
      // namedOnlyContentDict = null;
      return null;
    }

    return namedOnlyContentDict;
  }

  @override
  set namedOnlyContent(Map<String, RuntimeObject>? value) {
    var existingNamedOnly = namedOnlyContent;
    if (existingNamedOnly != null) {
      existingNamedOnly.forEach((key, value) {
        namedContent.remove(key);
      });
    }

    if (value == null) {
      return;
    }

    value.forEach((key, value) {
      var named = asOrNull<NamedContent>(value);
      if (named != null) {
        addToNamedContentOnly(named);
      }
    });
  }

  Path? _pathToFirstLeafContent;

  Path get internalPathToFirstLeafContent {
    var components = <Component>[];
    Container? container = this;
    while (container != null) {
      if (container.content.isNotEmpty) {
        components.add(Component.index(0));
        container = asOrNull<Container>(container.content[0]);
      }
    }
    return Path(components);
  }

  @override
  void AddContentsOfContainer(Container otherContainer) {
    content.addAll(otherContainer.content);
    for (var obj in otherContainer.content) {
      obj.parent = this;
      tryAddNamedContent(obj);
    }
  }

  @override
  void AddContent(List<RuntimeObject> contentList) => addContents(contentList);

  @override
  void AddToNamedContentOnly(NamedContent namedContentObj) =>
      addToNamedContentOnly(namedContentObj);

  @override
  void InsertContent(RuntimeObject contentObj, int index) =>
      insertContent(contentObj, index);

  @override
  void TryAddNamedContent(RuntimeObject contentObj) =>
      tryAddNamedContent(contentObj);

  @override
  Path? get pathToFirstLeafContent => _pathToFirstLeafContent;

  @override
  set content(List<RuntimeObject> value) => AddContent(value);
}
