import 'container.dart';
import 'logger.dart';
import 'path.dart';
import 'pointer.dart';
import 'push_pop.dart';
import 'runtime_object.dart';

final _logger = Logger('Divert');

class Divert extends RuntimeObject {
  Path? _targetPath;
  Pointer _targetPointer = Pointer.Null;

  Path? get targetPath {
    // Resolve any relative paths to global ones as we come across them
    if (_targetPath != null && _targetPath!.isRelative) {
      var targetObj = targetPointer.resolve();

      if (targetObj != null) {
        _targetPath = targetObj.path;
      }
    }
    return _targetPath;
  }

  set targetPath(value) {
    _targetPath = value;
    _targetPointer = Pointer.Null;
  }

  Pointer get targetPointer {
    _logger.finest('[get targetPointer]');

    if (_targetPointer.isNull) {
      var targetObj = resolvePath(_targetPath!).obj;

      if (_targetPath!.lastComponent!.isIndex) {
        _targetPointer.container = targetObj!.parent as Container;
        _targetPointer.index = _targetPath!.lastComponent!.index;
      } else {
        _targetPointer = Pointer.startOf(targetObj as Container);
      }
    }
    return _targetPointer.copyWith();
  }

  String? get targetPathString {
    if (targetPath == null) {
      return null;
    } else {
      return compactPathString(targetPath!);
    }
  }

  set targetPathString(String? value) {
    if (value == null) {
      targetPath = null;
    } else {
      targetPath = Path(value);
    }
  }

  String? variableDivertName;

  bool get hasVariableTarget => variableDivertName != null;

  bool pushesToStack = false;

  PushPopType? stackPushType;

  bool isExternal = false;

  int externalArgs = 0;

  bool isConditional = false;

  Divert([PushPopType? stackPushType]) : super() {
    pushesToStack = false;
    if (stackPushType != null) {
      pushesToStack = true;
      this.stackPushType = stackPushType;
    }
  }

  @override
  bool operator ==(other) {
    if (other is Divert) {
      if (hasVariableTarget == other.hasVariableTarget) {
        if (hasVariableTarget) {
          return variableDivertName == other.variableDivertName;
        } else {
          return targetPath == (other.targetPath);
        }
      }
    }
    return false;
  }

  @override
  int get hashCode {
    if (hasVariableTarget) {
      const variableTargetSalt = 12345;
      return variableDivertName.hashCode + variableTargetSalt;
    } else {
      const pathTargetSalt = 54321;
      return targetPath.hashCode + pathTargetSalt;
    }
  }

  @override
  String toString() {
    if (hasVariableTarget) {
      return 'Divert(variable: $variableDivertName)';
    } else if (targetPath == null) {
      return 'Divert(null)';
    } else {
      var sb = StringBuffer();

      var targetStr = targetPath.toString();
      var targetLineNum = debugLineNumberOfPath(targetPath);
      _logger.finest('targetLineNum : $targetLineNum');
      if (targetLineNum != null) {
        targetStr = 'line $targetLineNum';
      }

      sb.write('Divert');

      if (isConditional) sb.write('?');

      if (pushesToStack) {
        if (stackPushType == PushPopType.Function) {
          sb.write(' function');
        } else {
          sb.write(' tunnel');
        }
      }

      sb.write(' -> ');
      sb.write(targetPathString);

      sb.write(' (');
      sb.write(targetStr);
      sb.write(')');

      return sb.toString();
    }
  }
}
