import 'package:freezed_annotation/freezed_annotation.dart';

import 'list_definition.dart';
import 'story.dart';

part 'ink_list.freezed.dart';

/// <summary>
/// The underlying type for a list item in ink. It stores the original list definition
/// name as well as the item name, but without the value of the item. When the value is
/// stored, it's stored in a KeyValuePair of InkListItem and int.
/// </summary>
@freezed
class InkListItem with _$InkListItem {
  const InkListItem._();

  /// <summary>
  /// Create an item with the given original list definition name, and the name of this
  /// item.
  /// </summary>
  /// The name of the list where the item was originally defined.
  /// The main name of the item as defined in ink.
  factory InkListItem({String? originName, String? itemName}) = _InkListItem;

  /// <summary>
  /// Create an item from a dot-separted string of the form "listDefinitionName.listItemName".
  /// </summary>
  factory InkListItem.fullName(String fullName) {
    var nameParts = fullName.split('.');
    return InkListItem(originName: nameParts[0], itemName: nameParts[1]);
  }

  static InkListItem get Null => InkListItem();

  bool get isNull => originName == null && itemName == null;

  /// Get the full dot-separated name of the item, in the form "listDefinitionName.itemName".
  String get fullName => '${(originName ?? '?')}.$itemName';

  /// <summary>
  /// Get the full dot-separated name of the item, in the form "listDefinitionName.itemName".
  /// Calls fullName internally.
  /// </summary>
  @override
  String toString() {
    return fullName;
  }

  /// <summary>
  /// Get the hashcode for an item.
  /// </summary>
  @override
  int get hashCode {
    var originCode = 0;
    var itemCode = itemName.hashCode;
    if (originName != null) {
      originCode = originName.hashCode;
    }
    return originCode + itemCode;
  }
}

/// The InkList is the underlying type that's used to store an instance of a
/// list in ink. It's not used for the *definition* of the list, but for a list
/// value that's stored in a variable.
/// Somewhat confusingly, it's backed by a C# Dictionary, and has nothing to
/// do with a C# List!
class InkList implements Map<InkListItem, int> {
  final Map<InkListItem, int> _map;

  List<String>? _originNames;

  // Story has to set this so that the value knows its origin,
  // necessary for certain operations (e.g. interacting with ints).
  // Only the story has access to the full set of lists, so that
  // the origin can be resolved from the originListName.
  List<ListDefinition>? _origins;

  List<ListDefinition>? get origins => _origins;

  //_origins == null ? null : List.unmodifiable(_origins!);

  set origins(value) => _origins = value;

  InkList() : _map = {};

  InkList.of(InkList otherList) : _map = Map.of(otherList) {
    var otherOriginNames = otherList.originNames;
    if (otherOriginNames != null) {
      _originNames = List.of(otherOriginNames);
    }
    if (otherList._origins != null) {
      _origins = List<ListDefinition>.of(otherList._origins!);
    }
  }

  InkList.fromOrigin(String singleOriginListName, Story originStory)
      : _map = {} {
    setInitialOriginName(singleOriginListName);
    ArgumentError.checkNotNull(
        originStory.listDefinitions, 'originStory.listDefinitions');
    var def = originStory.listDefinitions!.getDefinition(singleOriginListName);
    if (def == null) {
      throw Exception(
          'InkList origin could not be found in story when constructing new list: $singleOriginListName');
    }
    _origins = List<ListDefinition>.of([def]);
  }

  InkList.fromEntry(MapEntry<InkListItem, int> singleElement) : _map = {} {
    this[singleElement.key] = singleElement.value;
  }

  /// <summary>
  /// Converts a string to an ink list and returns for use in the story.
  /// </summary>
  /// <returns>InkList created from string list item</returns>
  /// <param name="itemKey">Item key.</param>
  /// <param name="originStory">Origin story.</param>
  factory InkList.fromString(String myListItem, Story originStory) {
    var listValue =
        originStory.listDefinitions?.findSingleItemListWithName(myListItem);
    if (listValue != null) {
      return InkList.of(listValue.value);
    } else {
      throw Exception("Could not find the InkListItem from the string '" +
          myListItem +
          "' to create an InkList because it doesn't exist in the original list definition in ink.");
    }
  }

  // public void SetInitialOriginName (string initialOriginName)
  void setInitialOriginName(String value) => _originNames = [value];

  //public void SetInitialOriginNames (List<string> initialOriginNames)
  void setInitialOriginNames(List<String>? value) {
    if (value == null) {
      _originNames = null;
    } else {
      _originNames = List<String>.of(value);
    }
  }

  void addItem(dynamic item) {
    if (item is String) {
      _addItemWithName(item);
    } else if (item is InkListItem) {
      _addItem(item);
    }
    throw Exception('not supported type');
  }

  /// <summary>
  /// Adds the given item to the ink list. Note that the item must come from a list definition that
  /// is already "known" to this list, so that the item's value can be looked up. By "known", we mean
  /// that it already has items in it from that source, or it did at one point - it can't be a
  /// completely fresh empty list, or a list that only contains items from a different list definition.
  /// </summary>
  void _addItem(InkListItem item) {
    if (item.originName == null) {
      addItem(item.itemName);
      return;
    }
    for (var origin in _origins!) {
      if (origin.name == item.originName) {
        var intVal = origin.getValueForItem(item);
        if (intVal != null) {
          this[item] = intVal;
          return;
        } else {
          throw Exception(
              "Could not add the item $item to this list because it doesn't exist in the original list definition in ink.");
        }
      }
    }
    throw Exception(
        "Failed to add item to list because the item was from a new list definition that wasn't previously known to this list. Only items from previously known lists can be used, so that the int value can be found.");
  }

  /// <summary>
  /// Adds the given item to the ink list, attempting to find the origin list definition that it belongs to.
  /// The item must therefore come from a list definition that is already "known" to this list, so that the
  /// item's value can be looked up. By "known", we mean that it already has items in it from that source, or
  /// it did at one point - it can't be a completely fresh empty list, or a list that only contains items from
  /// a different list definition.
  /// </summary>
  void _addItemWithName(String itemName) {
    ListDefinition? foundListDef;

    for (var origin in _origins!) {
      if (origin.containsItemWithName(itemName)) {
        if (foundListDef != null) {
          throw Exception(
              'Could not add the item $itemName to this list because it could come from either ${origin.name} or ${foundListDef.name}');
        } else {
          foundListDef = origin;
        }
      }
    }
    ;

    if (foundListDef == null) {
      throw Exception(
          "Could not add the item $itemName to this list because it isn't known to any list definitions previously associated with this list.");
    }

    var item = InkListItem(originName: foundListDef.name, itemName: itemName);
    var itemVal = foundListDef.valueForItem(item);
    this[item] = itemVal;
  }

  /// Returns true if this ink list contains an item with the given short name
  /// (ignoring the original list where it was defined).
  bool containsItemNamed(String itemName) {
    for (var entry in entries) {
      if (entry.key.itemName == itemName) {
        return true;
      }
    }
    return false;
  }

  /// Get the maximum item in the list, equivalent to calling LIST_MAX(list) in ink.
  MapEntry<InkListItem, int> get maxItem {
    var max = MapEntry<InkListItem, int>(InkListItem(), 0);

    for (var entry in entries) {
      if (max.key.isNull || entry.value > max.value) {
        max = entry;
      }
    }

    return max;
  }

  /// <summary>
  /// Get the minimum item in the list, equivalent to calling LIST_MIN(list) in ink.
  /// </summary>
  MapEntry<InkListItem, int> get minItem {
    var min = MapEntry<InkListItem, int>(InkListItem(), 0);

    entries.forEach((entry) {
      if (min.key.isNull || entry.value < min.value) {
        min = entry;
      }
    });

    return min;
  }

  /// The inverse of the list, equivalent to calling LIST_INVERSE(list) in ink
  InkList get inverse {
    var list = InkList();
    if (_origins != null) {
      for (var origin in origins!) {
        for (var itemAndValue in origin.items.entries) {
          if (!containsKey(itemAndValue.key)) {
            list[itemAndValue.key] = itemAndValue.value;
          }
        }
      }
    }
    return list;
  }

  /// The list of all items from the original list definition, equivalent to calling
  /// LIST_ALL(list) in ink.
  InkList get all {
    var list = InkList();
    if (_origins! != null) {
      for (var origin in origins!) {
        for (var itemAndValue in origin.items.entries) {
          list[itemAndValue.key] = itemAndValue.value;
        }
      }
    }
    return list;
  }

  /// <summary>
  /// Returns a new list that is the combination of the current list and one that's
  /// passed in. Equivalent to calling (list1 + list2) in ink.
  /// </summary>
  InkList union(InkList otherList) {
    var union = InkList.of(this);
    union.addAll(otherList);

    return union;
  }

  /// <summary>
  /// Returns a new list that is the intersection of the current list with another
  /// list that's passed in - i.e. a list of the items that are shared between the
  /// two other lists. Equivalent to calling (list1 ^ list2) in ink.
  /// </summary>
  InkList intersect(InkList otherList) {
    var intersection = InkList();
    for (var kv in entries) {
      if (otherList.containsKey(kv.key)) {
        intersection[kv.key] = kv.value;
      }
    }

    return intersection;
  }

  /// <summary>
  /// Returns true if the current list contains all the items that are in the list that
  /// is passed in. Equivalent to calling (list1 ? list2) in ink.
  /// </summary>
  /// <param name="otherList">Other list.</param>
  bool contains(InkList otherList) {
    if (otherList.isEmpty || isEmpty) return false;
    for (var kv in otherList.entries) {
      if (!containsKey(kv.key)) {
        return false;
      }
    }
    return true;
  }

  /// <summary>
  /// Returns true if all the item values in the current list are greater than all the
  /// item values in the passed in list. Equivalent to calling (list1 > list2) in ink.
  /// </summary>
  bool greaterThan(InkList otherList) {
    if (isEmpty) return false;
    if (otherList.isEmpty) return true;

    // All greater
    return minItem.value > otherList.maxItem.value;
  }

  /// <summary>
  /// Returns true if the item values in the current list overlap or are all greater than
  /// the item values in the passed in list. None of the item values in the current list must
  /// fall below the item values in the passed in list. Equivalent to (list1 >= list2) in ink,
  /// or LIST_MIN(list1) >= LIST_MIN(list2) &amp;&amp; LIST_MAX(list1) >= LIST_MAX(list2).
  /// </summary>
  bool greaterThanOrEquals(InkList otherList) {
    if (isEmpty) return false;
    if (otherList.isEmpty) return true;

    return minItem.value >= otherList.minItem.value &&
        maxItem.value >= otherList.maxItem.value;
  }

  /// <summary>
  /// Returns true if all the item values in the current list are less than all the
  /// item values in the passed in list. Equivalent to calling (list1 &lt; list2) in ink.
  /// </summary>
  bool lessThan(InkList otherList) {
    if (otherList.isEmpty) return false;
    if (isEmpty) return true;

    return maxItem.value < otherList.minItem.value;
  }

  /// <summary>
  /// Returns a new list that's the same as the current one, except with the given items
  /// removed that are in the passed in list. Equivalent to calling (list1 - list2) in ink.
  /// </summary>
  /// <param name="listToRemove">List to remove.</param>
  InkList without(InkList listToRemove) {
    var result = InkList.of(this);
    result.removeWhere((k, v) => listToRemove.containsKey(k));
    return result;
  }

  /// <summary>
  /// Returns true if the item values in the current list overlap or are all less than
  /// the item values in the passed in list. None of the item values in the current list must
  /// go above the item values in the passed in list. Equivalent to (list1 &lt;= list2) in ink,
  /// or LIST_MAX(list1) &lt;= LIST_MAX(list2) &amp;&amp; LIST_MIN(list1) &lt;= LIST_MIN(list2).
  /// </summary>
  bool lessThanOrEquals(InkList otherList) {
    if (otherList.isEmpty) return false;
    if (isEmpty) return true;

    return maxItem.value <= otherList.maxItem.value &&
        minItem.value <= otherList.minItem.value;
  }

  InkList maxAsList() {
    if (isNotEmpty) {
      return InkList.fromEntry(maxItem);
    } else {
      return InkList();
    }
  }

  InkList MinAsList() {
    if (isNotEmpty) {
      return InkList.fromEntry(minItem);
    } else {
      return InkList();
    }
  }

  /// <summary>
  /// Returns a sublist with the elements given the minimum and maxmimum bounds.
  /// The bounds can either be ints which are indices into the entire (sorted) list,
  /// or they can be InkLists themselves. These are intended to be single-item lists so
  /// you can specify the upper and lower bounds. If you pass in multi-item lists, it'll
  /// use the minimum and maximum items in those lists respectively.
  /// WARNING: Calling this method requires a full sort of all the elements in the list.
  /// </summary>
  InkList listWithSubRange(dynamic minBound, dynamic maxBound) {
    if (isEmpty) {
      return InkList();
    }

    var ordered = orderedItems;

    int minValue = 0;
    int maxValue = 1 << 32;

    if (minBound is int) {
      minValue = minBound;
    } else {
      if (minBound is InkList && minBound.isNotEmpty) {
        minValue = minBound.minItem.value;
      }
    }

    if (maxBound is int) {
      maxValue = maxBound;
    } else {
      if (minBound is InkList && minBound.isNotEmpty) {
        maxValue = maxBound.maxItem.value;
      }
    }

    var subList = InkList();
    subList.setInitialOriginNames(originNames);
    for (var item in ordered) {
      if (item.value >= minValue && item.value <= maxValue) {
        subList[item.key] = item.value;
      }
    }

    return subList;
  }

  /// <summary>
  /// Returns true if the passed object is also an ink list that contains
  /// the same items as the current list, false otherwise.
  /// </summary>
  @override
  bool operator ==(Object other) {
    //if (other == null) return false;
    var otherRawList = other as InkList;
    if (otherRawList.length != length) return false;

    for (var kv in entries) {
      if (!otherRawList.containsKey(kv.key)) {
        return false;
      }
    }

    return true;
  }

  /// <summary>
  /// Return the hashcode for this object, used for comparisons and inserting into dictionaries.
  /// </summary>
  @override
  int get hashCode {
    var ownHash = 0;
    forEach((k, v) {
      ownHash += k.hashCode;
    });
    return ownHash;
  }

  List<MapEntry<InkListItem, int>> get orderedItems {
    var ordered = <MapEntry<InkListItem, int>>[];
    ordered.addAll(entries);
    ordered.sort((x, y) {
      // Ensure consistent ordering of mixed lists.
      if (x.value == y.value) {
        return x.key.originName!.compareTo(y.key.originName!);
      } else {
        return x.value.compareTo(y.value);
      }
    });
    return ordered;
  }

  /// <summary>
  /// Returns a string in the form "a, b, c" with the names of the items in the list, without
  /// the origin list definition names. Equivalent to writing {list} in ink.
  /// </summary>
  @override
  String toString() {
    var ordered = orderedItems;

    var sb = StringBuffer();
    for (var i = 0; i < ordered.length; i++) {
      if (i > 0) sb.write(', ');

      var item = ordered[i].key;
      sb.write(item.itemName);
    }
    return sb.toString();
  }

  ListDefinition? get originOfMaxItem {
    if (_origins == null) return null;

    var maxOriginName = maxItem.key.originName;
    for (var origin in _origins!) {
      if (origin.name == maxOriginName) {
        return origin;
      }
    }

    return null;
  }

  // Origin name needs to be serialised when content is empty,
  // assuming a name is availble, for list definitions with variable
  // that is currently empty.
  List<String>? get originNames {
    if (isNotEmpty) {
      if (_originNames == null && isNotEmpty) {
        _originNames = <String>[];
      } else {
        _originNames!.clear();
      }
      forEach((k, v) {
        _originNames!.add(k.originName!);
      });
    }
    return _originNames;
  }

  @override
  int? operator [](Object? key) => _map[key];

  @override
  void operator []=(InkListItem key, int value) => _map[key] = value;

  @override
  void addAll(Map<InkListItem, int> other) => _map.addAll(other);

  @override
  void addEntries(Iterable<MapEntry<InkListItem, int>> newEntries) =>
      _map.addEntries(newEntries);

  @override
  Map<RK, RV> cast<RK, RV>() => _map.cast();

  @override
  void clear() => _map.clear();

  @override
  bool containsKey(Object? key) => _map.containsKey(key);

  @override
  bool containsValue(Object? value) => _map.containsValue(value);

  @override
  Iterable<MapEntry<InkListItem, int>> get entries => _map.entries;

  @override
  void forEach(void Function(InkListItem key, int value) f) => _map.forEach(f);

  @override
  bool get isEmpty => _map.isEmpty;

  @override
  bool get isNotEmpty => _map.isNotEmpty;

  @override
  Iterable<InkListItem> get keys => _map.keys;

  @override
  int get length => _map.length;

  int get count => length;

  @override
  Map<K2, V2> map<K2, V2>(
          MapEntry<K2, V2> Function(InkListItem key, int value) f) =>
      _map.map(f);

  @override
  int putIfAbsent(InkListItem key, int Function() ifAbsent) =>
      _map.putIfAbsent(key, ifAbsent);

  @override
  int? remove(Object? key) => _map.remove(key);

  @override
  void removeWhere(bool Function(InkListItem key, int value) predicate) =>
      _map.removeWhere(predicate);

  @override
  int update(InkListItem key, int Function(int value) update,
          {int Function()? ifAbsent}) =>
      _map.update(key, update, ifAbsent: ifAbsent);

  @override
  void updateAll(int Function(InkListItem key, int value) update) =>
      _map.updateAll(update);

  @override
  Iterable<int> get values => _map.values;
}
