
import 'choice.dart';
import 'choice_point.dart';
import 'container.dart';
import 'control_command.dart';
import 'divert.dart';
import 'extension.dart';
import 'glue.dart';
import 'ink_list.dart';
import 'list_definition.dart';
import 'list_definitions_origin.dart';
import 'native_function_call.dart';
import 'path.dart';
import 'push_pop.dart';
import 'runtime_object.dart';
import 'runtime_void.dart';
import 'simple_json.dart';
import 'tag.dart';
import 'value.dart';
import 'variable_assignment.dart';
import 'variable_reference.dart';

class JsonSerialization {
  static List<String>? _controlCommandNames;
  static final JsonSerialization _singvaron = JsonSerialization._();

  static List<String> get controlCommandNames {
    if (_controlCommandNames == null) {
      _createControlCommandNames();
    }
    return _controlCommandNames!;
  }

  factory JsonSerialization() {
    return _singvaron;
  }

  static void _createControlCommandNames() {
    _controlCommandNames = List.filled(CommandType.values.length, '');
    _controlCommandNames![CommandType.EvalStart.index] = 'ev';
    _controlCommandNames![CommandType.EvalOutput.index] = 'out';
    _controlCommandNames![CommandType.EvalEnd.index] = '/ev';
    _controlCommandNames![CommandType.Duplicate.index] = 'du';
    _controlCommandNames![CommandType.PopEvaluatedValue.index] = 'pop';
    _controlCommandNames![CommandType.PopFunction.index] = '~ret';
    _controlCommandNames![CommandType.PopTunnel.index] = '->->';
    _controlCommandNames![CommandType.BeginString.index] = 'str';
    _controlCommandNames![CommandType.EndString.index] = '/str';
    _controlCommandNames![CommandType.NoOp.index] = 'nop';
    _controlCommandNames![CommandType.ChoiceCount.index] = 'choiceCnt';
    _controlCommandNames![CommandType.Turns.index] = 'turn';
    _controlCommandNames![CommandType.TurnsSince.index] = 'turns';
    _controlCommandNames![CommandType.ReadCount.index] = 'readc';
    _controlCommandNames![CommandType.Random.index] = 'rnd';
    _controlCommandNames![CommandType.SeedRandom.index] = 'srnd';
    _controlCommandNames![CommandType.VisitIndex.index] = 'visit';
    _controlCommandNames![CommandType.SequenceShuffleIndex.index] = 'seq';
    _controlCommandNames![CommandType.StartThread.index] = 'thread';
    _controlCommandNames![CommandType.Done.index] = 'done';
    _controlCommandNames![CommandType.End.index] = 'end';
    _controlCommandNames![CommandType.ListFromInt.index] = 'listInt';
    _controlCommandNames![CommandType.ListRange.index] = 'range';
    _controlCommandNames![CommandType.ListRandom.index] = 'lrnd';
    _controlCommandNames![CommandType.BeginTag.index] = '#';
    _controlCommandNames![CommandType.EndTag.index] = '/#';

    for (var i = 1; i < CommandType.TOTAL_VALUES.index; ++i) {
      if (_controlCommandNames![i] == null) {
        throw Exception('Control command not accounted for in serialisation');
      }
    }
  }

  JsonSerialization._();

  static JArrayToRuntimeObjList<T extends RuntimeObject>(List<dynamic> jArray,
      [bool skipLast = false]) {
    var count = jArray.length;
    if (skipLast) count--;

    var list = <T>[];

    for (var i = 0; i < count; i++) {
      var jTok = jArray[i];
      var runtimeObj = JTokenToRuntimeObject(jTok);
      if (runtimeObj == null) {
        return ArgumentError.notNull('runtimeObj');
      }
      list.add(runtimeObj as T);
    }

    return list;
  }

  static WriteDictionaryRuntimeObjs(
      SimpleJsonWriter writer, Map<String, RuntimeObject> dictionary) {
    writer.WriteObjectStart();
    for (var kv in dictionary.entries) {
      writer.WritePropertyStart(kv.key);
      WriteRuntimeObject(writer, kv.value);
      writer.WritePropertyEnd();
    }
    writer.WriteObjectEnd();
  }

  static WriteListRuntimeObjs(
      SimpleJsonWriter writer, List<RuntimeObject> list) {
    writer.WriteArrayStart();
    for (var value in list) {
      WriteRuntimeObject(writer, value);
    }
    writer.WriteArrayEnd();
  }

  static WriteIntDictionary(SimpleJsonWriter writer, Map<String, int> dict) {
    writer.WriteObjectStart();
    for (var kv in dict.entries) {
      writer.WriteIntProperty(kv.key, kv.value);
    }
    writer.WriteObjectEnd();
  }

  static WriteRuntimeObject(SimpleJsonWriter writer, RuntimeObject obj) {
    var container = asOrNull<Container>(obj);
    if (container != null) {
      WriteRuntimeContainer(writer, container);
      return;
    }

    var divert = asOrNull<Divert>(obj);
    if (divert != null) {
      var divTypeKey = "->";
      if (divert.isExternal) {
        divTypeKey = "x()";
      } else if (divert.pushesToStack) {
        if (divert.stackPushType == PushPopType.Function) {
          divTypeKey = "f()";
        } else if (divert.stackPushType == PushPopType.Tunnel) {
          divTypeKey = "->t->";
        }
      }

      var targetStr;
      if (divert.hasVariableTarget) {
        targetStr = divert.variableDivertName;
      } else {
        targetStr = divert.targetPathString;
      }

      writer.WriteObjectStart();
      writer.WriteProperty(divTypeKey, targetStr);

      if (divert.hasVariableTarget) {
        writer.WriteProperty("var", true);
      }

      if (divert.isConditional) {
        writer.WriteProperty("c", true);
      }

      if (divert.externalArgs > 0) {
        writer.WriteIntProperty("exArgs", divert.externalArgs);
      }

      writer.WriteObjectEnd();
      return;
    }

    var choicePoint = asOrNull<ChoicePoint>(obj);
    if (choicePoint != null) {
      writer.WriteObjectStart();
      writer.WriteProperty("*", choicePoint.pathStringOnChoice);
      writer.WriteIntProperty("flg", choicePoint.flags);
      writer.WriteObjectEnd();
      return;
    }

    var boolVal = asOrNull<BoolValue>(obj);
    if (boolVal != null) {
      writer.WriteBool(boolVal.value);
      return;
    }

    var intVal = asOrNull<IntValue>(obj);
    if (intVal != null) {
      writer.WriteInt(intVal.value);
      return;
    }

    var floatVal = asOrNull<FloatValue>(
      obj,
    );
    if (floatVal != null) {
      writer.WriteFloat(floatVal.value);
      return;
    }

    var strVal = asOrNull<StringValue>(obj);
    if (strVal != null) {
      if (strVal.isNewLine) {
        writer.Write("\n", escape: false);
      } else {
        writer.WriteStringStart();
        writer.WriteStringInner("^");
        writer.WriteStringInner(strVal.value);
        writer.WriteStringEnd();
      }
      return;
    }

    var listVal = asOrNull<ListValue>(obj);
    if (listVal != null) {
      WriteInkList(writer, listVal);
      return;
    }

    var divTargetVal = asOrNull<DivertTargetValue>(obj);
    if (divTargetVal != null) {
      writer.WriteObjectStart();
      if (divTargetVal.value == null) {
        return ArgumentError.notNull("divTargetVal.value");
      }
      writer.WriteProperty("^->", divTargetVal.value.componentsString);
      writer.WriteObjectEnd();

      return;
    }

    var varPtrVal = asOrNull<VariablePointerValue>(obj);
    if (varPtrVal != null) {
      writer.WriteObjectStart();
      writer.WriteProperty("^var", varPtrVal.value);
      writer.WriteIntProperty("ci", varPtrVal.contextIndex);
      writer.WriteObjectEnd();
      return;
    }

    var glue = asOrNull<Glue>(obj);
    if (glue != null) {
      writer.Write("<>");
      return;
    }

    var controlCmd = asOrNull<ControlCommand>(obj);
    if (controlCmd != null) {
      writer.Write(
          JsonSerialization.controlCommandNames[controlCmd.commandType.index]);
      return;
    }

    var nativeFunc = asOrNull<NativeFunctionCall>(obj);
    if (nativeFunc != null) {
      var name = nativeFunc.name;

      if (name == "^") name = "L^";

      writer.Write(name);
      return;
    }

    var varRef = asOrNull<VariableReference>(obj);
    if (varRef != null) {
      writer.WriteObjectStart();
      var readCountPath = varRef.pathStringForCount;
      if (readCountPath != null) {
        writer.WriteProperty("CNT?", readCountPath);
      } else {
        writer.WriteProperty("VAR?", varRef.name);
      }

      writer.WriteObjectEnd();
      return;
    }

    var varAss = asOrNull<VariableAssignment>(obj);
    if (varAss != null) {
      writer.WriteObjectStart();

      var key = varAss.isGlobal ? "VAR=" : "temp=";
      writer.WriteProperty(key, varAss.variableName);

      // Reassignment?
      if (!varAss.isNewDeclaration) writer.WriteProperty("re", true);

      writer.WriteObjectEnd();

      return;
    }

    var voidObj = asOrNull<RuntimeVoid>(obj);
    if (voidObj != null) {
      writer.Write("void");
      return;
    }

    // Legacy tag
    var tag = asOrNull<Tag>(obj);
    if (tag != null) {
      writer.WriteObjectStart();
      writer.WriteProperty("#", tag.text);
      writer.WriteObjectEnd();
      return;
    }

    var choice = asOrNull<Choice>(obj);
    if (choice != null) {
      WriteChoice(writer, choice);
      return;
    }

    throw Exception("Failed to convert runtime object to Json token: $obj");
  }

  static WriteRuntimeContainer(SimpleJsonWriter writer, Container? container,
      [bool withoutName = false]) {
    writer.WriteArrayStart();
    if (container == null) {
      return ArgumentError.notNull('container');
    }
    for (var c in container.content) {
      WriteRuntimeObject(writer, c);
    }

    var namedOnlyContent = container.namedOnlyContent;
    var countFlags = container.countFlags;
    var hasNameProperty = container.name != null && !withoutName;

    var hasTerminator =
        namedOnlyContent != null || countFlags > 0 || hasNameProperty;
    if (hasTerminator) {
      writer.WriteObjectStart();
    }

    if (namedOnlyContent != null) {
      for (var kv in namedOnlyContent.entries) {
        var name = kv.key;
        var namedContainer = asOrNull<Container>(kv.value);
        writer.WritePropertyStart(name);
        JsonSerialization.WriteRuntimeContainer(writer, namedContainer, true);
        writer.WritePropertyEnd();
      }
    }

    if (hasNameProperty) writer.WriteProperty('#n', container.name);

    if (hasTerminator) {
      writer.WriteObjectEnd();
    } else {
      writer.WriteNull();
    }

    writer.WriteArrayEnd();
  }

  static JArrayToContainer(List<dynamic> jArray) {
    var container = Container();
    container.content = JArrayToRuntimeObjList(jArray, true);

    var terminatingObj = jArray[jArray.length - 1] as Map<String, dynamic>?;
    if (terminatingObj != null) {
      var namedOnlyContent = <String, RuntimeObject>{};

      for (var key in terminatingObj.keys) {
        if (key == "#f") {
          container.countFlags = terminatingObj[key] as int;
        } else if (key == "#n") {
          container.name = terminatingObj[key].toString();
        } else {
          var namedContentItem = JTokenToRuntimeObject(terminatingObj[key]);
          // var namedSubContainer = namedContentItem as Container;
          var namedSubContainer = asOrNull<Container>(namedContentItem);
          if (namedSubContainer != null) namedSubContainer.name = key;
          namedOnlyContent[key] = namedContentItem!;
        }
      }

      container.namedOnlyContent = namedOnlyContent;
    }

    return container;
  }

  static JObjectToChoice(Map<String, dynamic> jObj) {
    var choice = Choice();
    choice.text = jObj["text"].toString();
    choice.index = jObj["index"] as int;
    choice.sourcePath = jObj["originalChoicePath"].toString();
    choice.originalThreadIndex = jObj["originalThreadIndex"] as int;
    choice.pathStringOnChoice = jObj["targetPath"].toString();
    return choice;
  }

  static WriteChoice(SimpleJsonWriter writer, Choice choice) {
    writer.WriteObjectStart();
    writer.WriteProperty("text", choice.text);
    writer.WriteIntProperty("index", choice.index);
    writer.WriteProperty("originalChoicePath", choice.sourcePath);
    writer.WriteIntProperty("originalThreadIndex", choice.originalThreadIndex);
    writer.WriteProperty("targetPath", choice.pathStringOnChoice);
    writer.WriteObjectEnd();
  }

  static WriteInkList(SimpleJsonWriter writer, ListValue listVal) {
    var rawList = listVal.value;
    if (rawList == null) {
      return ArgumentError.notNull("rawList");
    }

    writer.WriteObjectStart();
    writer.WritePropertyStart("list");
    writer.WriteObjectStart();

    for (var kv in rawList.entries) {
      var item = kv.key;
      var itemVal = kv.value;

      if (item.itemName == null) {
        return ArgumentError.notNull("item.itemName");
      }

      writer.WritePropertyNameStart();
      writer.WritePropertyNameInner(item.originName ?? "?");
      writer.WritePropertyNameInner(".");
      writer.WritePropertyNameInner(item.itemName!);
      writer.WritePropertyNameEnd();

      writer.Write(itemVal);

      writer.WritePropertyEnd();
    }

    writer.WriteObjectEnd();

    writer.WritePropertyEnd();

    if (rawList.isEmpty &&
        rawList.originNames != null &&
        rawList.originNames!.isNotEmpty) {
      writer.WritePropertyStart("origins");
      writer.WriteArrayStart();
      for (var name in rawList.originNames!) {
        writer.Write(name);
      }
      writer.WriteArrayEnd();
      writer.WritePropertyEnd();
    }

    writer.WriteObjectEnd();
  }

  static ListDefinitionsOrigin JTokenToListDefinitions(dynamic obj) {
    var defsObj = obj as Map<String, dynamic>;

    var allDefs = <ListDefinition>[];

    for (var kv in defsObj.entries) {
      var name = kv.key;
      var listDefJson = Map<String, dynamic>.from(kv.value);

      // Cast (string, object) to (string, int) for items
      var items = <String, int>{};
      for (var nameValue in listDefJson.entries) {
        items[nameValue.key] = nameValue.value as int;
      }

      var def = ListDefinition(name, items);
      allDefs.add(def);
    }

    return ListDefinitionsOrigin(allDefs);
  }

  // ----------------------
  // JSON ENCODING SCHEME
  // ----------------------
  //
  // Glue:           "<>", "G<", "G>"
  //
  // ControlCommand: "ev", "out", "/ev", "du" "pop", "->->", "~ret", "str", "/str", "nop",
  //                 "choiceCnt", "turns", "visit", "seq", "thread", "done", "end"
  //
  // NativeFunction: "+", "-", "/", "*", "%" "~", "==", ">", "<", ">=", "<=", "!=", "!"... etc
  //
  // Void:           "void"
  //
  // Value:          "^string value", "^^string value beginning with ^"
  //                 5, 5.2
  //                 {"^->": "path.target"}
  //                 {"^var": "varname", "ci": 0}
  //
  // Container:      [...]
  //                 [...,
  //                     {
  //                         "subContainerName": ...,
  //                         "#f": 5,                    // flags
  //                         "#n": "containerOwnName"    // only if not redundant
  //                     }
  //                 ]
  //
  // Divert:         {"->": "path.target", "c": true }
  //                 {"->": "path.target", "var": true}
  //                 {"f()": "path.func"}
  //                 {"->t->": "path.tunnel"}
  //                 {"x()": "externalFuncName", "exArgs": 5}
  //
  // Var Assign:     {"VAR=": "varName", "re": true}   // reassignment
  //                 {"temp=": "varName"}
  //
  // Var ref:        {"VAR?": "varName"}
  //                 {"CNT?": "stitch name"}
  //
  // ChoicePoint:    {"*": pathString,
  //                  "flg": 18 }
  //
  // Choice:         Nothing too clever, it's only used in the save state,
  //                 there's not likely to be many of them.
  //
  // Tag:            {"#": "the tag text"}
  static RuntimeObject? JTokenToRuntimeObject(dynamic token) {
    if (token is int || token is double || token is bool) {
      return Value.create(token);
    }
    if (token is String) {
      var str = token;
      // String Value
      var firstChar = str[0];
      if (firstChar == '^') {
        return StringValue(str.substring(1));
      } else if (firstChar == '\n' && str.length == 1) {
        return StringValue('\n');
      }

      // Glue
      if (str == '<>') {
        return Glue();
      }

      // Control commands (would looking up in a hash set be faster?)
      for (var i = 0; i < controlCommandNames.length; ++i) {
        var cmdName = controlCommandNames[i];
        if (str == cmdName) {
          return ControlCommand(CommandType.values[i]);
        }
      }

      // Native functions
      // "^" conflicts with the way to identify strings, so now
      // we know it's not a string, we can convert back to the proper
      // symbol for the operator.
      if (str == 'L^') str = '^';
      if (NativeFunctionCall.callExistsWithName(str)) {
        return NativeFunctionCall.callWithName(str);
      }
      // Pop
      if (str == '->->') {
        return ControlCommand.PopTunnel();
      } else if (str == '~ret') {
        return ControlCommand.PopFunction();
      }

      // Void
      if (str == 'void') {
        return RuntimeVoid();
      }
    }

    if (token is Map<String, dynamic>) {
      var obj = token;
      dynamic propValue;

      // Divert target value to path
      propValue = obj['^->'];
      if (propValue != null) {
        return DivertTargetValue(Path(propValue as String));
      }

      // VariablePointerValue
      propValue = obj['^var'];
      if (propValue != null) {
        var varPtr = VariablePointerValue(obj['^var'] as String);

        propValue = obj['ci'];

        if (propValue != null) {
          varPtr.contextIndex = propValue as int;
        }
        return varPtr;
      }

      // Divert
      var isDivert = false;
      var pushesToStack = false;
      var divPushType = PushPopType.Function;
      var external = false;

      propValue = obj['->'];
      if (propValue != null) {
        isDivert = true;
      } else {
        propValue = obj['f()'];

        if (propValue != null) {
          isDivert = true;
          pushesToStack = true;
          divPushType = PushPopType.Function;
        } else {
          propValue = obj['->t->'];
          if (propValue != null) {
            isDivert = true;
            pushesToStack = true;
            divPushType = PushPopType.Tunnel;
          } else {
            propValue = obj['x()'];
            if (propValue != null) {
              isDivert = true;
              external = true;
              pushesToStack = false;
              divPushType = PushPopType.Function;
            }
          }
        }
      }
      if (isDivert) {
        var divert = Divert();
        divert.pushesToStack = pushesToStack;
        divert.stackPushType = divPushType;
        divert.isExternal = external;

        var target = propValue.toString();

        propValue = obj['var'];

        if (propValue != null) {
          divert.variableDivertName = target;
        } else {
          divert.targetPathString = target;
        }

        propValue = obj['c'];
        divert.isConditional = obj['c'] ?? false;

        if (external) {
          propValue = obj['exArgs'];
          if (propValue != null) {
            divert.externalArgs = obj['exArgs'] as int;
          }
        }

        return divert;
      }

      // Choice
      propValue = obj['*'];
      if (propValue != null) {
        var choice = ChoicePoint();
        choice.pathStringOnChoice = propValue.toString();
        propValue = obj['flg'];

        if (propValue != null) {
          choice.flags = obj['flg'] as int;
        }
        return choice;
      }

      // Variable reference
      propValue = obj['VAR?'];
      if (propValue != null) {
        return VariableReference(propValue.toString());
      } else {
        propValue = obj['CNT?'];
        if (propValue != null) {
          var readCountVarRef = VariableReference();
          readCountVarRef.pathStringForCount = propValue.toString();
          return readCountVarRef;
        }
      }

      // Variable assignment
      var isVarAss = false;
      var isGlobalVar = false;

      propValue = obj['VAR='];
      if (propValue != null) {
        isVarAss = true;
        isGlobalVar = true;
      } else {
        propValue = obj['temp='];
        if (propValue != null) {
          isVarAss = true;
          isGlobalVar = false;
        }
      }
      if (isVarAss) {
        var varName = propValue.toString();
        propValue = obj['re'];
        bool isNewDecl = propValue == null;

        var varAss = VariableAssignment(varName, isNewDecl);
        varAss.isGlobal = isGlobalVar;
        return varAss;
      }

      // Legacy Tag with text
      propValue = obj['#'];
      if (propValue != null) {
        return Tag(obj['#'].toString());
      }

      // List value
      propValue = obj['list'];
      // List value
      if (propValue != null) {
        var listContent = Map<String, dynamic>.from(propValue);
        var rawList = InkList();
        propValue = obj['origins'];

        if (propValue != null) {
          var namesAsObjs = List<String>.from(propValue);
          rawList.setInitialOriginNames(namesAsObjs);
        }
        listContent.forEach((key, value) {
          var item = InkListItem.fullName(key);
          var val = value;
          rawList[item] = val;
        });
        return ListValue(rawList);
      }

      // Used when serialising save state only
      if (obj['originalChoicePath'] != null) {
        return JObjectToChoice(obj);
      }
    }

    // Array is always a Runtime.Container
    if (token is List<dynamic>) {
      return JArrayToContainer(token);
    }

    if (token == null) return null;

    throw Exception('Failed to convert token to runtime object: $token');
  }

  static Map<String, RuntimeObject> JObjectToDictionaryRuntimeObjs(
      Map<String, dynamic> jElementObj) {
    var dict = <String, RuntimeObject>{};
    jElementObj.forEach((key, value) {
      dict[key] = JsonSerialization.JTokenToRuntimeObject(value)!;
    });
    return dict;
  }
}
