import 'dart:math' as math;

import 'package:collection/collection.dart';

import 'ink_list.dart';
import 'list_definition.dart';
import 'logger.dart';
import 'path.dart';
import 'runtime_object.dart';
import 'runtime_void.dart';
import 'story_exception.dart';
import 'value.dart';

typedef BinaryOp<T> = Object Function(T left, T right);
typedef UnaryOp<T> = Object Function(T val);

final _logger = Logger('NativeFunctionCall');

class NativeFunctionCall extends RuntimeObject {
  static const String Add = '+';
  static const String Subtract = '-';
  static const String Divide = '/';
  static const String Multiply = '*';
  static const String Mod = '%';
  static const String Negate = '_'; // distinguish from '-' for subtraction

  static const String Equal = '==';
  static const String Greater = '>';
  static const String Less = '<';
  static const String GreaterThanOrEquals = '>=';
  static const String LessThanOrEquals = '<=';
  static const String NotEquals = '!=';
  static const String Not = '!';

  static const String And = '&&';
  static const String Or = '||';

  static const String Min = 'MIN';
  static const String Max = 'MAX';

  static const String Pow = 'POW';
  static const String Floor = 'FLOOR';
  static const String Ceiling = 'CEILING';
  static const String Int = 'INT';
  static const String Float = 'FLOAT';

  static const String Has = '?';
  static const String Hasnt = '!?';
  static const String Intersect = '^';

  static const String ListMin = 'LIST_MIN';
  static const String ListMax = 'LIST_MAX';
  static const String All = 'LIST_ALL';
  static const String Count = 'LIST_COUNT';
  static const String ValueOfList = 'LIST_VALUE';
  static const String Invert = 'LIST_INVERT';

  NativeFunctionCall? _prototype;
  bool _isPrototype = false;

  // Operations for each data type, for a single operation (e.g. "+")
  Map<ValueType, Function>? _operationFuncs;

  static Map<String, NativeFunctionCall>? _nativeFunctions;

  static NativeFunctionCall callWithName(String functionName) {
    return NativeFunctionCall(functionName);
  }

  static bool callExistsWithName(String functionName) {
    GenerateNativeFunctionsIfNecessary();
    return _nativeFunctions!.containsKey(functionName);
  }

  String? _name;

  String? get name => _name;

  set name(value) {
    _logger.finest('set name');

    _name = value;
    if (!_isPrototype) {
      _logger.finest('!_isPrototype');
      _prototype = _nativeFunctions![_name];
    }
  }

  int _numberOfParameters = 0;

  int get numberOfParameters {
    if (_prototype != null) {
      return _prototype!.numberOfParameters;
    } else {
      return _numberOfParameters;
    }
  }

  RuntimeObject? call(List<RuntimeObject> parameters) {
    if (_prototype != null) {
      return _prototype!.call(parameters);
    }

    if (numberOfParameters != parameters.length) {
      throw Exception('Unexpected number of parameters');
    }

    var hasList = false;
    for (var p in parameters) {
      if (p is RuntimeVoid) {
        throw StoryException(
            "Attempting to perform operation on a void value. Did you forget to 'return' a value from a function you called here?");
      }
      if (p is ListValue) {
        hasList = true;
      }
    }

    // Binary operations on lists are treated outside of the standard coerscion rules
    if (parameters.length == 2 && hasList) {
      _logger.finest(
          'Binary operations on lists are treated outside of the standard coerscion rules');
      return _callBinaryListOperation(parameters);
    }

    var coercedParams = _coerceValuesToSingleType(parameters);
    _logger.finest('coercedParams : $coercedParams');

    var coercedType = coercedParams.first.valueType;

    if (coercedType == ValueType.Int) {
      return _Call<int>(coercedParams);
    } else if (coercedType == ValueType.Float) {
      return _Call<double>(coercedParams);
    } else if (coercedType == ValueType.String) {
      return _Call<String>(coercedParams);
    } else if (coercedType == ValueType.DivertTarget) {
      return _Call<Path>(coercedParams);
    } else if (coercedType == ValueType.List) {
      return _Call<InkList>(coercedParams);
    }

    return null;
  }

  Value _Call<T>(List<Value> parametersOfSingleType) {
    var param1 = parametersOfSingleType[0];
    var valType = param1.valueType;

    var val1 = param1 as Value<T>;

    var paramCount = parametersOfSingleType.length;

    if (paramCount == 2 || paramCount == 1) {
      var opForTypeObj = _operationFuncs?[valType];

      if (opForTypeObj == null) {
        throw StoryException("Cannot perform operation '$name' on $valType");
      }

      // Binary
      if (paramCount == 2) {
        var param2 = parametersOfSingleType[1];

        var val2 = param2 as Value<T>;

        var opForType = opForTypeObj as BinaryOp<T>;

        // Return value unknown until it's evaluated
        var resultVal = opForType(val1.value, val2.value);

        return Value.create(resultVal)!;
      }

      // Unary
      else {
        var opForType = opForTypeObj as UnaryOp<T>;

        var resultVal = opForType(val1.value);

        return Value.create(resultVal)!;
      }
    } else {
      throw Exception(
          'Unexpected number of parameters to NativeFunctionCall: ' +
              parametersOfSingleType.length.toString());
    }
  }

  Value _callBinaryListOperation(List<RuntimeObject> parameters) {
    // List-Int addition/subtraction returns a List (e.g. "alpha" + 1 = "beta")
    if ((name == '+' || name == '-') &&
        parameters[0] is ListValue &&
        parameters[1] is IntValue) {
      return _callListIncrementOperation(parameters);
    }

    var v1 = parameters[0] as Value;
    var v2 = parameters[1] as Value;

    // And/or with any other type requires coerscion to bool (int)
    if ((name == '&&' || name == '||') &&
        (v1.valueType != ValueType.List || v2.valueType != ValueType.List)) {
      var op = _operationFuncs![ValueType.Int] as BinaryOp<int>;
      var result = op(v1.isTruthy ? 1 : 0, v2.isTruthy ? 1 : 0) as bool;
      return BoolValue(result);
    }

    // Normal (list • list) operation
    if (v1.valueType == ValueType.List && v2.valueType == ValueType.List) {
      return _Call<InkList>(List<Value>.of([v1, v2]));
    }

    throw StoryException("Can not call use '" +
        "$name' operation on " +
        v1.valueType.toString() +
        ' and ' +
        v2.valueType.toString());
  }

  // Value CallListIncrementOperation (List<Runtime.Object> listIntParams)
  Value _callListIncrementOperation(List<RuntimeObject> listIntParams) {
    var listVal = listIntParams[0] as ListValue;
    var intVal = listIntParams[1] as IntValue;

    var resultRawList = InkList();

    listVal.value.forEach((listItem, listItemValue) {
      // Find + or - operation
      var intOp = _operationFuncs![ValueType.Int] as BinaryOp<int>;

      // Return value unknown until it's evaluated
      var targetInt = intOp(listItemValue, intVal.value) as int;

      // Find this item's origin (linear search should be ok, should be short haha)
      ListDefinition? itemOrigin;
      for (var origin in listVal.value.origins ?? []) {
        if (origin.name == listItem.originName) {
          itemOrigin = origin;
          break;
        }
      }
      if (itemOrigin != null) {
        var incrementedItem = itemOrigin.getItemWithValue(targetInt);
        if (incrementedItem.isValue) {
          resultRawList[incrementedItem.asValue!.value] = targetInt;
        }
      }
    });
    return ListValue(resultRawList);
  }

  // List<Value> CoerceValuesToSingleType(List<Runtime.Object> parametersIn)
  List<Value> _coerceValuesToSingleType(List<RuntimeObject> parametersIn) {
    _logger
        .finest('_coerceValuesToSingleType ( parametersIn : $parametersIn )  ');
    var valType = ValueType.Int;

    ListValue? specialCaseList;

    // Find out what the output type is
    // "higher level" types infect both so that binary operations
    // use the same type on both sides. e.g. binary operation of
    // int and float causes the int to be casted to a float.
    for (var obj in parametersIn) {
      var val = obj as Value;
      if (val.valueType.effectiveIndex > valType.effectiveIndex) {
        valType = val.valueType;
      }

      if (val.valueType == ValueType.List) {
        specialCaseList = val as ListValue;
      }
    }

    // Coerce to this chosen type
    var parametersOut = <Value>[];

    // Special case: Coercing to Ints to Lists
    // We have to do it early when we have both parameters
    // to hand - so that we can make use of the List's origin
    if (valType == ValueType.List) {
      for (var val in parametersIn) {
        final value = val as Value;
        if (val.valueType == ValueType.List) {
          parametersOut.add(val);
        } else if (val.valueType == ValueType.Int) {
          var intVal = val.valueObject as int;
          var list = specialCaseList!.value.originOfMaxItem;

          var item = list?.getItemWithValue(intVal);
          if (item != null && item.isValue) {
            var castedValue = ListValue.single(item.asValue!.value, intVal);
            parametersOut.add(castedValue);
          } else {
            throw StoryException('Could not find List item with the value ' +
                intVal.toString() +
                ' in ${list?.name}');
          }
        } else {
          throw StoryException('Cannot mix Lists and ' +
              val.valueType.toString() +
              ' values in this operation');
        }
      }
    }

    // Normal Coercing (with standard casting)
    else {
      for (var val in parametersIn) {
        val as Value;
        var castedValue = val.cast(valType)!;
        parametersOut.add(castedValue);
      }
    }

    return parametersOut;
  }

  // Require default constructor for serialisation
  NativeFunctionCall([String? name]) {
    GenerateNativeFunctionsIfNecessary();
    if (name != null) {
      this.name = name;
    }
  }

  // Only called internally to generate prototypes
  NativeFunctionCall._(String name, int numberOfParameters) {
    _isPrototype = true;
    this.name = name;
    _numberOfParameters = numberOfParameters;
  }

  static Function deepEq = const DeepCollectionEquality().equals;

// For defining operations that do nothing to the specific type
  // (but are still supported), such as floor/ceil on int and float
  // cast on float.
  static Object Identity<T>(T t) {
    return t as Object;
  }

  static void GenerateNativeFunctionsIfNecessary() {
    if (_nativeFunctions == null) {
      _nativeFunctions = <String, NativeFunctionCall>{};

      // Why no bool operations?
      // Before evaluation, all bools are coerced to ints in
      // CoerceValuesToSingleType (see default value for valType at top).
      // So, no operations are ever directly done in bools themselves.
      // This also means that 1 == true works, since true is always converted
      // to 1 first.
      // However, many operations return a "native" bool (equals, etc).

      // Int operations
      AddIntBinaryOp(Add, (x, y) => x + y);
      AddIntBinaryOp(Subtract, (x, y) => x - y);
      AddIntBinaryOp(Multiply, (x, y) => x * y);
      AddIntBinaryOp(Divide, (x, y) => x ~/ y);
      AddIntBinaryOp(Mod, (x, y) => x % y);
      AddIntUnaryOp(Negate, (x) => -x);

      AddIntBinaryOp(Equal, (x, y) => x == y);
      AddIntBinaryOp(Greater, (x, y) => x > y);
      AddIntBinaryOp(Less, (x, y) => x < y);
      AddIntBinaryOp(GreaterThanOrEquals, (x, y) => x >= y);
      AddIntBinaryOp(LessThanOrEquals, (x, y) => x <= y);
      AddIntBinaryOp(NotEquals, (x, y) => x != y);
      AddIntUnaryOp(Not, (x) => x == 0);

      AddIntBinaryOp(And, (x, y) => x != 0 && y != 0);
      AddIntBinaryOp(Or, (x, y) => x != 0 || y != 0);

      AddIntBinaryOp(Max, (x, y) => math.max(x, y));
      AddIntBinaryOp(Min, (x, y) => math.min(x, y));

      // Have to cast to float since you could do POW(2, -1)
      AddIntBinaryOp(Pow, (x, y) => math.pow(x, y).toDouble());
      AddIntUnaryOp(Floor, (x) => x);
      AddIntUnaryOp(Ceiling, (x) => x);
      AddIntUnaryOp(Int, (x) => x);
      AddIntUnaryOp(Float, (x) => x.toDouble());

      // Float operations
      AddFloatBinaryOp(Add, (x, y) => x + y);
      AddFloatBinaryOp(Subtract, (x, y) => x - y);
      AddFloatBinaryOp(Multiply, (x, y) => x * y);
      AddFloatBinaryOp(Divide, (x, y) => x / y);
      AddFloatBinaryOp(Mod,
          (x, y) => x % y); // TODO: Is this the operation we want for floats?
      AddFloatUnaryOp(Negate, (x) => -x);

      AddFloatBinaryOp(Equal, (x, y) => x == y);
      AddFloatBinaryOp(Greater, (x, y) => x > y);
      AddFloatBinaryOp(Less, (x, y) => x < y);
      AddFloatBinaryOp(GreaterThanOrEquals, (x, y) => x >= y);
      AddFloatBinaryOp(LessThanOrEquals, (x, y) => x <= y);
      AddFloatBinaryOp(NotEquals, (x, y) => x != y);
      AddFloatUnaryOp(Not, (x) => (x == 0.0) ? 1 : 0);

      AddFloatBinaryOp(And, (x, y) => x != 0.0 && y != 0.0);
      AddFloatBinaryOp(Or, (x, y) => x != 0.0 || y != 0.0);

      AddFloatBinaryOp(Max, (x, y) => math.max(x, y));
      AddFloatBinaryOp(Min, (x, y) => math.min(x, y));

      AddFloatBinaryOp(Pow, (x, y) => math.pow(x, y).toDouble());
      AddFloatUnaryOp(Floor, (x) => x.floor().toDouble());
      AddFloatUnaryOp(Ceiling, (x) => x.ceil().toDouble());
      AddFloatUnaryOp(Int, (x) => x.toInt());
      AddFloatUnaryOp(Float, (x) => x);

      // String operations
      AddStringBinaryOp(Add, (x, y) => x + y); // concat
      AddStringBinaryOp(Equal, (x, y) => x == y);
      AddStringBinaryOp(NotEquals, (x, y) => x != y);
      AddStringBinaryOp(Has, (x, y) => x.contains(y));
      AddStringBinaryOp(Hasnt, (x, y) => x.contains(y));

      // List operations
      AddListBinaryOp(Add, (x, y) => x.union(y));
      AddListBinaryOp(Subtract, (x, y) => x.without(y));
      AddListBinaryOp(Has, (x, y) => x.contains(y));
      AddListBinaryOp(Hasnt, (x, y) => !x.contains(y));
      AddListBinaryOp(Intersect, (x, y) => x.intersect(y));

      AddListBinaryOp(Equal, (x, y) => deepEq(x, y));
      AddListBinaryOp(Greater, (x, y) => x.greaterThan(y));
      AddListBinaryOp(Less, (x, y) => x.lessThan(y));
      AddListBinaryOp(GreaterThanOrEquals, (x, y) => x.greaterThanOrEquals(y));
      AddListBinaryOp(LessThanOrEquals, (x, y) => x.lessThanOrEquals(y));
      AddListBinaryOp(NotEquals, (x, y) => !deepEq(x, y));

      AddListBinaryOp(And, (x, y) => x.isNotEmpty && y.isNotEmpty);
      AddListBinaryOp(Or, (x, y) => x.isNotEmpty || y.isNotEmpty);

      AddListUnaryOp(Not, (x) => x.isEmpty ? 1 : 0);

      // Placeholders to ensure that these special case functions can exist,
      // since these function is never actually run, and is special cased in Call
      AddListUnaryOp(Invert, (x) => x.inverse);
      AddListUnaryOp(All, (x) => x.all);
      AddListUnaryOp(ListMin, (x) => x.MinAsList());
      AddListUnaryOp(ListMax, (x) => x.maxAsList());
      AddListUnaryOp(Count, (x) => x.length);
      AddListUnaryOp(ValueOfList, (x) => x.maxItem.value);

      // Special case: The only operations you can do on divert target values
      BinaryOp<Path> divertTargetsEqual =
          (Path d1, Path d2) {
        return d1 == d2;
      };
      BinaryOp<Path> divertTargetsNotEqual =
          (Path d1, Path d2) {
        return d1 != d2;
      };
      AddOpToNativeFunc(Equal, 2, ValueType.DivertTarget, divertTargetsEqual);
      AddOpToNativeFunc(
          NotEquals, 2, ValueType.DivertTarget, divertTargetsNotEqual);
    }
  }

  void AddOpFuncForType(ValueType valType, Function op) {
    _operationFuncs ??= <ValueType, Function>{};

    _operationFuncs![valType] = op;
  }

  static void AddOpToNativeFunc(
      String name, int args, ValueType valType, Function op) {
    var nativeFunc = _nativeFunctions![name];
    if (nativeFunc == null) {
      nativeFunc = NativeFunctionCall._(name, args);
      _nativeFunctions![name] = nativeFunc;
      _logger.finest('nativeFunc : $nativeFunc');
    }

    nativeFunc.AddOpFuncForType(valType, op);
  }

  static void AddIntBinaryOp(String name, BinaryOp<int> op) {
    AddOpToNativeFunc(name, 2, ValueType.Int, op);
  }

  static void AddIntUnaryOp(String name, UnaryOp<int> op) {
    AddOpToNativeFunc(name, 1, ValueType.Int, op);
  }

  static void AddFloatBinaryOp(String name, BinaryOp<double> op) {
    AddOpToNativeFunc(name, 2, ValueType.Float, op);
  }

  static void AddStringBinaryOp(String name, BinaryOp<String> op) {
    AddOpToNativeFunc(name, 2, ValueType.String, op);
  }

  static void AddListBinaryOp(String name, BinaryOp<InkList> op) {
    AddOpToNativeFunc(name, 2, ValueType.List, op);
  }

  static void AddListUnaryOp(String name, UnaryOp<InkList> op) {
    AddOpToNativeFunc(name, 1, ValueType.List, op);
  }

  static void AddFloatUnaryOp(String name, UnaryOp<double> op) {
    AddOpToNativeFunc(name, 1, ValueType.Float, op);
  }

  @override
  String toString() {
    return "Native '$name' ${this.numberOfParameters}";
  }
}
