import 'package:collection/collection.dart';
import 'package:logging/logging.dart';
import 'package:quiver/strings.dart';

final _logger = Logger('Path');
const String PARENT_ID = '^';
const String PATH_NULL = '__________';

class Path {
  String? _componentsString;

  static Path NULL = Path(PATH_NULL);

  late List<Component> _components;

  bool _isRelative = false;

  bool get isRelative => _isRelative;

  Path([dynamic arg1, dynamic arg2]) {
    _components = [];
    _componentsString = null;
    _isRelative = false;
    if (arg1 is String) {
      _setComponentsString(arg1);
    } else if (arg1 is Component && arg2 is Path) {
      var head = arg1;
      var tail = arg2;
      _components.add(head);
      _components.addAll(tail._components);
    } else if (arg1 is Iterable<Component>) {
      var head = arg1;
      bool relative = arg2 == true;
      _components.addAll(head);
      _isRelative = relative;
    }
  }

  static Path get self {
    var path = Path();
    path._isRelative = true;
    return path;
  }

  Component? get lastComponent => _components.lastOrNull;

  bool get containsNamedComponent => _components.any((comp) => comp.isIndex);

  Component? get head => _components.firstOrNull;

  Path get tail {
    if (_components.length >= 2) {
      Iterable<Component> tailComps = _components.sublist(1);
      return Path(tailComps);
    } else {
      return Path.self;
    }
  }

  String get componentsString {
    if (_componentsString == null) {
      _componentsString = _components.join('.');
      if (isRelative) _componentsString = '.$_componentsString';
    }

    return _componentsString!;
  }

  void _setComponentsString(String? value) {
    _components.clear();
    _componentsString = value;

    // Empty path, empty components
    // (path is to root, like "/" in file system)
    if (isBlank(_componentsString)) {
      return;
    }

    // When components start with ".", it indicates a relative path, e.g.
    //   .^.^.hello.5
    // is equivalent to file system style path:
    //  ../../hello/5
    if (_componentsString![0] == '.') {
      _isRelative = true;
      _componentsString = _componentsString!.substring(1);
    } else {
      _isRelative = false;
    }

    var componentStrings = _componentsString!.split('.');
    for (var str in componentStrings) {
      int? index = int.tryParse(str);
      if (index != null) {
        _components.add(Component.index(index));
      } else {
        _components.add(Component.name(str));
      }
    }
  }

  @override
  String toString() {
    return componentsString;
  }

  // public Path PathByAppendingComponent (Component c)
  Path pathByAppendingComponent(Component c) {
    final p = Path();
    p._components.addAll(_components);
    p._components.add(c);
    return p;
  }

  //public Path PathByAppendingPath(Path pathToAppend)
  Path pathByAppendingPath(Path pathToAppend) {
    final p = Path();
    var upwardMoves = 0;
    for (var i = 0; i < pathToAppend._components.length; ++i) {
      if (pathToAppend._components[i].isParent) {
        upwardMoves++;
      } else {
        break;
      }
    }
    for (var i = 0; i < _components.length - upwardMoves; ++i) {
      p._components.add(_components[i]);
    }
    for (var i = upwardMoves; i < pathToAppend._components.length; ++i) {
      p._components.add(pathToAppend._components[i]);
    }
    return p;
  }

  Component getComponent(int index) => _components[index];

  int get length {
    return _components.length;
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is Path &&
          runtimeType == other.runtimeType &&
          _componentsString == other._componentsString &&
          eq(_components, other._components) &&
          _isRelative == other._isRelative;

  @override
  int get hashCode => toString().hashCode;

  Function eq = const DeepCollectionEquality().equals;
}

class Component {
  int _index = -1;
  String? _name;

  Component._(dynamic value) {
    if (value is int) {
      assert(value >= 0);
      _index = value;
      _name = null;
    } else if (value is String) {
      assert(value.isNotEmpty);
      _index = -1;
      _name = value;
    }
  }

  factory Component.name(String name) => Component._(name);

  factory Component.index(int index) => Component._(index);

  bool get isParent => PARENT_ID == _name;

  bool get isIndex => _index >= 0;

  int get index => _index;

  static Component toParent() => Component.name(PARENT_ID);

  String get name => _name!;

  @override
  String toString() {
    if (isIndex) {
      return index.toString();
    } else {
      return name;
    }
  }

  @override
  int get hashCode {
    if (isIndex) {
      return _index;
    } else {
      return _name.hashCode;
    }
  }

  @override
  bool operator ==(Object other) {
    if (other is! Component) {
      return false;
    }
    var otherComp = other;

    if (otherComp.isIndex == isIndex) {
      if (isIndex) {
        return index == otherComp.index;
      } else {
        return name == otherComp.name;
      }
    }

    return false;
  }
}
