import 'package:darq/darq.dart';

import 'call_stack.dart';
import 'control_command.dart';
import 'runtime_object.dart';

/// <summary>
/// Simple ink profiler that logs every instruction in the story and counts frequency and timing.
/// To use:
///
///   var profiler = story.StartProfiling(),
///
///   (play your story for a bit)
///
///   var reportStr = profiler.Report();
///
///   story.EndProfiling();
///
/// </summary>
class Profiler {
  Stopwatch _continueWatch = Stopwatch();
  Stopwatch _stepWatch = Stopwatch();
  Stopwatch _snapWatch = Stopwatch();

  double _continueTotal = 0.0;
  double _snapTotal = 0.0;
  double _stepTotal = 0.0;

  List<String>? _currStepStack;
  StepDetails? _currStepDetails;
  ProfileNode _rootNode;
  int _numContinues = 0;
  final List<StepDetails> _stepDetails = <StepDetails>[];

  static double _millisecsPerTick = 1000.0 / Stopwatch().frequency;

  Profiler() : _rootNode = ProfileNode();

  /// <summary>
  /// The root node in the hierarchical tree of recorded ink timings.
  /// </summary>
  ProfileNode get rootNode => _rootNode;

  /// <summary>
  /// Generate a printable report based on the data recording during profiling.
  /// </summary>
  String report() {
    var sb = StringBuffer();
    sb.writeln('$_numContinues CONTINUES / LINES:');
    sb.writeln('TOTAL TIME: ${formatMillisecs(_continueTotal)}');
    sb.writeln(
      'SNAPSHOTTING: ${formatMillisecs(_snapTotal)}',
    );
    sb.writeln(
      'OTHER: ${formatMillisecs(_continueTotal - (_stepTotal + _snapTotal))}',
    );
    sb.write(_rootNode.toString());
    return sb.toString();
  }

  void preContinue() {
    _continueWatch.reset();
    _continueWatch.start();
  }

  void postContinue() {
    _continueWatch.stop();
    _continueTotal += millisecs(_continueWatch);
    _numContinues++;
  }

  void PreStep() {
    _currStepStack = null;
    _stepWatch.reset();
    _stepWatch.start();
  }

  void Step(CallStack callstack) {
    _stepWatch.stop();

    var stack = <String>[];
    for (var i = 0; i < callstack.elements.length; i++) {
      var stackElementName = '';
      if (!callstack.elements[i].currentPointer.isNull) {
        var objPath = callstack.elements[i].currentPointer.path!;

        for (var c = 0; c < objPath.length; c++) {
          var comp = objPath.getComponent(c);
          if (!comp.isIndex) {
            stackElementName = comp.name;
            break;
          }
        }
      }
      stack[i] = stackElementName;
    }

    _currStepStack = stack;

    var currObj = callstack.currentElement.currentPointer.resolve()!;

    String stepType;
    var controlCommandStep = currObj is ControlCommand ? currObj : null;
    if (controlCommandStep != null) {
      stepType = controlCommandStep.commandType.toString() + ' CC';
    } else {
      // TODO stepType = currObj.GetType().Name;
      stepType = currObj.runtimeType.toString();
    }

    _currStepDetails = StepDetails(type: stepType, obj: currObj);

    _stepWatch.start();
  }

  void PostStep() {
    _stepWatch.stop();

    var duration = millisecs(_stepWatch);
    _stepTotal += duration;

    _rootNode.addSample(_currStepStack!, duration);

    _currStepDetails!.time = duration;
    _stepDetails.add(_currStepDetails!);
  }

  /// <summary>
  /// Generate a printable report specifying the average and maximum times spent
  /// stepping over different internal ink instruction types.
  /// This report type is primarily used to profile the ink engine itself rather
  /// than your own specific ink.
  /// </summary>
  String stepLengthReport() {
    var sb = StringBuffer();

    sb.writeln('TOTAL: ${_rootNode.totalMillisecs}ms');

    var averageStepTimes = _stepDetails
        .groupBy((StepDetails s) => s.type)
        .select((e, i) => MapEntry(e.key, e.average((d) => d.time)))
        .orderByDescending((stepTypeToAverage) => stepTypeToAverage.value)
        .select((stepTypeToAverage, index) {
      var typeName = stepTypeToAverage.key;
      var time = stepTypeToAverage.value;
      return typeName + ': ${time}ms';
    }).toList();

    sb.writeln('AVERAGE STEP TIMES: ' + averageStepTimes.join(', '));

    var accumStepTimes = _stepDetails
        .groupBy((StepDetails s) => s.type)
        .select((typeToDetails, index) => MapEntry<String, double>(
            typeToDetails.key + ' (x${typeToDetails.count})',
            typeToDetails.sum((d) => d.time)))
        .orderByDescending((stepTypeToAccum) => stepTypeToAccum.value)
        .select((stepTypeToAccum, _) {
      var typeName = stepTypeToAccum.key;
      var time = stepTypeToAccum.value;
      return '$typeName: $time';
    }).toList();

    sb.writeln('ACCUMULATED STEP TIMES: ' + accumStepTimes.join(', '));

    return sb.toString();
  }

  /// <summary>
  /// Create a large log of all the internal instructions that were evaluated while profiling was active.
  /// Log is in a tab-separated format, for easy loading into a spreadsheet application.
  /// </summary>
  String megalog() {
    var sb = StringBuffer();

    sb.writeln('Step type\tDescription\tPath\tTime');

    _stepDetails.forEach((step) {
      sb.write(step.type);
      sb.write('\t');
      sb.write(step.obj.toString());
      sb.write('\t');
      sb.write(step.obj.path);
      sb.write('\t');
      sb.writeln(step.time.toStringAsFixed(8));
    });

    return sb.toString();
  }

  void PreSnapshot() {
    _snapWatch.reset();
    _snapWatch.start();
  }

  void postSnapshot() {
    _snapWatch.stop();
    _snapTotal += millisecs(_snapWatch);
  }

  double millisecs(Stopwatch watch) {
    var ticks = watch.elapsedTicks;
    return ticks * _millisecsPerTick;
  }

  static String formatMillisecs(double num) {
    if (num > 5000) {
      //return String.Format("{0:N1} secs", num / 1000.0);
      return '${(num / 1000.0).toStringAsFixed(1)} secs';
    }
    if (num > 1000) {
      //return string.Format("{0:N2} secs", num / 1000.0);
      return '${(num / 1000.0).toStringAsFixed(2)} secs';
    } else if (num > 100) {
      //return string.Format("{0:N0} ms", num);
      return '${num.toStringAsFixed(0)} ms';
    } else if (num > 1) {
      //return string.Format("{0:N1} ms", num);
      return '${num.toStringAsFixed(1)} ms';
    } else if (num > 0.01) {
      //return string.Format("{0:N3} ms", num);
      return '${num.toStringAsFixed(3)} ms';
    } else {
      //return string.Format("{0:N} ms", num);
      return '${num.toStringAsFixed(0)} ms';
    }
  }
}

class StepDetails {
  String type;
  RuntimeObject obj;
  double time;

  StepDetails({required this.type, required this.obj, this.time = 0.0});
}

/// <summary>
/// Node used in the hierarchical tree of timings used by the Profiler.
/// Each node corresponds to a single line viewable in a UI-based representation.
/// </summary>
class ProfileNode {
  /// The key for the node corresponds to the printable name of the callstack element.
  final String? key;

  /// <summary>
  /// Horribly hacky field only used by ink unity integration,
  /// but saves constructing an entire data structure that mirrors
  /// the one in here purely to store the state of whether each
  /// node in the UI has been opened or not  /// </summary>
  // bool openInUI;

  Map<String, ProfileNode>? _nodes;
  double _selfMillisecs = 0.0;
  double _totalMillisecs = 0.0;
  int _selfSampleCount = 0;
  int _totalSampleCount = 0;

  ProfileNode([this.key]);

  /// <summary>
  /// Whether this node contains any sub-nodes - i.e. does it call anything else
  /// that has been recorded?
  /// </summary>
  /// <value><c>true</c> if has children; otherwise, <c>false</c>.</value>
  bool get hasChildren => _nodes != null && _nodes!.isNotEmpty;

  /// <summary>
  /// Total number of milliseconds this node has been active for.
  /// </summary>
  int get totalMillisecs => _totalMillisecs.toInt();

  void addSample(List<String> stack, double duration, [int stackIdx = -1]) {
    _totalSampleCount++;
    _totalMillisecs += duration;

    if (stackIdx == stack.length - 1) {
      _selfSampleCount++;
      _selfMillisecs += duration;
    }

    if (stackIdx + 1 < stack.length) {
      addSampleToNode(stack, stackIdx + 1, duration);
    }
  }

  void addSampleToNode(List<String> stack, int stackIdx, double duration) {
    var nodeKey = stack[stackIdx];
    _nodes ??= <String, ProfileNode>{};

    var node = _nodes![nodeKey];
    if (node != null) {
      node = ProfileNode(nodeKey);
      _nodes![nodeKey] = node;
    }

    node!.addSample(stack, duration, stackIdx);
  }

  /// <summary>
  /// Returns a sorted enumerable of the nodes in descending order of
  /// how long they took to run.
  /// </summary>
  Iterable<MapEntry<String, ProfileNode>>? get descendingOrderedNodes {
    if (_nodes == null) return null;
    return _nodes!.entries
        .orderByDescending((keyNode) => keyNode.value._totalMillisecs);
  }

  void printHierarchy(StringBuffer sb, int indent) {
    Pad(sb, indent);

    sb.write(key);
    sb.write(': ');
    sb.write(ownReport);

    if (_nodes == null) return;
    descendingOrderedNodes!.forEach((keyNode) {
      keyNode.value.printHierarchy(sb, indent + 1);
    });
  }

  /// <summary>
  /// Generates a string giving timing information for this single node, including
  /// total milliseconds spent on the piece of ink, the time spent within itself
  /// (v.s. spent in children), as well as the number of samples (instruction steps)
  /// recorded for both too.
  /// </summary>
  /// <value>The own report.</value>
  String get ownReport {
    var sb = StringBuffer();
    sb.write('total ');
    sb.write(Profiler.formatMillisecs(_totalMillisecs));
    sb.write(', self ');
    sb.write(Profiler.formatMillisecs(_selfMillisecs));
    sb.write(' (');
    sb.write(_selfSampleCount);
    sb.write(' self samples, ');
    sb.write(_totalSampleCount);
    sb.write(' total)');
    return sb.toString();
  }

  void Pad(StringBuffer sb, int spaces) {
    for (var i = 0; i < spaces; i++) {
      sb.write('   ');
    }
  }

  /// <summary>
  /// String is a report of the sub-tree from this node, but without any of the header information
  /// that's prepended by the Profiler in its Report() method.
  /// </summary>
  @override
  String toString() {
    var sb = StringBuffer();
    printHierarchy(sb, 0);
    return sb.toString();
  }
}
