import 'dart:math';

import 'container.dart';
import 'debug_metadata.dart';
import 'extension.dart';
import 'named_content.dart';
import 'path.dart';
import 'search_result.dart';

abstract class RuntimeObject {
  Container? get rootContentContainer {
    var ancestor = this;
    while (ancestor.parent != null) {
      ancestor = ancestor.parent!;
    }
    return asOrNull<Container>(ancestor);
  }

  /// RuntimeObjects can be included in the main Story as a hierarchy.
  /// Usually parents are Container objects. (TODO: Always?)
  RuntimeObject? parent;
  Path? _path;

  /// TODO: Come up with some clever solution for not having
  /// to have debug metadata on the RTObject itself, perhaps
  /// for serialisation purposes at least.
  DebugMetadata? _debugMetadata;

  RuntimeObject({this.parent});

  DebugMetadata get ownDebugMetadata => _debugMetadata!;

  DebugMetadata? get debugMetadata => _debugMetadata ?? parent?.debugMetadata;

  set debugMetadata(v) => _debugMetadata = v;

  int? debugLineNumberOfPath(Path? path) {
    // FIXME Added path.isRelative() because orginal code not working
    if (path == null) {
      return null;
    }
    // Try to get a line number from debug metadata
    var root = rootContentContainer;
    if (root != null) {
      var targetContent = rootContentContainer!.contentAtPath(path).obj;
      if (targetContent != null) {
        var dm = targetContent.debugMetadata;
        if (dm != null) {
          return dm.startLineNumber;
        }
      }
    }
    return null;
  }

  @override
  Path get path {
    if (_path == null) {
      if (parent == null) {
        return _path = Path();
      } else {
        // Maintain a Stack so that the order of the components
        // is reversed when they're added to the Path.
        // We're iterating up the hierarchy from the leaves/children to the root.
        final comps = <Component>[];

        var child = this;
        Container? container = asOrNull<Container>(child.parent);

        while (container != null) {
          var namedChild = asOrNull<NamedContent>(child);
          if (namedChild != null && namedChild.hasValidName) {
            comps.insert(0, Component.name(namedChild.name));
          } else {
            comps.insert(0, Component.index(container.content.indexOf(child)));
          }
          child = container;
          container = asOrNull<Container>(container.parent);
        }
        _path = Path(comps);
      }
    }
    return _path!;
  }

  SearchResult resolvePath(Path? path) {
    ArgumentError.checkNotNull(path);
    if (path!.isRelative == true) {
      var nearestContainer = asOrNull<Container>(this);
      if (nearestContainer == null) {
        assert(parent != null,
            "Can't resolve relative path because we don't have a parent");
        nearestContainer = asOrNull<Container>(parent);
        assert(nearestContainer != null, 'Expected parent to be a container');
        assert(path.getComponent(0).isParent);
        path = path.tail;
      }

      ArgumentError.checkNotNull(nearestContainer, 'nearestContainer');

      return nearestContainer!.contentAtPath(path);
    } else {
      return rootContentContainer!.contentAtPath(path);
    }
  }

  Path convertPathToRelative(Path globalPath) {
    // 1. Find last shared ancestor
    // 2. Drill up using ".." style (actually represented as "^")
    // 3. Re-build downward chain from common ancestor

    var ownPath = path;

    var minPathLength = min(globalPath.length, ownPath.length);
    var lastSharedPathCompIndex = -1;

    for (var i = 0; i < minPathLength; ++i) {
      var ownComp = ownPath.getComponent(i);
      var otherComp = globalPath.getComponent(i);

      if (ownComp == otherComp) {
        lastSharedPathCompIndex = i;
      } else {
        break;
      }
    }

    // No shared path components, so just use global path
    if (lastSharedPathCompIndex == -1) return globalPath;

    var numUpwardsMoves = (ownPath.length - 1) - lastSharedPathCompIndex;

    var newPathComps = <Component>[];

    for (var up = 0; up < numUpwardsMoves; ++up) {
      newPathComps.add(Component.toParent());
    }

    for (var down = lastSharedPathCompIndex + 1;
        down < globalPath.length;
        ++down) {
      newPathComps.add(globalPath.getComponent(down));
    }

    var relativePath = Path(newPathComps, true);
    return relativePath;
  }

  // Find most compact representation for a path, whether relative or global
  String compactPathString(Path otherPath) {
    String globalPathStr;
    String relativePathStr;
    if (otherPath.isRelative) {
      relativePathStr = otherPath.componentsString;
      globalPathStr = path.pathByAppendingPath(otherPath).componentsString;
    } else {
      var relativePath = convertPathToRelative(otherPath);
      relativePathStr = relativePath.componentsString;
      globalPathStr = otherPath.componentsString;
    }

    if (relativePathStr.length < globalPathStr.length) {
      return relativePathStr;
    } else {
      return globalPathStr;
    }
  }

  void setChild<T extends RuntimeObject>(T? obj, T? value) {
    if (obj != null) {
      obj.parent = null;
    }
    obj = value;
    if (obj != null) {
      obj.parent = this;
    }
  }

  String CompactPathString(Path otherPath) {
    String globalPathStr;
    String relativePathStr;
    if (otherPath.isRelative) {
      relativePathStr = otherPath.componentsString;
      globalPathStr = path.pathByAppendingPath(otherPath).componentsString;
    } else {
      var relativePath = convertPathToRelative(otherPath);
      relativePathStr = relativePath.componentsString;
      globalPathStr = otherPath.componentsString;
    }

    if (relativePathStr.length < globalPathStr.length) {
      return relativePathStr;
    } else {
      return globalPathStr;
    }
  }
}
