import 'dart:collection';
import 'dart:convert';

import 'action.dart';

class SimpleJson {
  static TextToDictionary(String text) {
    return SimpleJsonReader(text).ToDictionary();
  }

  static TextToArray(String text) {
    return SimpleJsonReader(text).ToArray();
  }
}

class SimpleJsonReader {
  final dynamic _rootObject;

  SimpleJsonReader(String text) : _rootObject = jsonDecode(text);

  Map<String, dynamic> ToDictionary() => _rootObject as Map<String, dynamic>;

  List<dynamic> ToArray() => _rootObject as List<dynamic>;
}

class SimpleJsonWriter {
  // In addition to `_stateStack` present in the original code,
  // this implementation of SimpleJson use two other stacks and two
  // temporary variables holding the current context.

  // Used to keep track of the current property name being built
  // with `WritePropertyNameStart`, `WritePropertyNameInner` and
  // `WritePropertyNameEnd`.
  dynamic _currentPropertyName;

  // Used to keep track of the current string value being built
  // with `WriteStringStart`, `WriteStringInner` and
  // `WriteStringEnd`.
  dynamic _currentString = null;

  // Keep track of the current collection being built (either an array
  // or an object). For instance, at the '?' step during the hiarchy
  // creation, this hierarchy:
  // [3, {a: [b, ?]}] will have this corresponding stack:
  // (bottom) [Array, Object, Array] (top)
  List _collectionStack = [];

  // Keep track of the current property being assigned. For instance, at
  // the '?' step during the hiarchy creation, this hierarchy:
  // [3, {a: [b, {c: ?}]}] will have this corresponding stack:
  // (bottom) [a, c] (top)
  List<String> _propertyNameStack = [];
  dynamic _jsonObject;

  SimpleJsonWriter();

  final ListQueue<StateElement> _stateStack = ListQueue();

  void WriteObject(Action<SimpleJsonWriter> inner) {
    WriteObjectStart();
    inner.call(this);
    WriteObjectEnd();
  }

  void WriteObjectStart() {
    StartNewObject(container: true);
    var newObject = <String, dynamic>{};

    if (state == State.Property) {
      // This object is created as the value of a property,
      // inside an other object.
      assert(currentCollection != null);
      assert(currentPropertyName != null);

      var propertyName = _propertyNameStack.removeLast();
      currentCollection![propertyName] = newObject;
      _collectionStack.add(newObject);
    } else if (state == State.Array) {
      // This object is created as the child of an array.
      assert(currentCollection != null);

      currentCollection!.add(newObject);
      _collectionStack.add(newObject);
    } else {
      // This object is the root object.
      assert(state == State.None);
      _jsonObject = newObject;
      _collectionStack.add(newObject);
    }
    _stateStack.add(StateElement(type: State.Object));
  }

  void WriteObjectEnd() {
    assert(state == State.Object);
    _collectionStack.removeLast();
    _stateStack.removeLast();
  }

  void WriteProperty(dynamic name, dynamic innerOrContent) {
    WritePropertyStart(name);
    if (innerOrContent is Action) {
      var inner = innerOrContent;
      inner.call(this);
    } else {
      var content = innerOrContent;
      if (content is int) {
        WriteInt(content);
      } else if (content is double) {
        WriteFloat(content);
      } else {
        Write(content);
      }
    }
    WritePropertyEnd();
  }

  void WritePropertyEnd() {
    assert(state == State.Property);
    assert(childCount == 1);
    _stateStack.removeLast();
  }

  // Prepare a new property name, except this time, the property name
  // will be created by concatenating all the strings passed to
  // WritePropertyNameInner.
  void WritePropertyNameStart() {
    assert(state == State.Object);
    _currentPropertyName = "";

    IncrementChildCount();

    _stateStack.add(StateElement(type: State.Property));
    _stateStack.add(StateElement(type: State.PropertyName));
  }

  void WritePropertyNameEnd() {
    assert(state == State.PropertyName);
    assert(_currentPropertyName != null);

    _propertyNameStack.add(_currentPropertyName!);
    _currentPropertyName = null;
    // removeLast PropertyName, leaving Property state
    _stateStack.removeLast();
  }

  void WritePropertyNameInner(String str) {
    assert(state == State.PropertyName);
    assert(_currentPropertyName != null);
    _currentPropertyName += str;
  }

  void WritePropertyStart(dynamic name) {
    assert(state == State.Object);

    _propertyNameStack.add(name.toString());

    IncrementChildCount();

    _stateStack.add(new StateElement(type: State.Property));
  }

  // allow name to be string or int
  void WritePropertyStringOrInt<T>(T name, Action<SimpleJsonWriter> inner) {
    WritePropertyStart(name);

    inner(this);

    WritePropertyEnd();
  }

  void WriteArrayStart() {
    StartNewObject(container: true);
    var newObject = [];

    if (state == State.Property) {
      // This array is created as the value of a property,
      // inside an object.
      assert(currentCollection != null);
      assert(currentPropertyName != null);

      var propertyName = _propertyNameStack.removeLast();
      currentCollection![propertyName] = newObject;
      _collectionStack.add(newObject);
    } else if (state == State.Array) {
      // This array is created as the child of another array.
      assert(currentCollection != null);

      currentCollection!.add(newObject);
      _collectionStack.add(newObject);
    } else {
      // This array is the root object.
      assert(state == State.None);
      _jsonObject = newObject;
      _collectionStack.add(newObject);
    }

    _stateStack.add(StateElement(type: State.Array));
  }

  void WriteArrayEnd() {
    assert(state == State.Array);
    _collectionStack.removeLast();
    _stateStack.removeLast();
  }

  void WriteIntProperty(dynamic name, int content) {
    WritePropertyStart(name);
    WriteInt(content);
    WritePropertyEnd();
  }

  void WriteFloatProperty(dynamic name, double content) {
    WritePropertyStart(name);
    WriteFloat(content);
    WritePropertyEnd();
  }

// Add the value to the appropriate collection (array / object), given the current
  // context.
  void Write(dynamic value, {bool escape = true}) {
    if (value == null) {
      //console.error("Warning: trying to write a null string");
      return;
    }
    StartNewObject(container: false);
    _addToCurrentObject(value);
  }

  void WriteInt(int i) {
    StartNewObject(container: false);
    // Math.floor is used as a precaution:
    //     1. to ensure that the value is written as an integer
    //        (without a fractional part -> 1 instead of 1.0), even
    //        though it should be the default behaviour of
    //        JSON.serialize;
    //     2. to ensure that if a floating number is passed
    //        accidentally, it's converted to an integer.
    //
    // This guarantees savegame compatibility with the reference
    // implementation.
    _addToCurrentObject(i.floor());
  }

  // Since JSON doesn't support NaN and Infinity, these values
  // are converted here.
  void WriteFloat(double f) {
    StartNewObject(container: false);

    // TODO: Find an heap-allocation-free way to do this please!
    // _writer.write(formatStr, obj (the float)) requires boxing
    // Following implementation seems to work ok but requires creating temporary garbage string.
    //String floatStr = f.toString();
    if (f == double.infinity) {
      _addToCurrentObject(
          3.4E+38); // JSON doesn't support, do our best alternative
    } else if (f == double.negativeInfinity) {
      _addToCurrentObject(
          -3.4E+38); // JSON doesn't support, do our best alternative
    } else if (f == double.nan || f.isNaN) {
      _addToCurrentObject(0.0); // JSON doesn't support, not much we can do
    } else {
      _addToCurrentObject(f);
      //if (!floatStr.contains('.') && !floatStr.contains('E')) {
      //_addToCurrentObject(.0);
      //} // ensure it gets read back in as a floating point value
    }
  }

  void WriteNull() {
    StartNewObject(container: false);
    _addToCurrentObject(null);
  }

  void WriteStringStart() {
    StartNewObject(container: false);
    _currentString = '';
    _stateStack.add(StateElement(type: State.String));
  }

  void WriteStringEnd() {
    assert(state == State.String);
    _stateStack.removeLast();
    _addToCurrentObject(_currentString);
    _currentString = null;
  }

  void WriteStringInner(String str, {bool escape = true}) {
    assert(state == State.String);
    if (escape) {
      WriteEscapedString(str);
    } else {
      _currentString += str;
    }
  }

  void WriteEscapedString(String str) {
    for (var i = 0; i < str.length; i++) {
      var c = str[i];
      if (c.codeUnitAt(0) < ' '.codeUnitAt(0)) {
        // Don't write any control characters except \n and \t
        switch (c) {
          case '\n':
            _currentString += ('\\n');
            break;
          case '\t':
            _currentString += ('\\t');
            break;
        }
      } else {
        switch (c) {
          case '\\':
          case '"':
            _currentString += ('\\');
            _currentString += (c);
            break;
          default:
            _currentString += (c);
            break;
        }
      }
    }
  }

  @override
  String toString() {
    if (_jsonObject == null) {
      return '';
    }

    return jsonEncode(_jsonObject);
  }

  Map<String, dynamic> toJson() {
    return _jsonObject as Map<String, dynamic>;
  }

  // Prepare the state stack when adding new objects / values.
  void StartNewObject({required bool container}) {
    if (container) {
      assert(state == State.None ||
          state == State.Property ||
          state == State.Array);
    } else {
      assert(state == State.Property || state == State.Array);
    }

    if (state == State.Property) {
      assert(childCount == 0);
    }

    if (state == State.Array || state == State.Property) {
      IncrementChildCount();
    }
  }

  State get state {
    if (_stateStack.isNotEmpty) {
      return _stateStack.last.type;
    } else {
      return State.None;
    }
  }

  int get childCount {
    if (_stateStack.isNotEmpty) {
      return _stateStack.last.childCount;
    } else {
      return 0;
    }
  }

  // These getters peek all the different stacks.

  dynamic get currentCollection {
    if (_collectionStack.isNotEmpty) {
      return _collectionStack.last;
    } else {
      return null;
    }
  }

  String? get currentPropertyName {
    if (_propertyNameStack.isNotEmpty) {
      return _propertyNameStack.last;
    } else {
      return null;
    }
  }

  void IncrementChildCount() {
    assert(_stateStack.isNotEmpty);
    var currEl = _stateStack.removeLast();
    currEl.childCount++;
    _stateStack.add(currEl);
  }

  // This method did not exist in the original C# code. It adds
  // the given value to the current collection (used by Write methods).
  void _addToCurrentObject(dynamic value) {
    assert(currentCollection != null);
    if (state == State.Array) {
      (currentCollection as List).add(value);
    } else if (state == State.Property) {
      assert(currentPropertyName != null);
      (currentCollection as Map<String, dynamic>)[currentPropertyName!] = value;
      _propertyNameStack.removeLast();
    }
  }

  void WriteBool(bool? value) {
    if (value == null) {
      return;
    }

    StartNewObject(container: false);
    _addToCurrentObject(value);
  }
}

enum State { None, Object, Array, Property, PropertyName, String }

class StateElement {
  State type;
  int childCount = 0;

  StateElement({required this.type, this.childCount = 0});
}
