// ignore_for_file: non_constant_identifier_names

import 'dart:collection';
import 'dart:convert';
import 'dart:math';

import 'package:darq/darq.dart';
import 'package:meta/meta.dart';
import 'package:quiver/check.dart';
import 'package:quiver/collection.dart';
import 'package:quiver/strings.dart';

import 'choice.dart';
import 'choice_point.dart';
import 'container.dart';
import 'control_command.dart';
import 'debug_metadata.dart';
import 'divert.dart';
import 'error.dart';
import 'extension.dart';
import 'ink_list.dart';
import 'json.dart';
import 'list_definition.dart';
import 'list_definitions_origin.dart';
import 'logger.dart';
import 'native_function_call.dart';
import 'path.dart';
import 'pointer.dart';
import 'profiler.dart';
import 'push_pop.dart';
import 'runtime_object.dart';
import 'runtime_void.dart';
import 'search_result.dart';
import 'stack.dart';
import 'story_exception.dart';
import 'story_state.dart';
import 'tag.dart';
import 'value.dart';
import 'variable_assignment.dart';
import 'variable_reference.dart';
import 'variable_state.dart';

// Assumption: prevText is the snapshot where we saw a newline, and we're
// checking whether we're really done
// with that line. Therefore prevText will definitely end in a newline.
//
// We take tags into account too, so that a tag following a content line:
// Content
// # tag
// ... doesn't cause the tag to be wrongly associated with the content above.
enum OutputStateChange { NoChange, ExtendedBeyondNewline, NewlineRemoved }

/// General purpose delegate definition for bound EXTERNAL function definitions
/// from ink. Note that this version isn't necessary if you have a function
/// with three arguments or less - see the overloads of BindExternalFunction.
typedef ExternalFunction = Object? Function(List<Object> args);

class ExternalFunctionDef {
  ExternalFunction? function;
  bool lookaheadSafe = false;
}

final _logger = Logger('Story');

/// <summary>
/// Delegate definition for variable observation - see ObserveVariable.
/// </summary>
typedef VariableObserver = Object Function(
    String variableName, Object newValue);

/// <summary>
/// A Story is the core class that represents a complete Ink narrative, and
/// manages the evaluation and state of it.
/// </summary>
class Story extends RuntimeObject {
  /// The current version of the ink story file format.
  static const int inkVersionCurrent = 21;

  static var debug = false;

  // Version numbers are for engine itself and story file, rather
  // than the story state save format
  //  -- old engine, new format: always fail
  //  -- new engine, old format: possibly cope, based on this number
  // When incrementing the version number above, the question you
  // should ask yourself is:
  //  -- Will the engine be able to load an old story file from
  //     before I made these changes to the engine?
  //     If possible, you should support it, though it's not as
  //     critical as loading old save games, since it's an
  //     in-development problem only.

  /// <summary>
  /// The minimum legacy version of ink that can be loaded by the current version of the code.
  /// </summary>
  static const int inkVersionMinimumCompatible = 18;

  int _recursiveContinueCount = 0;

  bool _asyncSaving = false;
  Profiler? _profiler;
  Container? _temporaryEvaluationContainer;

  /// `state` is almost guaranteed to be set in the constructor, unless
  /// using the compiler-specific constructor which will likely not be used in
  /// the real world.
  late StoryState _state;

  /// `_mainContentContainer` is almost guaranteed to be set in the
  /// constructor, unless the json is malformed.
  late Container _mainContentContainer;
  ListDefinitionsOrigin? _listDefinitions;

  late Map<String, ExternalFunctionDef> _externals;
  ListMultimap<String, VariableObserver>? _variableObservers;
  bool _hasValidatedExternals = false;

  bool _asyncContinueActive = false;
  StoryState? _stateSnapshotAtLastNewline;

  bool _sawLookaheadUnsafeFunctionAfterNewline = false;

  /// An ink file can provide a fallback functions for when when an EXTERNAL has been left
  /// unbound by the client, and the fallback function will be called instead. Useful when
  /// testing a story in playmode, when it's not possible to write a client-side C# external
  /// function, but you don't want it to fail to run.
  bool allowExternalFunctionFallbacks = true;

  /// <summary>
  /// The list of Choice objects available at the current point in
  /// the Story. This list will be populated as the Story is stepped
  /// through with the Continue() method. Once canContinue becomes
  /// false, this list will be populated, and is usually
  /// (but not always) on the final Continue() step.
  /// </summary>
  List<Choice> get currentChoices {
    // Don't include invisible choices for external usage.
    var choices = <Choice>[];

    if (_state == null) {
      throw StateError('this._state');
    }

    for (var c in _state.currentChoices) {
      if (!c.isInvisibleDefault) {
        c.index = choices.length;
        choices.add(c);
      }
    }
    return choices;
  }

  /// The latest line of text to be generated from a Continue() call.
  String? get currentText {
    ifAsyncWeCant("call currentText since it's a work in progress");
    return state.currentText;
  }

  /// Gets a list of tags as defined with '#' in source that were seen
  /// during the latest Continue() call.
  @override
  List<String> get currentTags {
    ifAsyncWeCant("call currentTags since it's a work in progress");
    return state.currentTags;
  }

  /// Any errors generated during evaluation of the Story.
  @override
  List<String>? get currentErrors {
    return state.currentErrors;
  }

  /// Any warnings generated during evaluation of the Story.
  List<String>? get currentWarnings {
    return state.currentWarnings;
  }

  /// Whether the currentErrors list contains any errors.
  /// THIS MAY BE REMOVED - you should be setting an error handler directly
  /// using Story.onError.
  @override
  bool get hasError {
    return state.hasError;
  }

  /// Whether the currentWarnings list contains any warnings.
  @override
  bool get hasWarning {
    return state.hasWarning;
  }

  /// <summary>
  /// The VariablesState object contains all the global variables in the story.
  /// However, note that there's more to the state of a Story than just the
  /// global variables. This is a convenience accessor to the full state object.
  /// </summary>
  VariablesState get variablesState {
    return state.variablesState;
  }

  ListDefinitionsOrigin? get listDefinitions {
    return _listDefinitions;
  }

  //Container? get rootContentContainer => _r;

  SearchResult contentAtPath(Path path) {
    return mainContentContainer.contentAtPath(path);
  }

  void warning(String s) {
    addError(s, isWarning: true);
  }

  Pointer pointerAtPath(Path path) {
    if (path.length == 0) {
      return Pointer.Null;
    }

    Pointer p = Pointer();

    var pathLengthToUse = path.length;

    SearchResult? result;
    if (path.lastComponent!.isIndex) {
      _logger.finest('lastComponent is index');
      pathLengthToUse = path.length - 1;
      result = mainContentContainer.contentAtPath(path,
          partialPathLength: pathLengthToUse);
      p.container = result.container;
      p.index = path.lastComponent!.index;
    } else {
      _logger.finest('lastComponent is not index');
      result = mainContentContainer.contentAtPath(path);
      p.container = result.container;
      p.index = -1;
    }

    if (result.obj == null ||
        result.obj == mainContentContainer && pathLengthToUse > 0) {
      error("Failed to find content at path '$path',"
          ' and no approximation of it was possible.');
    } else if (result.approximate) {
      warning("Failed to find content at path '$path',"
          " so it was approximated to: '${result.obj!.path}'.");
    }

    return p;
  }

  /// <summary>
  /// The entire current state of the story including (but not limited to):
  ///
  ///  * Global variables
  ///  * Temporary variables
  ///  * Read/visit and turn counts
  ///  * The callstack and evaluation stacks
  ///  * The current threads
  ///
  /// </summary>
  @override
  StoryState get state => _state;

  /// Error handler for all runtime errors in ink - i.e. problems
  /// with the source ink itself that are only discovered when playing
  /// the story.
  /// It's strongly recommended that you assign an error handler to your
  /// story instance to avoid getting exceptions for ink errors.
  ErrorHandler? onError;

  /// Callback for when ContinueInternal is complete
  void Function()? onDidContinue;

  /// Callback for when a choice is about to be executed
  void Function(Choice choice)? onMakeChoice;

  /// Callback for when a function is about to be evaluated
  void Function(String? m, List<Object> objects)? onEvaluateFunction;

  /// Callback for when a function has been evaluated
  /// This is necessary because evaluating a function can cause continuing
  void Function(String m, List<Object> objects, String string, Object object)?
      onCompleteEvaluateFunction;

  /// Callback for when a path string is chosen
  void Function(String m, List<Object> objects)? onChoosePathString;

  /// <summary>
  /// Start recording ink profiling information during calls to Continue on Story.
  /// Return a Profiler instance that you can request a report from when you're finished.
  /// </summary>
  Profiler StartProfiling() {
    ifAsyncWeCant('start profiling');
    _profiler = Profiler();
    return _profiler!;
  }

  /// <summary>
  /// Stop recording ink profiling information during calls to Continue on Story.
  /// To generate a report from the profiler, call
  /// </summary>
  void EndProfiling() {
    _profiler = null;
  }

  // Warning: When creating a Story using this constructor, you need to
  // call ResetState on it before use. Intended for compiler use only.
  // For normal use, use the constructor that takes a json string.
  Story(Container? contentContainer, [List<ListDefinition>? lists]) {
    if (contentContainer != null) {
      _mainContentContainer = contentContainer;
    }

    if (lists != null) _listDefinitions = ListDefinitionsOrigin(lists);

    _externals = <String, ExternalFunctionDef>{};
  }

  /// <summary>
  /// Construct a Story object using a JSON string compiled through inklecate.
  /// </summary>
  factory Story.fromJson(String jsonString) {
    var story = Story(null);
    Map<String, dynamic> rootObject = jsonDecode(jsonString);

    dynamic versionObj = rootObject['inkVersion'];
    if (versionObj == null) {
      throw Exception(
          "ink version number not found. Are you sure it's a valid .ink.json file?");
    }

    var formatFromFile = versionObj as int;
    if (formatFromFile > inkVersionCurrent) {
      throw Exception(
          'Version of ink used to build story was newer than the current version of the engine');
    } else if (formatFromFile < inkVersionMinimumCompatible) {
      throw Exception(
          'Version of ink used to build story is too old to be loaded by this version of the engine');
    } else if (formatFromFile != inkVersionCurrent) {
      print(
          "WARNING: Version of ink used to build story ($formatFromFile) doesn't match current version of engine. Non-critical, but recommend synchronising.");
    }

    var rootToken = rootObject['root'];
    if (rootToken == null) {
      throw Exception(
          "Root node for ink not found. Are you sure it's a valid .ink.json file?");
    }

    if (rootObject.containsKey('listDefs')) {
      Map<String, dynamic> listDefsObj =
          Map<String, dynamic>.from(rootObject['listDefs']);
      story._listDefinitions =
          JsonSerialization.JTokenToListDefinitions(listDefsObj);
    }

    story._mainContentContainer =
        JsonSerialization.JTokenToRuntimeObject(rootToken) as Container;

    story.resetState();
    return story;
  }

  // Throw an exception that gets caught and causes AddError to be called,
  // then exits the flow.
  void error(String message, [bool useEndLineNumber = false]) {
    var e = StoryException(message);
    e.useEndLineNumber = useEndLineNumber;
    throw e;
  }

  void ifAsyncWeCant(String activityStr) {
    if (_asyncContinueActive) {
      throw Exception("Can't " +
          activityStr +
          '. Story is in the middle of a ContinueAsync(). Make more ContinueAsync() calls or a single Continue() call beforehand.');
    }
  }

  Container get mainContentContainer =>
      _temporaryEvaluationContainer ?? _mainContentContainer;

  /// <summary>
  /// Reset the Story back to its initial state as it was when it was
  /// first constructed.
  /// </summary>
  void resetState() {
    // TODO: Could make this possible
    ifAsyncWeCant('ResetState');

    _state = StoryState(this);
    _state.variablesState.variableChangedEvent = _variableStateDidChangeEvent;

    resetGlobals();
  }

  void ResetErrors() {
    _state.resetErrors();
  }

  /// <summary>
  /// Unwinds the callstack. Useful to reset the Story's evaluation
  /// without actually changing any meaningful state, for example if
  /// you want to exit a section of story prematurely and tell it to
  /// go elsewhere with a call to ChoosePathString(...).
  /// Doing so without calling ResetCallstack() could cause unexpected
  /// issues if, for example, the Story was in a tunnel already.
  /// </summary>
  void resetCallstack() {
    ifAsyncWeCant('ResetCallstack');
    _state.ForceEnd();
  }

  void resetGlobals() {
    if (_mainContentContainer.namedContent.containsKey('global decl')) {
      var originalPointer = state.currentPointer.copyWith();

      choosePath(Path('global decl'), incrementingTurnIndex: false);

      // Continue, but without validating external bindings,
      // since we may be doing this reset at initialisation time.
      _continueInternal();

      state.currentPointer = originalPointer;
    }

    state.variablesState.snapshotDefaultGlobals();
  }

  void choosePath(Path p, {bool incrementingTurnIndex = true}) {
    state.SetChosenPath(p, incrementingTurnIndex);

    // Take a note of newly visited containers for read counts etc
    _visitChangedContainersDueToDivert();
  }

  /// <summary>
  /// Chooses the Choice from the currentChoices list with the given
  /// index. Internally, this sets the current content path to that
  /// pointed to by the Choice, ready to continue story evaluation.
  /// </summary>
  void chooseChoiceIndex(int choiceIdx) {
    var choices = currentChoices;
    assert(choiceIdx >= 0 && choiceIdx < choices.length, 'choice out of range');

    // Replace callstack with the one from the thread at the choosing point,
    // so that we can jump into the right place in the flow.
    // This is important in case the flow was forked by a new thread, which
    // can create multiple leading edges for the story, each of
    // which has its own context.
    var choiceToChoose = choices[choiceIdx];
    onMakeChoice?.call(choiceToChoose);
    state.callStack.currentThread = choiceToChoose.threadAtGeneration;

    _logger.finest('choiceToChoose : ${choiceToChoose.toJson()}');
    choosePath(choiceToChoose.targetPath!);
  }

  /// <summary>
  /// Checks if a function exists.
  /// </summary>
  /// <returns>True if the function exists, else false.</returns>
  /// <param name="functionName">The name of the function as declared in ink.</param>
  bool hasFunction(String functionName) {
    try {
      return knotContainerWithName(functionName) != null;
    } catch (e) {
      return false;
    }
  }

  Container? knotContainerWithName(String name) {
    _logger.finest('[knotContainerWithName] ($name)');
    if (mainContentContainer.namedContent.containsKey(name)) {
      _logger.finest('[knotContainerWithName] contains $name');

      return mainContentContainer.namedContent[name] as Container;
    } else {
      _logger.finest('[knotContainerWithName] not contains $name');

      return null;
    }
  }

  final List<Container> _prevContainers = <Container>[];

  //void VisitChangedContainersDueToDivert()
  void _visitChangedContainersDueToDivert() {
    _logger.finest('[_visitChangedContainersDueToDivert]');
    var previousPointer = state.previousPointer;
    var pointer = state.currentPointer;

    // Unless we're pointing *directly* at a piece of content, we don't do
    // counting here. Otherwise, the main stepping function will do the counting.
    if (pointer.isNull || pointer.index == -1) {
      _logger.info(
          "[_visitChangedContainersDueToDivert] Unless we're pointing *directly* at a piece of content, we don't do"
          'counting here. Otherwise, the main stepping function will do the counting.');
      return;
    }

    // First, find the previously open set of containers
    _logger.finest(
        '[_visitChangedContainersDueToDivert] First, find the previously open set of containers');

    _prevContainers.clear();
    if (!previousPointer.isNull) {
      _logger.finest(
          '[_visitChangedContainersDueToDivert] previousPointer is not Null');

      var prevAncestor = asOrNull<Container>(previousPointer.resolve()) ??
          asOrNull<Container>(previousPointer.container);
      _logger.finest(
          '[_visitChangedContainersDueToDivert] prevAncestor $prevAncestor');
      while (prevAncestor != null) {
        _prevContainers.add(prevAncestor);
        prevAncestor = asOrNull<Container>(prevAncestor.parent);
      }
    }

    // If the new object is a container itself, it will be visited automatically at the next actual
    // content step. However, we need to walk up the new ancestry to see if there are more new containers
    var currentChildOfContainer = pointer.resolve();

    // Invalid pointer? May happen if attemptingto
    if (currentChildOfContainer == null) {
      _logger.warning('[_visitChangedContainersDueToDivert] invalid pointer?');
      return;
    }
    var currentContainerAncestor =
        asOrNull<Container>(currentChildOfContainer.parent);
    _logger.info(
        '[_visitChangedContainersDueToDivert] currentContainerAncestor : $currentContainerAncestor');

    var allChildrenEnteredAtStart = true;
    while (currentContainerAncestor != null &&
        (!_prevContainers.contains(currentContainerAncestor) ||
            currentContainerAncestor.countingAtStartOnly)) {
      // Check whether this ancestor container is being entered at the start,
      // by checking whether the child object is the first.
      _logger.finest(
          '[_visitChangedContainersDueToDivert]  Check whether this ancestor container is being entered at the start,'
          "by checking whether the child object is the first.");
      var enteringAtStart = currentContainerAncestor.content.isNotEmpty &&
          currentChildOfContainer == currentContainerAncestor.content[0] &&
          allChildrenEnteredAtStart;

      // Don't count it as entering at start if we're entering random somewhere within
      // a container B that happens to be nested at index 0 of container A. It only counts
      // if we're diverting directly to the first leaf node.
      if (!enteringAtStart) {
        allChildrenEnteredAtStart = false;
      }

      // Mark a visit to this container
      _logger.finest(
          '[_visitChangedContainersDueToDivert] Mark a visit to this container');
      _visitContainer(currentContainerAncestor, atStart: enteringAtStart);

      currentChildOfContainer = currentContainerAncestor;
      currentContainerAncestor = asOrNull(currentContainerAncestor.parent);
    }
  }

  // string PopChoiceStringAndTags(ref List<string> tags)
  String PopChoiceStringAndTags(List<String> tags) {
    var choiceOnlyStrVal = state.popEvaluationStack() as StringValue;

    while (state.evaluationStack.isNotEmpty &&
        asOrNull<Tag>(state.peekEvaluationStack()) != null) {
      //tags ??= List<string>();
      var tag = state.popEvaluationStack() as Tag;
      tags.insert(0, tag.text); // popped in reverse order
    }

    return choiceOnlyStrVal.value;
  }

  Choice? _processChoice(ChoicePoint choicePoint) {
    _logger.fine('=== _processChoice ===');
    var showChoice = true;

    // Don't create choice if choice point doesn't pass conditional
    if (choicePoint.hasCondition) {
      _logger.finest('choicePoint has condition');
      var conditionValue = state.popEvaluationStack();
      if (!_isTruthy(conditionValue)) {
        showChoice = false;
      }
    }

    var startText = '';
    var choiceOnlyText = '';
    List<String> tags = [];

    if (choicePoint.hasChoiceOnlyContent) {
      choiceOnlyText = PopChoiceStringAndTags(tags);
    }

    if (choicePoint.hasStartContent) {
      startText = PopChoiceStringAndTags(tags);
    }

    // Don't create choice if player has already read this content
    if (choicePoint.onceOnly) {
      _logger.finest('choicePoint onceOnly');
      var visitCount = state.visitCountForContainer(choicePoint.choiceTarget!);
      if (visitCount > 0) {
        showChoice = false;
      }
    }

    // We go through the full process of creating the choice above so
    // that we consume the content for it, since otherwise it'll
    // be shown on the output stream.
    if (!showChoice) {
      return null;
    }

    var choice = Choice();
    choice.targetPath = choicePoint.pathOnChoice;
    choice.sourcePath = choicePoint.path.toString();
    choice.isInvisibleDefault = choicePoint.isInvisibleDefault;
    choice.tags = tags;

    // We need to capture the state of the callstack at the point where
    // the choice was generated, since after the generation of this choice
    // we may go on to pop out from a tunnel (possible if the choice was
    // wrapped in a conditional), or we may pop out from a thread,
    // at which point that thread is discarded.
    // Fork clones the thread, gives it a new ID, but without affecting
    // the thread stack itself.
    choice.threadAtGeneration = state.callStack.forkThread();

    // Set final text for the choice
    choice.text = (startText + choiceOnlyText).trim();

    return choice;
  }

  // Mark a container as having been visited
  // https://github.com/inkle/ink/blob/01898140be43d29baac70c1cb6544bdb10164209/ink-engine-runtime/Story.cs#L843
  void _visitContainer(Container container, {required bool atStart}) {
    _logger.finest('[visitContainer]');

    if (!container.countingAtStartOnly || atStart) {
      _logger
          .finest('[visitContainer] !container.countingAtStartOnly || atStart');

      if (container.visitsShouldBeCounted) {
        _logger.finest('[visitContainer] visitsShouldBeCounted');
        state.incrementVisitCountForContainer(container);
      }

      if (container.turnIndexShouldBeCounted) {
        _logger.finest('[visitContainer] turnIndexShouldBeCounted');
        state.recordTurnIndexVisitToContainer(container);
      }
    }
  }

  // Does the expression result represented by this object evaluate to true?
  // e.g. is it a Number that's not equal to 1?
  bool _isTruthy(RuntimeObject obj) {
    var truthy = false;
    if (obj is Value) {
      var val = obj;

      if (val is DivertTargetValue) {
        var divTarget = val;
        error("Shouldn't use a divert target (to " +
            divTarget.targetPath.toString() +
            ") as a conditional value. Did you intend a function call 'likeThis()' or a read count check 'likeThis'? (no arrows)");
        return false;
      }

      return val.isTruthy;
    }
    return truthy;
  }

  void _variableStateDidChangeEvent(
      String variableName, RuntimeObject newValueObj) {
    if (_variableObservers == null) {
      return;
    }

    var observers = _variableObservers![variableName];
    if (observers.isNotEmpty) {
      for (var observer in observers) {
        if (newValueObj is! Value) {
          throw Exception(
              "Tried to get the value of a variable that isn't a standard type");
        }
        var val = newValueObj;

        observer(variableName, val.valueObject);
      }
    }
  }

  // Maximum snapshot stack:
  //  - stateSnapshotDuringSave -- not retained, but returned to game code
  //  - _stateSnapshotAtLastNewline (has older patch)
  //  - _state (current, being patched)

  void _stateSnapshot() {
    _stateSnapshotAtLastNewline = _state;
    _state = _state.copyAndStartPatching();
  }

  void _restoreStateSnapshot() {
    _logger.finest('_restoreStateSnapshot()');
    // Patched state had temporarily hijacked our
    // VariablesState and set its own callstack on it,
    // so we need to restore that.
    // If we're in the middle of saving, we may also
    // need to give the VariablesState the old patch.

    _logger.finest('prev.curr : ${_stateSnapshotAtLastNewline!.currentText}');

    _stateSnapshotAtLastNewline!.restoreAfterPatch();

    _state = _stateSnapshotAtLastNewline!;
    _stateSnapshotAtLastNewline = null;

    // If save completed while the above snapshot was
    // active, we need to apply any changes made since
    // the save was started but before the snapshot was made.
    if (!_asyncSaving) {
      _state.applyAnyPatch();
    }
    _logger.finest('state.curr : ${state.currentText}');
  }

  void _discardSnapshot() {
    // Normally we want to integrate the patch
    // into the main global/counts dictionaries.
    // However, if we're in the middle of async
    // saving, we simply stay in a "patching" state,
    // albeit with the newer cloned patch.
    if (!_asyncSaving) {
      _state.applyAnyPatch();
    }

    // No longer need the snapshot.
    _stateSnapshotAtLastNewline = null;
  }

  /// <summary>
  /// Advanced usage!
  /// If you have a large story, and saving state to JSON takes too long for your
  /// framerate, you can temporarily freeze a copy of the state for saving on
  /// a separate thread. Internally, the engine maintains a "diff patch".
  /// When you've finished saving your state, call BackgroundSaveComplete()
  /// and that diff patch will be applied, allowing the story to continue
  /// in its usual mode.
  /// </summary>
  /// <returns>The state for background thread save.</returns>
  StoryState CopyStateForBackgroundThreadSave() {
    ifAsyncWeCant('start saving on a background thread');
    if (_asyncSaving) {
      throw Exception(
          "Story is already in background saving mode, can't call CopyStateForBackgroundThreadSave again!");
    }
    var stateToSave = _state;
    _state = _state.copyAndStartPatching();
    _asyncSaving = true;
    return stateToSave;
  }

  /// <summary>
  /// See CopyStateForBackgroundThreadSave. This method releases the
  /// "frozen" save state, applying its patch that it was using internally.
  /// </summary>
  void backgroundSaveComplete() {
    // CopyStateForBackgroundThreadSave must be called outside
    // of any async ink evaluation, since otherwise you'd be saving
    // during an intermediate state.
    // However, it's possible to *complete* the save in the middle of
    // a glue-lookahead when there's a state stored in _stateSnapshotAtLastNewline.
    // This state will have its own patch that is newer than the save patch.
    // We hold off on the final apply until the glue-lookahead is finished.
    // In that case, the apply is always done, it's just that it may
    // apply the looked-ahead changes OR it may simply apply the changes
    // made during the save process to the old _stateSnapshotAtLastNewline state.
    if (_stateSnapshotAtLastNewline == null) {
      _state.applyAnyPatch();
    }

    _asyncSaving = false;
  }

  void callExternalFunction(String funcName, int numberOfArguments) {
    ExternalFunctionDef? funcDef;
    Container? fallbackFunctionContainer;

    var foundExternal = _externals.containsKey(funcName);
    if (foundExternal) {
      funcDef = _externals[funcName];
    }

    if (foundExternal &&
        funcDef?.lookaheadSafe == false &&
        _stateSnapshotAtLastNewline != null) {
      _sawLookaheadUnsafeFunctionAfterNewline = true;
      return;
    }

    // Try to use fallback function?
    if (!foundExternal) {
      if (allowExternalFunctionFallbacks) {
        fallbackFunctionContainer = knotContainerWithName(funcName);
        assert(
            fallbackFunctionContainer != null,
            "Trying to call EXTERNAL function '" +
                funcName +
                "' which has not been bound, and fallback ink function could not be found.");

        // Divert direct into fallback function and we're done
        state.callStack.push(PushPopType.Function,
            outputStreamLengthWithPushed: state.outputStream.length);
        state.divertedPointer = Pointer.startOf(fallbackFunctionContainer);
        return;
      } else {
        assert(
            false,
            "Trying to call EXTERNAL function '" +
                funcName +
                "' which has not been bound (and ink fallbacks disabled).");
      }
    }

    // Pop arguments
    var arguments = <Object>[];
    for (var i = 0; i < numberOfArguments; ++i) {
      var poppedObj = state.popEvaluationStack() as Value;
      var valueObj = poppedObj.valueObject;
      arguments.add(valueObj);
    }

    // Reverse arguments from the order they were popped,
    // so they're the right way round again.
    arguments = arguments.reversed.toList();

    // Run the function!
    var funcResult = funcDef!.function?.call(arguments.toList());

    // Convert return value (if any) to the a type that the ink engine can use
    RuntimeObject? returnObj;
    if (funcResult != null) {
      returnObj = Value.create(funcResult);
      assert(
          returnObj != null,
          'Could not create ink value from returned object of type ' +
              funcResult.runtimeType.toString());
    } else {
      returnObj = RuntimeVoid();
    }

    state.pushEvaluationStack(returnObj!);
  }

  /// Checks whether contentObj is a control or flow object rather than a piece of content,
  /// and performs the required command if necessary.
  ///
  ///  returns `true` if object was logic or flow control, `false` if it's normal content.
  /// [contentObj] Content object.
  //        bool PerformLogicAndFlowControl(Runtime.Object contentObj)
  bool _performLogicAndFlowControl(RuntimeObject? contentObj) {
    _logger.fine('[_performLogicAndFlowControl] ($contentObj)');
    if (contentObj == null) {
      return false;
    }

    // Divert
    if (contentObj is Divert) {
      _logger.info(
          '[_performLogicAndFlowControl] contentObj is Divert : $contentObj');
      var currentDivert = contentObj;

      if (currentDivert.isConditional) {
        _logger.finest(
            '[_performLogicAndFlowControl] currentDivert is Conditional');

        var conditionValue = state.popEvaluationStack();

        // False conditional? Cancel divert
        if (!_isTruthy(conditionValue)) {
          _logger.info(
              '[_performLogicAndFlowControl] False conditional? Cancel divert');

          return true;
        }
      }

      if (currentDivert.hasVariableTarget) {
        _logger.finest(
            '[_performLogicAndFlowControl] currentDivert($currentDivert) has VariableTarget');

        var varName = currentDivert.variableDivertName;

        var varContents = state.variablesState.getVariableWithName(varName!);
        _logger.finest(
            '[_performLogicAndFlowControl] varContents($varName) : $varContents');
        if (varContents == null) {
          error(
              '[_performLogicAndFlowControl]  Tried to divert using a target from a variable that could not be found (' +
                  varName +
                  ')');
        } else if (varContents is! DivertTargetValue) {
          var intContent = asOrNull<IntValue>(varContents);

          var errorMessage =
              '[_performLogicAndFlowControl]  Tried to divert to a target from a variable, but the variable (' +
                  varName +
                  ") didn't contain a divert target, it ";
          if (intContent != null && intContent.value == 0) {
            errorMessage += "was empty/null (the value 0).";
          } else {
            errorMessage += "contained '" + varContents.toString() + "'.";
          }

          error(errorMessage);
        }

        var target = varContents as DivertTargetValue;
        state.divertedPointer = pointerAtPath(target.targetPath);
        _logger.finest('state.divertedPointer : ${state.divertedPointer}');
      } else if (currentDivert.isExternal) {
        _logger
            .finest('[_performLogicAndFlowControl] currentDivert isExternal');

        callExternalFunction(
            currentDivert.targetPathString!, currentDivert.externalArgs);
        return true;
      } else {
        _logger.finest('[_performLogicAndFlowControl] else - $currentDivert');

        state.divertedPointer = currentDivert.targetPointer.copyWith();
        _logger
            .finest('[_performLogicAndFlowControl] ${state.divertedPointer}');
      }

      if (currentDivert.pushesToStack) {
        _logger.finest(
            '[_performLogicAndFlowControl] currentDivert.pushesToStack is true');

        state.callStack.push(currentDivert.stackPushType!,
            outputStreamLengthWithPushed: state.outputStream.length);
      }

      if (state.divertedPointer.isNull && !currentDivert.isExternal) {
        // Human readable name available - runtime divert is part of a hard-written divert that to missing content
        if (currentDivert.debugMetadata?.sourceName != null) {
          error("Divert target doesn't exist: " +
              currentDivert.debugMetadata!.sourceName!);
        } else {
          error('Divert resolution failed: ' + currentDivert.toString());
        }
      }

      return true;
    }

    // Start/end an expression evaluation? Or print out the result?
    else if (contentObj is ControlCommand) {
      _logger
          .info('[_performLogicAndFlowControl]  ControlCommand : $contentObj');

      var evalCommand = contentObj;

      switch (evalCommand.commandType) {
        case CommandType.EvalStart:
          assert(state.inExpressionEvaluation == false,
              'Already in expression evaluation?');
          state.inExpressionEvaluation = true;
          break;

        case CommandType.EvalEnd:
          assert(state.inExpressionEvaluation == true,
              'Not in expression evaluation mode');
          state.inExpressionEvaluation = false;
          break;

        case CommandType.EvalOutput:
          // If the expression turned out to be empty, there may not be anything on the stack
          if (state.evaluationStack.isNotEmpty) {
            var output = state.popEvaluationStack();

            // Functions may evaluate to Void, in which case we skip output
            if (output is! RuntimeVoid) {
              // TODO: Should we really always blanket convert to string?
              // It would be okay to have numbers in the output stream the
              // only problem is when exporting text for viewing, it skips over numbers etc.
              var text = StringValue(output.toString());

              state.PushToOutputStream(text);
            }
          }
          break;

        case CommandType.NoOp:
          break;

        case CommandType.Duplicate:
          state.pushEvaluationStack(state.peekEvaluationStack());
          break;

        case CommandType.PopEvaluatedValue:
          state.popEvaluationStack();
          break;

        case CommandType.PopFunction:
        case CommandType.PopTunnel:
          var popType = evalCommand.commandType == CommandType.PopFunction
              ? PushPopType.Function
              : PushPopType.Tunnel;

          // Tunnel onwards is allowed to specify an optional override
          // divert to go to immediately after returning: ->-> target
          DivertTargetValue? overrideTunnelReturnTarget;
          if (popType == PushPopType.Tunnel) {
            var popped = state.popEvaluationStack();
            overrideTunnelReturnTarget = asOrNull<DivertTargetValue>(popped);
            if (overrideTunnelReturnTarget == null) {
              assert(popped is RuntimeVoid,
                  "Expected void if ->-> doesn't override target");
            }
          }

          if (state.tryExitFunctionEvaluationFromGame()) {
            break;
          } else if (state.callStack.currentElement.type != popType ||
              !state.callStack.canPop()) {
            var names = <PushPopType, String>{};
            names[PushPopType.Function] =
                'function return statement (~ return)';
            names[PushPopType.Tunnel] = 'tunnel onwards statement (->->)';

            var expected = names[state.callStack.currentElement.type];
            if (!state.callStack.canPop()) {
              expected = 'end of flow (-> END or choice)';
            }

            var errorMsg = 'Found ${names[popType]}, when expected $expected}';

            error(errorMsg);
          } else {
            state.popCallstack();

            // Does tunnel onwards override by diverting to a new ->-> target?
            if (overrideTunnelReturnTarget != null) {
              state.divertedPointer =
                  pointerAtPath(overrideTunnelReturnTarget.targetPath);
            }
          }

          break;

        case CommandType.BeginString:
          state.PushToOutputStream(evalCommand);

          assert(state.inExpressionEvaluation == true,
              'Expected to be in an expression when evaluating a string');
          state.inExpressionEvaluation = false;
          break;

        // Leave it to story.currentText and story.currentTags to sort out the text from the tags
        // This is mostly because we can't always rely on the existence of EndTag, and we don't want
        // to try and flatten dynamic tags to strings every time \n is pushed to output
        case CommandType.BeginTag:
          state.PushToOutputStream(evalCommand);
          break;

        case CommandType.EndTag:
          // EndTag has 2 modes:
          //  - When in string evaluation (for choices)
          //  - Normal
          //
          // The only way you could have an EndTag in the middle of
          // string evaluation is if we're currently generating text for a
          // choice, such as:
          //
          //   + choice # tag
          //
          // In the above case, the ink will be run twice:
          //  - First, to generate the choice text. String evaluation
          //    will be on, and the final string will be pushed to the
          //    evaluation stack, ready to be popped to make a Choice
          //    object.
          //  - Second, when ink generates text after choosing the choice.
          //    On this ocassion, it's not in string evaluation mode.
          //
          // On the writing side, we disallow manually putting tags within
          // strings like this:
          //
          //   {"hello # world"}
          //
          // So we know that the tag must be being generated as part of
          // choice content. Therefore, when the tag has been generated,
          // we push it onto the evaluation stack in the exact same way
          // as the string for the choice content.
          if (state.inStringEvaluation) {
            var contentStackForTag = Stack<StringValue>();
            int outputCountConsumed = 0;

            for (int i = state.outputStream.length - 1; i >= 0; --i) {
              var obj = state.outputStream[i];

              outputCountConsumed++;

              var command = asOrNull<ControlCommand>(obj);
              if (command != null) {
                if (command.commandType == CommandType.BeginTag) {
                  break;
                } else {
                  Exception(
                      "Unexpected ControlCommand while extracting tag from choice");
                  break;
                }
              }

              if (obj is StringValue) {
                contentStackForTag.push(obj);
              }
            }

            // Consume the content that was produced for this string
            state.PopFromOutputStream(outputCountConsumed);

            var sb = StringBuffer();
            for (StringValue strVal in contentStackForTag) {
              sb.write(strVal.value);
            }

            var choiceTag = Tag(state.cleanOutputWhitespace(sb.toString()));

            // Pushing to the evaluation stack means it gets picked up
            // when a Choice is generated from the next Choice Point.
            state.pushEvaluationStack(choiceTag);
          }

          // Otherwise! Simply push EndTag, so that in the output stream we
          // have a structure of: [BeginTag, "the tag content", EndTag]
          else {
            state.PushToOutputStream(evalCommand);
          }
          break;

        // Dynamic strings and tags are built in the same way
        case CommandType.EndString:
          // Since we're iterating backward through the content,
          // build a stack so that when we build the string,
          // it's in the right order
          var contentStackForString = Stack<RuntimeObject>();
          var contentToRetain = Stack<RuntimeObject>();

          var outputCountConsumed = 0;
          for (var i = state.outputStream.length - 1; i >= 0; --i) {
            var obj = state.outputStream[i];

            outputCountConsumed++;

            var command = asOrNull<ControlCommand>(obj);
            if (command != null &&
                command.commandType == CommandType.BeginString) {
              break;
            }

            if (obj is Tag) {
              contentToRetain.push(obj);
            }

            if (obj is StringValue) {
              contentStackForString.push(obj);
            }
          }

          // Consume the content that was produced for this string
          state.PopFromOutputStream(outputCountConsumed);

          // Rescue the tags that we want actually to keep on the output stack
          // rather than consume as part of the string we're building.
          // At the time of writing, this only applies to Tag objects generated
          // by choices, which are pushed to the stack during string generation.
          for (var rescuedTag in contentToRetain) {
            state.PushToOutputStream(rescuedTag);
          }

          // Build string out of the content we collected
          var sb = StringBuffer();
          for (var c in contentStackForString) {
            sb.write(c.toString());
          }

          // Return to expression evaluation (from content mode)
          state.inExpressionEvaluation = true;
          state.pushEvaluationStack(StringValue(sb.toString()));
          break;

        case CommandType.ChoiceCount:
          var choiceCount = state.generatedChoices.length;
          state.pushEvaluationStack(IntValue(choiceCount));
          break;

        case CommandType.Turns:
          state.pushEvaluationStack(IntValue(state.currentTurnIndex + 1));
          break;

        case CommandType.TurnsSince:
        case CommandType.ReadCount:
          var target = state.popEvaluationStack();
          if (target is! DivertTargetValue) {
            var extraNote = '';
            if (target is IntValue) {
              extraNote =
                  ". Did you accidentally pass a read count ('knot_name') instead of a target ('-> knot_name')?";
            }
            error(
                'TURNS_SINCE expected a divert target (knot, stitch, label name), but saw ' +
                    target.toString() +
                    extraNote);
            break;
          }

          var divertTarget = target as DivertTargetValue;
          var container = asOrNull<Container>(
              contentAtPath(divertTarget.targetPath).correctObj);

          int eitherCount;
          if (container != null) {
            if (evalCommand.commandType == CommandType.TurnsSince) {
              eitherCount = state.turnsSinceForContainer(container);
            } else {
              eitherCount = state.visitCountForContainer(container);
            }
          } else {
            if (evalCommand.commandType == CommandType.TurnsSince) {
              eitherCount = -1; // turn count, default to never/unknown
            } else {
              eitherCount = 0;
            } // visit count, assume 0 to default to allowing entry

            warning('Failed to find container for ' +
                evalCommand.toString() +
                ' lookup at ' +
                divertTarget.targetPath.toString());
          }

          state.pushEvaluationStack(IntValue(eitherCount));
          break;

        case CommandType.Random:
          {
            var maxInt = asOrNull<IntValue>(state.popEvaluationStack());
            var minInt = asOrNull<IntValue>(state.popEvaluationStack());

            if (minInt == null) {
              error('Invalid value for minimum parameter of RANDOM(min, max)');
            }

            if (maxInt == null) {
              error('Invalid value for maximum parameter of RANDOM(min, max)');
            }

            // +1 because it's inclusive of min and max, for e.g. RANDOM(1,6) for a dice roll.
            int randomRange;

            try {
              checkState(maxInt!.value - minInt!.value + 1 < 1 << 32);
              randomRange = maxInt.value - minInt.value + 1;
            } catch (e) {
              randomRange = 1 << 32;
              error(
                  'RANDOM was called with a range that exceeds the size that ink numbers can use.');
            }
            if (randomRange <= 0) {
              error('RANDOM was called with minimum as ${minInt!.value}'
                  ' and maximum as ${maxInt!.value}'
                  '. The maximum must be larger');
            }

            var resultSeed = state.storySeed + state.previousRandom;
            var random = Random(resultSeed);

            var nextRandom = random.nextInt(1 << 32);
            var chosenValue = (nextRandom % randomRange) + minInt!.value;
            state.pushEvaluationStack(IntValue(chosenValue));

            // Next random number (rather than keeping the Random object around)
            state.previousRandom = nextRandom;
            break;
          }

        case CommandType.SeedRandom:
          var seed = asOrNull(state.popEvaluationStack());
          if (seed == null) {
            error('Invalid value passed to SEED_RANDOM');
          }

          // Story seed affects both RANDOM and shuffle behaviour
          state.storySeed = seed!.value;
          state.previousRandom = 0;

          // SEED_RANDOM returns nothing.
          state.pushEvaluationStack(RuntimeVoid());
          break;

        case CommandType.VisitIndex:
          var count =
              state.visitCountForContainer(state.currentPointer.container!) -
                  1; // index not count
          state.pushEvaluationStack(IntValue(count));
          break;

        case CommandType.SequenceShuffleIndex:
          var shuffleIndex = _nextSequenceShuffleIndex();
          state.pushEvaluationStack(IntValue(shuffleIndex));
          break;

        case CommandType.StartThread:
          // Handled in main step function
          break;

        case CommandType.Done:
          _logger.finest('CommandType.Done');
          // We may exist in the context of the initial
          // act of creating the thread, or in the context of
          // evaluating the content.
          if (state.callStack.canPopThread) {
            state.callStack.popThread();
          }
          // In normal flow - allow safe exit without warning
          else {
            _logger.finest('In normal flow - allow safe exit without warning');

            state.didSafeExit = true;

            // Stop flow in current thread
            _logger.finest('Stop flow in current thread');
            state.currentPointer = Pointer.Null;
          }

          break;

        // Force flow to end completely
        case CommandType.End:
          state.ForceEnd();
          break;

        case CommandType.ListFromInt:
          var intVal = asOrNull<IntValue>(state.popEvaluationStack());
          var listNameVal = state.popEvaluationStack() as StringValue;

          if (intVal == null) {
            throw StoryException(
                'Passed non-integer when creating a list element from a numerical value.');
          }

          ListValue? generatedListValue;
          if (listDefinitions == null) {
            throw ArgumentError.notNull('this.listDefinitions');
          }
          var foundListDef = listDefinitions!.getDefinition(listNameVal.value);
          if (foundListDef != null) {
            var foundItem = foundListDef.getItemWithValue(intVal.value);
            if (foundItem.isValue) {
              generatedListValue =
                  ListValue.single(foundItem.asValue!.value, intVal.value);
            }
          } else {
            throw StoryException(
                'Failed to find LIST called ' + listNameVal.value);
          }

          generatedListValue ??= ListValue();

          state.pushEvaluationStack(generatedListValue);
          break;

        case CommandType.ListRange:
          {
            var max = asOrNull<Value>(state.popEvaluationStack());
            var min = asOrNull<Value>(state.popEvaluationStack());

            var targetList = asOrNull<ListValue>(state.popEvaluationStack());

            if (targetList == null || min == null || max == null) {
              throw StoryException(
                  'Expected list, minimum and maximum for LIST_RANGE');
            }

            var result = targetList.value
                .listWithSubRange(min.valueObject, max.valueObject);

            state.pushEvaluationStack(ListValue(result));
            break;
          }

        case CommandType.ListRandom:
          {
            var listVal = state.popEvaluationStack();
            if (listVal is! ListValue) {
              throw StoryException('Expected list for LIST_RANDOM');
            }

            var list = listVal.value;

            InkList newList;

            // List was empty: return empty list
            if (list.isEmpty) {
              newList = InkList();
            }

            // Non-empty source list
            else {
              // Generate a random index for the element to take
              var resultSeed = state.storySeed + state.previousRandom;
              var random = Random(resultSeed);

              var nextRandom = random.nextInt(4294967296);
              var listItemIndex = nextRandom % list.length;

              // Iterate through to get the random element
              var listEnumerator = list.entries.iterator;
              for (var i = 0; i <= listItemIndex; i++) {
                listEnumerator.moveNext();
              }
              var randomItem = listEnumerator.current;

              // Origin list is simply the origin of the one element
              newList = InkList.fromOrigin(randomItem.key.originName!, this);
              newList[randomItem.key] = randomItem.value;

              state.previousRandom = nextRandom;
            }

            state.pushEvaluationStack(ListValue(newList));
            break;
          }

        default:
          error('unhandled ControlCommand: ' + evalCommand.toString());
          break;
      }

      return true;
    }

    // Variable assignment
    else if (contentObj is VariableAssignment) {
      var varAss = contentObj;
      var assignedVal = state.popEvaluationStack();

      // When in temporary evaluation, don't create new variables purely within
      // the temporary context, but attempt to create them globally
      //var prioritiseHigherInCallStack = _temporaryEvaluationContainer != null;

      state.variablesState.assign(varAss, assignedVal);

      return true;
    }

    // Variable reference
    else if (contentObj is VariableReference) {
      var varRef = contentObj as VariableReference;
      RuntimeObject? foundValue;

      // Explicit read count value
      if (varRef.pathForCount != null) {
        var container = varRef.containerForCount;
        var count = state.visitCountForContainer(container);
        foundValue = IntValue(count);
      }

      // Normal variable reference
      else {
        foundValue = state.variablesState.getVariableWithName(varRef.name);

        if (foundValue == null) {
          warning("Variable not found: '${varRef.name}'."
              "Using default value of 0 (false). This can happen with temporary variables if the declaration hasn't yet been hit. Globals are always given a default value on load if a value doesn't exist in the save state.");
          foundValue = IntValue(0);
        }
      }

      state.pushEvaluationStack(foundValue);

      return true;
    }

    // Native function call
    else if (contentObj is NativeFunctionCall) {
      _logger.finest('Native function call');

      var func = contentObj;
      var funcParams = state.PopEvaluationStack(func.numberOfParameters);
      _logger.finest('func : $func');
      _logger.finest('funcParams : $funcParams');

      var result = func.call(funcParams);
      state.pushEvaluationStack(result);
      return true;
    }

    // No control content, must be ordinary content
    return false;
  }

// Note that this is O(n), since it re-evaluates the shuffle indices
  // from a consistent seed each time.
  // TODO: Is this the best algorithm it can be?
  // int NextSequenceShuffleIndex()
  int _nextSequenceShuffleIndex() {
    var numElementsIntVal = asOrNull<IntValue>(state.popEvaluationStack());
    if (numElementsIntVal == null) {
      error('expected number of elements in sequence for shuffle index');
      return 0;
    }

    var seqContainer = state.currentPointer.container;
    if (seqContainer == null) {
      throw ArgumentError.notNull('seqContainer');
    }

    var numElements = numElementsIntVal.value;

    var seqCountVal = state.popEvaluationStack() as IntValue;
    var seqCount = seqCountVal.value;
    var loopIndex = seqCount / numElements;
    var iterationIndex = seqCount % numElements;

    // Generate the same shuffle based on:
    //  - The hash of this container, to make sure it's consistent
    //    each time the runtime returns to the sequence
    //  - How many times the runtime has looped around this full shuffle
    var seqPathStr = seqContainer.path.toString();
    var sequenceHash = 0;

    for (var c in seqPathStr.codeUnits) {
      sequenceHash += c;
    }
    var randomSeed = sequenceHash + loopIndex + state.storySeed;
    var random = Random(randomSeed.toInt());

    var unpickedIndices = <int>[];
    for (var i = 0; i < numElements; ++i) {
      unpickedIndices.add(i);
    }

    for (var i = 0; i <= iterationIndex; ++i) {
      var chosen = random.nextInt(4294967296) % unpickedIndices.length;
      var chosenIndex = unpickedIndices[chosen];
      unpickedIndices.removeAt(chosen);

      if (i == iterationIndex) {
        return chosenIndex;
      }
    }

    throw Exception('Should never reach here');
  }

  OutputStateChange _calculateNewlineOutputStateChange(
      String? prevText, String? currText, int prevTagCount, int currTagCount) {
    // Simple case: nothing's changed, and we still have a newline
    // at the end of the current content
    _logger.finest(
        '[_calculateNewlineOutputStateChange] ( prevText : $prevText, currText : $currText, prevTagCount : $prevTagCount, currTagCount : $currTagCount  )');
    ArgumentError.checkNotNull(prevText);
    ArgumentError.checkNotNull(currText);

    var newlineStillExists = currText!.length >= prevText!.length &&
        prevText.isNotEmpty &&
        currText[prevText!.length - 1] == '\n';
    _logger.finest(
        '[_calculateNewlineOutputStateChange] newlineStillExists : $newlineStillExists');
    if (prevTagCount == currTagCount &&
        prevText.length == currText.length &&
        newlineStillExists) {
      return OutputStateChange.NoChange;
    }

    //
    if (!newlineStillExists) {
      _logger.finest(
          '[_calculateNewlineOutputStateChange] Old newline has been removed, '
          "it wasn't the end of the line after all");
      return OutputStateChange.NewlineRemoved;
    }

    // Tag added - definitely the start of a new line
    if (currTagCount > prevTagCount) {
      return OutputStateChange.ExtendedBeyondNewline;
    }

    // There must be new content - check whether it's just whitespace
    for (var i = prevText.length; i < currText.length; i++) {
      var c = currText[i];
      if (c != ' ' && c != '\t') {
        _logger.finest(
            "There must be new content - check whether it's just whitespace");
        return OutputStateChange.ExtendedBeyondNewline;
      }
    }

    // There's new text but it's just spaces and tabs, so there's still the
    // potential
    // for glue to kill the newline.
    return OutputStateChange.NoChange;
  }

  void _nextContent() {
    _logger.fine('[_nextContent]');
    // Setting previousContentObject is critical for VisitChangedContainersDueToDivert
    _logger.info('[_nextContent] previousPointer : ${state.previousPointer}');
    _logger.fine('_nextContent] currentPointer : ${state.currentPointer}');

    state.previousPointer = state.currentPointer.copyWith();

    // Divert step?
    if (state.divertedPointer.isNotNull) {
      _logger.fine('[_nextContent] divert step');
      state.currentPointer = state.divertedPointer.copyWith();
      state.divertedPointer = Pointer.Null;

      // Internally uses state.previousContentObject and state.currentContentObject
      _visitChangedContainersDueToDivert();

      // Diverted location has valid content?
      if (state.currentPointer.isNotNull) {
        _logger.finest('[_nextContent] currentPointer is not null');
        return;
      }

      // Otherwise, if diverted location doesn't have valid content,
      // drop down and attempt to increment.
      // This can happen if the diverted path is intentionally jumping
      // to the end of a container - e.g. a Conditional that's re-joining
    } else {
      _logger.fine('[_nextContent] not divert');
    }
    var successfulPointerIncrement = _incrementContentPointer();
    _logger.fine(
        '[_nextContent] successfulPointerIncrement : ${successfulPointerIncrement}');
    // Ran out of content? Try to auto-exit from a function,
    // or finish evaluating the content of a thread
    if (!successfulPointerIncrement) {
      _logger.finest('[_nextContent] Ran out of content');
      var didPop = false;

      if (state.callStack.canPop(PushPopType.Function)) {
        // Pop from the call stack
        state.popCallstack(PushPopType.Function);

        // This pop was due to dropping off the end of a function that didn't return anything,
        // so in this case, we make sure that the evaluator has something to chomp on if it needs it
        if (state.inExpressionEvaluation) {
          state.pushEvaluationStack(RuntimeVoid());
        }

        didPop = true;
      } else if (state.callStack.canPopThread) {
        state.callStack.popThread();

        didPop = true;
      } else {
        state.tryExitFunctionEvaluationFromGame();
      }

      // Step past the point where we last called out
      if (didPop && !state.currentPointer.isNull) {
        _nextContent();
      }
    }
  }

  //bool IncrementContentPointer()
  bool _incrementContentPointer() {
    _logger.fine('[_incrementContentPointer]');
    var successfulIncrement = true;

    var pointer = state.callStack.currentElement.currentPointer.copyWith();
    pointer.index++;

    // Each time we step off the end, we fall out to the next container, all the
    // while we're in indexed rather than named content
    while (pointer.index >= pointer.container!.content.length) {
      successfulIncrement = false;

      var nextAncestor = asOrNull<Container>(pointer.container?.parent);
      if (nextAncestor == null) {
        break;
      }

      var indexInAncestor = nextAncestor.content.indexOf(pointer.container!);
      if (indexInAncestor == -1) {
        break;
      }
      pointer = Pointer(container: nextAncestor, index: indexInAncestor);

      // Increment to next content in outer container
      pointer.index++;

      successfulIncrement = true;
    }

    if (!successfulIncrement) {
      pointer = Pointer.Null;
    }
    state.callStack.currentElement.currentPointer = pointer;

    return successfulIncrement;
  }

  /// <summary>
  /// Check whether more content is available if you were to call <c>Continue()</c> - i.e.
  /// are we mid story rather than at a choice point or at the end.
  /// </summary>
  /// <value><c>true</c> if it's possible to call <c>Continue()</c>.</value>
  @override
  bool get canContinue => state.canContinue;

  /// <summary>
  /// If ContinueAsync was called (with milliseconds limit > 0) then this property
  /// will return false if the ink evaluation isn't yet finished, and you need to call
  /// it again in order for the Continue to fully complete.
  /// </summary>
  bool get asyncContinueComplete {
    return !_asyncContinueActive;
  }

  /// Continue the story for one line of content, if possible.
  /// If you're not sure if there's more content available, for example if you
  /// want to check whether you're at a choice point or at the end of the story,
  /// you should call <c>canContinue</c> before calling this function.
  /// <returns>The line of text content.</returns>
  @override
  String? Continue() {
    _logger.info('Continue');
    continueAsync(0);
    return currentText;
  }

  /// <summary>
  /// An "asnychronous" version of Continue that only partially evaluates the ink,
  /// with a budget of a certain time limit. It will exit ink evaluation early if
  /// the evaluation isn't complete within the time limit, with the
  /// asyncContinueComplete property being false.
  /// This is useful if ink evaluation takes a long time, and you want to distribute
  /// it over multiple game frames for smoother animation.
  /// If you pass a limit of zero, then it will fully evaluate the ink in the same
  /// way as calling Continue (and in fact, this exactly what Continue does internally).
  /// </summary>
  void continueAsync(double millisecsLimitAsync) {
    _logger.fine('=== continueAsync ===');
    if (!_hasValidatedExternals) validateExternalBindings();

    _continueInternal(millisecsLimitAsync);
    _logger.fine('=== continueAsync (end)===');
  }

  /// <summary>
  /// Check that all EXTERNAL ink functions have a valid bound C# function.
  /// Note that this is automatically called on the first call to Continue().
  /// </summary>
  void validateExternalBindings() {
    var missingExternals = HashSet<String>();

    _validateExternalBindings(_mainContentContainer, missingExternals);
    _hasValidatedExternals = true;

    // No problem! Validation complete
    if (missingExternals.isEmpty) {
      _hasValidatedExternals = true;
    }

    // Error for all missing externals
    else {
      var message =
          "ERROR: Missing function binding for external${missingExternals.length > 1 ? "s" : ''}: '${missingExternals..toList().join(', ')}' ${allowExternalFunctionFallbacks ? ", and no fallback ink function found." : " (ink fallbacks disabled)"}";

      error(message);
    }
  }

  void _validateExternalBindings(
      Container c, HashSet<String> missingExternals) {
    for (var innerContent in c.content) {
      var container = asOrNull<Container>(innerContent);
      if (container == null || !container.hasValidName) {
        _validateExternalBindingsObject(innerContent, missingExternals);
      }
    }

    c.namedContent.forEach((key, value) {
      _validateExternalBindingsObject(value as RuntimeObject, missingExternals);
    });
  }

  void _validateExternalBindingsObject(
      RuntimeObject o, HashSet<String> missingExternals) {
    if (o is Container) {
      _validateExternalBindings(o, missingExternals);
      return;
    }

    if (o is Divert && o.isExternal) {
      var divert = o;
      var name = divert.targetPathString;

      if (!_externals.containsKey(name)) {
        if (allowExternalFunctionFallbacks) {
          var fallbackFound =
              mainContentContainer.namedContent.containsKey(name);
          if (!fallbackFound) {
            missingExternals.add(name!);
          }
        } else {
          missingExternals.add(name!);
        }
      }
    }
  }

  void _continueInternal([double millisecsLimitAsync = 0]) {
    _logger.fine('_continueInternal');
    _profiler?.preContinue();

    var isAsyncTimeLimited = millisecsLimitAsync > 0;

    _recursiveContinueCount++;
    _logger.finest('_recursiveContinueCount : ${_recursiveContinueCount}');
    // Doing either:
    // - full run through non-async (so not active and don't want to be)
    // - Starting async run-through
    if (!_asyncContinueActive) {
      _asyncContinueActive = isAsyncTimeLimited;
      if (!canContinue) {
        throw StoryException(
            "Can't continue - should check canContinue before calling Continue");
      }

      state.didSafeExit = false;
      state.resetOutput();

      // It's possible for ink to call game to call ink to call game etc
      // In this case, we only want to batch observe variable changes
      // for the outermost call.
      if (_recursiveContinueCount == 1) {
        _logger.finest(
            "It's possible for ink to call game to call ink to call game etc"
                'In this case, we only want to batch observe variable changes',
            'for the outermost call.');
        state.variablesState.batchObservingVariableChanges = true;
      }
    }

    /// Start timing
    var durationStopwatch = Stopwatch();
    durationStopwatch.start();

    var outputStreamEndsInNewline = false;
    var count = 0;
    do {
      try {
        _logger.fine('call ${count++}');
        outputStreamEndsInNewline = _continueSingleStep();
      } on StoryException catch (e) {
        addError(e.toString(), useEndLineNumber: e.useEndLineNumber);
        break;
      }
      _logger.finest('outputStreamEndsInNewline : $outputStreamEndsInNewline');
      if (outputStreamEndsInNewline) {
        break;
      }

      // Run out of async time?
      if (_asyncContinueActive &&
          durationStopwatch.elapsedMilliseconds > millisecsLimitAsync) {
        break;
      }
    } while (canContinue);

    _logger.fine('while loop end');

    durationStopwatch.stop();

    // 4 outcomes:
    //  - got newline (so finished this line of text)
    //  - can't continue (e.g. choices or ending)
    //  - ran out of time during evaluation
    //  - error
    //
    // Successfully finished evaluation in time (or in error)
    if (outputStreamEndsInNewline || !canContinue) {
      // Need to rewind, due to evaluating further than we should?
      if (_stateSnapshotAtLastNewline != null) {
        _logger.finest(
            'Need to rewind, due to evaluating further than we should?');
        _restoreStateSnapshot();
      }

      // Finished a section of content / reached a choice point?
      _logger.finest('Finished a section of content / reached a choice point?');
      if (!canContinue) {
        _logger.finest("Can't continue");

        if (state.callStack.canPopThread) {
          _logger.severe(
              'Thread available to pop, threads should always be flat by the end of evaluation?');
          addError(
              'Thread available to pop, threads should always be flat by the end of evaluation?');
        }

        if (state.generatedChoices.isEmpty &&
            !state.didSafeExit &&
            _temporaryEvaluationContainer == null) {
          if (state.callStack.canPop(PushPopType.Tunnel)) {
            _logger.severe(
                "unexpectedly reached end of content. Do you need a '->->' to return from a tunnel?");
            addError(
                "unexpectedly reached end of content. Do you need a '->->' to return from a tunnel?");
          } else if (state.callStack.canPop(PushPopType.Function)) {
            _logger.severe(
                "unexpectedly reached end of content. Do you need a '~ return'?");
            addError(
                "unexpectedly reached end of content. Do you need a '~ return'?");
          } else if (!state.callStack.canPop()) {
            _logger.severe(
                "ran out of content. Do you need a '-> DONE' or '-> END'?");
            addError(
                "ran out of content. Do you need a '-> DONE' or '-> END'?");
          } else {
            _logger.severe(
                'unexpectedly reached end of content for unknown reason. Please debug compiler!');
            addError(
                'unexpectedly reached end of content for unknown reason. Please debug compiler!');
          }
        }
      }

      state.didSafeExit = false;

      if (_recursiveContinueCount == 1) {
        _state.variablesState.batchObservingVariableChanges = false;
      }

      _asyncContinueActive = false;
    }

    _recursiveContinueCount--;

    _profiler?.postContinue();
  }

  bool _continueSingleStep() {
    _logger.fine('[_continueSingleStep]');
    _profiler?.PreStep();

    // Run main step function (walks through content)
    _logger.finest(
        '[_continueSingleStep] Run main step function (walks through content)');
    _step();

    _profiler?.PostStep();

    // Run out of content and we have a default invisible choice that we can follow?
    _logger.finest(
        '[_continueSingleStep] Run out of content and we have a default invisible choice that we can follow?');
    if (!canContinue && !state.callStack.elementIsEvaluateFromGame) {
      _logger.fine('[_continueSingleStep] _tryFollowDefaultInvisibleChoice');
      _tryFollowDefaultInvisibleChoice();
    }

    _profiler?.PreSnapshot();

    // Don't save/rewind during string evaluation, which is e.g. used for choices
    if (!state.inStringEvaluation) {
      _logger.fine(
          "[_continueSingleStep] Don't save/rewind during string evaluation, which is e.g. used for choices");
      // We previously found a newline, but were we just double checking that
      // it wouldn't immediately be removed by glue?
      if (_stateSnapshotAtLastNewline != null) {
        // Has proper text or a tag been added? Then we know that the newline
        // that was previously added is definitely the end of the line.
        var change = _calculateNewlineOutputStateChange(
            _stateSnapshotAtLastNewline!.currentText,
            state.currentText,
            _stateSnapshotAtLastNewline!.currentTags.length,
            state.currentTags.length);

        _logger.finest(
            '[_continueSingleStep] calculateNewlineOutputStateChange: $change');

        // The last time we saw a newline, it was definitely the end of the line, so we
        // want to rewind to that point.
        if (change == OutputStateChange.ExtendedBeyondNewline) {
          _logger.finest(
              '[_continueSingleStep] The last time we saw a newline, it was definitely the end of the line, so we'
              'want to rewind to that point.');

          _restoreStateSnapshot();

          // Hit a newline for sure, we're done
          _logger.finest(
              "[_continueSingleStep] Hit a newline for sure, we're done");
          return true;
        }

        // Newline that previously existed is no longer valid - e.g.
        // glue was encounted that caused it to be removed.
        else if (change == OutputStateChange.NewlineRemoved) {
          _logger.finest(
              '[_continueSingleStep] Newline that previously existed is no longer valid - e.g.'
              'glue was encounted that caused it to be removed.');
          _discardSnapshot();
        }
      }

      // Current content ends in a newline - approaching end of our evaluation
      if (state.outputStreamEndsInNewline) {
        _logger.finest(
            '[_continueSingleStep] Current content ends in a newline - approaching end of our evaluation');
        // If we can continue evaluation for a bit:
        // Create a snapshot in case we need to rewind.
        // We're going to continue stepping in case we see glue or some
        // non-text content such as choices.
        if (canContinue) {
          _logger.finest(
              '[_continueSingleStep] If we can continue evaluation for a bit:'
              'Create a snapshot in case we need to rewind.'
              "We're going to continue stepping in case we see glue or some"
              'non-text content such as choices.');
          // Don't bother to record the state beyond the current newline.
          // e.g.:
          // Hello world\n            // record state at the end of here
          // ~ complexCalculation()   // don't actually need this unless it generates text
          if (_stateSnapshotAtLastNewline == null) {
            _stateSnapshot();
          }
        }

        // Can't continue, so we're about to exit - make sure we
        // don't have an old state hanging around.
        else {
          _logger.finest(
              "[_continueSingleStep] Can't continue, so we're about to exit - make sure we"
              "don't have an old state hanging around.");
          _discardSnapshot();
        }
      }
    }

    _profiler?.postSnapshot();

    // outputStreamEndsInNewline = false
    return false;
  }

  /// <summary>
  /// Continue the story until the next choice point or until it runs out of content.
  /// This is as opposed to the Continue() method which only evaluates one line of
  /// output at a time.
  /// </summary>
  /// <returns>The resulting text evaluated by the ink engine, concatenated together.</returns>
  @override
  String continueMaximally() {
    ifAsyncWeCant('ContinueMaximally');

    var sb = StringBuffer();

    while (canContinue) {
      sb.write(Continue());
    }

    return sb.toString();
  }

  void _step() {
    _logger.fine('[_step]');
    var shouldAddToStream = true;

    // Get current content
    _logger.finest('[_step] Get current content');
    var pointer = state.currentPointer.copyWith();
    if (pointer.isNull) {
      _logger.finest('[_step] pointer.isNull');
      return;
    }
    _logger.finest('[_step] currentPointer is ${pointer}');

    // Step directly to the first element of content in a container (if necessary)
    _logger.finest(
        '[_step] Step directly to the first element of content in a container (if necessary)');
    var r = pointer.resolve();
    _logger.finest('r : $r');
    var containerToEnter = asOrNull<Container>(r);
    while (containerToEnter != null) {
      // Mark container as being entered
      _logger.finest('[_step] Mark container as being entered');
      _logger.finest('[_step] visitContainer : $containerToEnter');
      _visitContainer(containerToEnter, atStart: true);

      // No content? the most we can do is step past it
      _logger.finest('[_step] No content? ${containerToEnter.content.isEmpty}');
      if (containerToEnter.content.isEmpty) {
        _logger.finest('[_step] the most we can do is step past it');
        break;
      }

      pointer = Pointer.startOf(containerToEnter);

      r = pointer.resolve();
      containerToEnter = asOrNull<Container>(r);
    }

    state.currentPointer = pointer.copyWith();
    _logger.finest('[_step] currentPointer : $pointer');
    _profiler?.Step(state.callStack);

    // Is the current content object:
    //  - Normal content
    //  - Or a logic/flow statement - if so, do it
    // Stop flow if we hit a stack pop when we're unable to pop (e.g. return/done statement in knot
    // that was diverted to rather than called as a function)
    _logger.finest("""
    [_step] Is the current content object:
    - Normal content
    - Or a logic/flow statement - if so, do it
    Stop flow if we hit a stack pop when we're unable to pop (e.g. return/done statement in knot
    that was diverted to rather than called as a function)
    """);
    var currentContentObj = pointer.resolve();
    var isLogicOrFlowControl = _performLogicAndFlowControl(currentContentObj);

    _logger.fine('[_step] isLogicOrFlowControl : $isLogicOrFlowControl');

    // Has flow been forced to end by flow control above?
    _logger.fine(
        '[_step] Has flow been forced to end by flow control above? ${state.currentPointer}');
    if (state.currentPointer.isNull) {
      return;
    }

    if (isLogicOrFlowControl) {
      shouldAddToStream = false;
    }

    // Choice with condition?
    var choicePoint = asOrNull<ChoicePoint>(currentContentObj);
    if (choicePoint != null) {
      _logger.finest('[_step] Choice with condition');
      var choice = _processChoice(choicePoint);
      if (choice != null) {
        _logger.finest('[_step] add choice');
        state.generatedChoices.add(choice);
      }

      currentContentObj = null;
      shouldAddToStream = false;
    }

    // If the container has no content, then it will be
    // the "content" itself, but we skip over it.
    if (currentContentObj is Container) {
      _logger.finest('[_step] If the container has no content, then it will be'
          'the "content" itself, but we skip over it.');

      shouldAddToStream = false;
    }

    // Content to add to evaluation stack or the output stream
    if (shouldAddToStream) {
      _logger.finest('Content to add to evaluation stack or the output stream');
      // If we're pushing a variable pointer onto the evaluation stack, ensure that it's specific
      // to our current (possibly temporary) context index. And make a copy of the pointer
      // so that we're not editing the original runtime object.
      var varPointer =
          currentContentObj is VariablePointerValue ? currentContentObj : null;
      if (varPointer != null && varPointer.contextIndex == -1) {
        _logger.finest('[_step] variable pointer');
        // Create new object so we're not overwriting the story's own data
        var contextIdx =
            state.callStack.contextForVariableNamed(varPointer.variableName!);
        currentContentObj =
            VariablePointerValue(varPointer.variableName, contextIdx);
        _logger.finest('[_step] currentContentObj : $currentContentObj');
      }

      // Expression evaluation content
      if (state.inExpressionEvaluation) {
        _logger.info('[_step] Expression evaluation content');
        state.pushEvaluationStack(currentContentObj!);
      }
      // Output stream content (i.e. not expression evaluation)
      else {
        _logger.finest(
            '[_step] Output stream content (i.e. not expression evaluation) : ${currentContentObj.toString()}');
        state.PushToOutputStream(currentContentObj!);
      }
    }

    // Increment the content pointer, following diverts if necessary
    _logger.finest(
        '[_step] Output stream content (i.e. not expression evaluation) : ${currentContentObj.toString()}');
    _nextContent();

    // Starting a thread should be done after the increment to the content pointer,
    // so that when returning from the thread, it returns to the content after this instruction.
    if (currentContentObj is ControlCommand &&
        currentContentObj.commandType == CommandType.StartThread) {
      state.callStack.pushThread();
    }
  }

  bool _tryFollowDefaultInvisibleChoice() {
    _logger.fine('[_tryFollowDefaultInvisibleChoice]');
    var allChoices = _state.currentChoices;

    // Is a default invisible choice the ONLY choice?
    var invisibleChoices =
        allChoices.where((c) => c.isInvisibleDefault).toList();
    if (invisibleChoices.isEmpty ||
        allChoices.length > invisibleChoices.length) {
      return false;
    }

    var choice = invisibleChoices[0];

    checkNotNull(choice.targetPath);

    checkNotNull(choice.threadAtGeneration);

    // Invisible choice may have been generated on a different thread,
    // in which case we need to restore it before we continue
    state.callStack.currentThread = choice.threadAtGeneration;

    // If there's a chance that this state will be rolled back to before
    // the invisible choice then make sure that the choice thread is
    // left intact, and it isn't re-entered in an old state.
    if (_stateSnapshotAtLastNewline != null) {
      state.callStack.currentThread = state.callStack.forkThread();
    }

    choosePath(choice.targetPath!, incrementingTurnIndex: false);

    return true;
  }

  void addError(String message,
      {bool isWarning = false, bool useEndLineNumber = false}) {
    _logger.finest('[addError] : $message');
    var dm = _currentDebugMetadata;

    var errorTypeStr = isWarning ? "WARNING" : "ERROR";

    if (dm != null) {
      var lineNum = useEndLineNumber ? dm.endLineNumber : dm.startLineNumber;
      message =
          "RUNTIME $errorTypeStr: '${dm.fileName}' line ${lineNum}: $message";
    } else if (!state.currentPointer.isNull) {
      message =
          'RUNTIME $errorTypeStr: (${state.currentPointer.path}): $message';
    } else {
      message = 'RUNTIME ' + errorTypeStr + ': ' + message;
    }

    state.addError(message, isWarning);

    // In a broken state don't need to know about any other errors.
    if (!isWarning) state.ForceEnd();
  }

  DebugMetadata? get _currentDebugMetadata {
    DebugMetadata? dm;

    // Try to get from the current path first
    var pointer = state.currentPointer.copyWith();
    if (!pointer.isNull) {
      dm = pointer.resolve()?.debugMetadata;
      if (dm != null) {
        return dm;
      }
    }

    // Move up callstack if possible
    for (var i = state.callStack.elements.length - 1; i >= 0; --i) {
      pointer = state.callStack.elements[i].currentPointer.copyWith();
      if (!pointer.isNull && pointer.resolve() != null) {
        dm = pointer.resolve()?.debugMetadata;
        if (dm != null) {
          return dm;
        }
      }
    }

    // Current/previous path may not be valid if we've just had an error,
    // or if we've simply run out of content.
    // As a last resort, try to grab something from the output stream
    for (var i = state.outputStream.length - 1; i >= 0; --i) {
      var outputObj = state.outputStream[i];
      dm = outputObj.debugMetadata;
      if (dm != null) {
        return dm;
      }
    }

    return null;
  }

  int get _currentLineNumber {
    var dm = _currentDebugMetadata;
    if (dm != null) {
      return dm.startLineNumber;
    }
    return 0;
  }

  /// <summary>
  /// Evaluates a function defined in ink.
  /// </summary>
  /// <returns>The return value as returned from the ink function with `~ return myValue`, or null if nothing is returned.</returns>
  /// <param name="functionName">The name of the function as declared in ink.</param>
  /// <param name="arguments">The arguments that the ink function takes, if any. Note that we don't (can't) do any validation on the number of arguments right now, so make sure you get it right!</param>
  Object evaluateFunction(String functionName,
      [List<Object> arguments = const []]) {
    return evaluateFunctionString(functionName, null, arguments);
  }

  /// <summary>
  /// Evaluates a function defined in ink, and gathers the possibly multi-line text as generated by the function.
  /// This text output is any text written as normal content within the function, as opposed to the return value, as returned with `~ return`.
  /// </summary>
  /// <returns>The return value as returned from the ink function with `~ return myValue`, or null if nothing is returned.</returns>
  /// <param name="functionName">The name of the function as declared in ink.</param>
  /// <param name="textOutput">The text content produced by the function via normal ink, if any.</param>
  /// <param name="arguments">The arguments that the ink function takes, if any. Note that we don't (can't) do any validation on the number of arguments right now, so make sure you get it right!</param>
  Object evaluateFunctionString(String functionName, StringBuffer? textOutput,
      [List<Object> arguments = const []]) {
    _logger.finest('[evaluateFunctionString] $functionName');
    if (onEvaluateFunction != null) {
      onEvaluateFunction?.call(functionName, arguments);
    }
    ifAsyncWeCant('evaluate a function');

    if (isBlank(functionName)) {
      throw Exception('Function is empty or white space.');
    }

    // Get the content that we need to run
    var funcContainer = knotContainerWithName(functionName);
    if (funcContainer == null) {
      throw Exception("Function doesn't exist: '" + functionName + "'");
    }

    // Snapshot the output stream
    var outputStreamBefore = List<RuntimeObject>.of(state.outputStream);
    _state.resetOutput();

    // State will temporarily replace the callstack in order to evaluate
    state.startFunctionEvaluationFromGame(funcContainer, arguments);

    // Evaluate the function, and collect the string output
    while (canContinue) {
      textOutput?.write(Continue());
    }

    // Restore the output stream in case this was called
    // during main story evaluation.
    _state.resetOutput(outputStreamBefore);

    // Finish evaluation, and see whether anything was produced
    var result = state.completeFunctionEvaluationFromGame();
    if (onCompleteEvaluateFunction != null) {
      onCompleteEvaluateFunction?.call(
          functionName, arguments, textOutput.toString(), result);
    }
    return result;
  }

  // Evaluate a "hot compiled" piece of ink content, as used by the REPL-like
  // CommandLinePlayer.
  RuntimeObject? EvaluateExpression(Container exprContainer) {
    var startCallStackHeight = state.callStack.elements.length;

    state.callStack.push(PushPopType.Tunnel);

    _temporaryEvaluationContainer = exprContainer;

    state.goToStart();

    var evalStackHeight = state.evaluationStack.length;

    Continue();

    _temporaryEvaluationContainer = null;

    // Should have fallen off the end of the Container, which should
    // have auto-popped, but just in case we didn't for some reason,
    // manually pop to restore the state (including currentPath).
    if (state.callStack.elements.length > startCallStackHeight) {
      state.popCallstack();
    }

    var endStackHeight = state.evaluationStack.length;
    if (endStackHeight > evalStackHeight) {
      return state.popEvaluationStack();
    } else {
      return null;
    }
  }

  /// <summary>
  /// Most general form of function binding that returns an object
  /// and takes an array of object parameters.
  /// The only way to bind a function with more than 3 arguments.
  /// </summary>
  /// <param name="funcName">EXTERNAL ink function name to bind to.</param>
  /// <param name="func">The C# function to bind.</param>
  /// <param name="lookaheadSafe">The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass 'true'.
  /// Usually, you want to pass 'false', especially if you want some action
  /// to be performed in game code when this function is called.</param>
  void bindExternalFunctionGeneral(String funcName, ExternalFunction func,
      {bool lookaheadSafe = true}) {
    ifAsyncWeCant('bind an external function');
    assert(!_externals.containsKey(funcName),
        "Function '" + funcName + "' has already been bound.");
    _externals[funcName] = ExternalFunctionDef()..function = func;
  }

  dynamic tryCoerce<T>(dynamic value) {
    if (value == null) {
      return null;
    }

    if (value is T) {
      return value as T;
    }

    if (value is double && T == int) {
      var intVal = value.round();
      return intVal as T;
    }

    if (value is int && T == double) {
      var floatVal = value.toDouble();
      return floatVal as T;
    }

    if (value is int && T == bool) {
      var intVal = value;
      return intVal == 0 ? false : true as T;
    }

    if (value is bool && T == int) {
      bool boolVal = value;
      return boolVal ? 1 : 0;
    }

    if (T == String) {
      return value.toString() as T;
    }

    assert(
        false,
        'Failed to cast ' +
            value.runtimeType.toString() +
            ' to ' +
            value.toString());

    return null;
  }

  /// <summary>
  /// Remove a binding for a named EXTERNAL ink function.
  /// </summary>
  void unbindExternalFunction(String funcName) {
    ifAsyncWeCant('unbind an external a function');
    assert(_externals.containsKey(funcName),
        "Function '" + funcName + "' has not been bound.");
    _externals.remove(funcName);
  }

  /// <summary>
  /// When the named global variable changes it's value, the observer will be
  /// called to notify it of the change. Note that if the value changes multiple
  /// times within the ink, the observer will only be called once, at the end
  /// of the ink's evaluation. If, during the evaluation, it changes and then
  /// changes back again to its original value, it will still be called.
  /// Note that the observer will also be fired if the value of the variable
  /// is changed externally to the ink, by directly setting a value in
  /// story.variablesState.
  /// </summary>
  /// <param name="variableName">The name of the global variable to observe.</param>
  /// <param name="observer">A delegate function to call when the variable changes.</param>
  void observeVariable(String variableName, VariableObserver observer) {
    ifAsyncWeCant('observe a new variable');
    _variableObservers ??= ListMultimap<String, VariableObserver>();

    if (!state.variablesState.globalVariableExistsWithName(variableName)) {
      throw Exception("Cannot observe variable '" +
          variableName +
          "' because it wasn't declared in the ink story.");
    }
    _variableObservers!.add(variableName, observer);
  }

  @override
  void ChooseChoiceIndex(int choiceIdx) => chooseChoiceIndex(choiceIdx);

  @override
  void ChoosePathString(String path,
      {bool resetCallstack = true, List<Object> arguments = const []}) {
    onChoosePathString?.call(path, arguments);

    if (resetCallstack) {
      ResetCallstack();
    } else {
      // ChoosePathString is potentially dangerous since you can call it when the stack is
      // pretty much in any state. Let's catch one of the worst offenders.
      if (state.callStack.currentElement.type == PushPopType.Function) {
        String funcDetail = "";
        var container = state.callStack.currentElement.currentPointer.container;
        if (container != null) {
          funcDetail = "(" + container.path.toString() + ") ";
        }
        throw Exception("Story was running a function " +
            funcDetail +
            "when you called ChoosePathString(" +
            path +
            ") - this is almost certainly not not what you want! Full stack trace: \n" +
            state.callStack.callStackTrace);
      }
    }
    state.PassArgumentsToEvaluationStack(arguments);

    choosePath(Path(path));
  }

  @override
  String ContinueMaximally() => continueMaximally();

  @override
  List<String> TagsForContentAtPath(String path) =>
      TagsAtStartOfFlowContainerWithPathString(path);

  @override
  List<String> get globalTags => TagsAtStartOfFlowContainerWithPathString("");

  @internal
  List<String> TagsAtStartOfFlowContainerWithPathString(String pathString) {
    var path = Path(pathString);

    // Expected to be global story, knot or stitch
    Container flowContainer = contentAtPath(path).container!;
    while (true) {
      var firstContent = flowContainer.content[0];
      if (firstContent is Container) {
        flowContainer = firstContent;
      } else {
        break;
      }
    }

    // Any initial tag objects count as the "main tags" associated with that story/knot/stitch
    bool inTag = false;
    List<String> tags = [];
    for (var c in flowContainer.content) {
      var command = asOrNull<ControlCommand>(c);
      if (command != null) {
        if (command.commandType == CommandType.BeginTag) {
          inTag = true;
        } else if (command.commandType == CommandType.EndTag) {
          inTag = false;
        }
      } else if (inTag) {
        var str = asOrNull<StringValue>(c);
        if (str != null) {
          tags.add(str.value);
        } else {
          Exception(
              'Tag contained non-text content. Only plain text is allowed when using globalTags or TagsAtContentPath. If you want to evaluate dynamic content, you need to use story.Continue().');
        }
      }
      // Legacy Tag ( ink version 20 )
      else if (c is Tag) {
        tags.add(c.text);
      }

      // Any other content - we're done
      // We only recognise initial text-only tags
      else {
        break;
      }
    }

    return tags;
  }

  @override
  void ResetCallstack() {
    ifAsyncWeCant("ResetCallstack");
    _state.ForceEnd();
  }

  @override
  EvaluateFunctionTextOutput EvaluateFunction(String functionName,
      [List<Object> arguments = const []]) {
    onEvaluateFunction?.call(functionName, arguments);

    ifAsyncWeCant('evaluate a function');

    checkArgument(functionName.isNotEmpty);

    // Get the content that we need to run
    var funcContainer = KnotContainerWithName(functionName);
    if (funcContainer == null) {
      throw Exception("Function doesn't exist: '" + functionName + "'");
    }

    // Snapshot the output stream
    var outputStreamBefore = List<RuntimeObject>.of(state.outputStream);
    _state.ResetOutput();

    // State will temporarily replace the callstack in order to evaluate
    state.StartFunctionEvaluationFromGame(funcContainer, arguments);

    // Evaluate the function, and collect the string output
    var stringOutput = StringBuffer();
    while (canContinue) {
      stringOutput.write(Continue());
    }
    var textOutput = stringOutput.toString();

    // Restore the output stream in case this was called
    // during main story evaluation.
    _state.ResetOutput(outputStreamBefore);

    // Finish evaluation, and see whether anything was produced
    var result = state.CompleteFunctionEvaluationFromGame();
    onCompleteEvaluateFunction?.call(
        functionName, arguments, textOutput, result);
    return EvaluateFunctionTextOutput(result, textOutput);
  }

  @override
  void IfAsyncWeCant(String activityStr) => ifAsyncWeCant(activityStr);

  @override
  TryCoerce<T>(Object value) => tryCoerce(value);

  @override
  void UnbindExternalFunction(String funcName) =>
      unbindExternalFunction(funcName);

  @override
  void ValidateExternalBindings() => validateExternalBindings();

  @override
  void BindExternalFunctionGeneral(String funcName, ExternalFunction func,
          {bool lookaheadSafe = true}) =>
      bindExternalFunctionGeneral(funcName, func, lookaheadSafe: lookaheadSafe);

  @override
  Container? KnotContainerWithName(String name) => knotContainerWithName(name);

  void switchFlow(String flowName) {
    ifAsyncWeCant("switch flow");
    if (_asyncSaving) {
      throw Exception(
          "Story is already in background saving mode, can't switch flow to $flowName");
    }
    _state.SwitchFlow_Internal(flowName);
  }

  @override
  List<String> get aliveFlowNames => _state.aliveFlowNames;

  @override
  bool get currentFlowIsDefaultFlow => _state.currentFlowIsDefaultFlow;

  @override
  String get currentFlowName => _state.currentFlowName;

  @override
  void SwitchFlow(String flowName) => switchFlow(flowName);

  void choosePathString(String path,
          {bool resetCallstack = true, List<Object> arguments = const []}) =>
      ChoosePathString(path,
          resetCallstack: resetCallstack, arguments: arguments);

  /// Bind a dart function to an ink EXTERNAL function declaration.
  ///
  /// [funcName] EXTERNAL ink function name to bind to
  /// [func] The dart function to bind.
  /// [lookaheadSafe] The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass `true`.
  /// Usually, you want to pass `false`, especially if you want some action
  /// to be performed in game code when this function is called.
  void BindExternalFunction(String funcName, Object? Function() func,
      {bool lookaheadSafe = false}) {
    BindExternalFunctionGeneral(funcName, (List<Object> args) {
      assert(args.isEmpty, 'External function expected no arguments');
      return func.call();
    }, lookaheadSafe: lookaheadSafe);
  }

  /// Bind a dart function to an ink EXTERNAL function declaration.
  ///
  /// [funcName] EXTERNAL ink function name to bind to
  /// [func] The dart function to bind.
  /// [lookaheadSafe] The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass `true`.
  /// Usually, you want to pass `false`, especially if you want some action
  /// to be performed in game code when this function is called.
  void BindExternalFunction1<T1>(String funcName, Object? Function(T1 t1) func,
      {bool lookaheadSafe = false}) {
    BindExternalFunctionGeneral(funcName, (List<Object> args) {
      assert(args.length == 1, 'External function expected one argument');
      return func.call(TryCoerce<T1>(args[0]));
    }, lookaheadSafe: lookaheadSafe);
  }

  /// Bind a dart function to an ink EXTERNAL function declaration.
  ///
  /// [funcName] EXTERNAL ink function name to bind to
  /// [func] The dart function to bind.
  /// [lookaheadSafe] The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass `true`.
  /// Usually, you want to pass `false`, especially if you want some action
  /// to be performed in game code when this function is called.
  void BindExternalFunction2<T1, T2>(
      String funcName, Object? Function(T1 t1, T2 t2) func,
      {bool lookaheadSafe = false}) {
    BindExternalFunctionGeneral(funcName, (List<Object> args) {
      assert(args.length == 2, 'External function expected two arguments');
      return func.call(TryCoerce<T1>(args[0]), TryCoerce<T2>(args[1]));
    }, lookaheadSafe: lookaheadSafe);
  }

  /// Bind a dart function to an ink EXTERNAL function declaration.
  ///
  /// [funcName] EXTERNAL ink function name to bind to
  /// [func] The dart function to bind.
  /// [lookaheadSafe] The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass `true`.
  /// Usually, you want to pass `false`, especially if you want some action
  /// to be performed in game code when this function is called.
  void BindExternalFunction3<T1, T2, T3>(
      String funcName, Object? Function(T1 t1, T2 t2, T3 t3) func,
      {bool lookaheadSafe = false}) {
    BindExternalFunctionGeneral(funcName, (List<Object> args) {
      assert(args.length == 3, 'External function expected three arguments');
      return func.call(TryCoerce<T1>(args[0]), TryCoerce<T2>(args[1]),
          TryCoerce<T3>(args[2]));
    }, lookaheadSafe: lookaheadSafe);
  }

  /// Bind a dart function to an ink EXTERNAL function declaration.
  ///
  /// [funcName] EXTERNAL ink function name to bind to
  /// [func] The dart function to bind.
  /// [lookaheadSafe] The ink engine often evaluates further
  /// than you might expect beyond the current line just in case it sees
  /// glue that will cause the two lines to become one. In this case it's
  /// possible that a function can appear to be called twice instead of
  /// just once, and earlier than you expect. If it's safe for your
  /// function to be called in this way (since the result and side effect
  /// of the function will not change), then you can pass `true`.
  /// Usually, you want to pass `false`, especially if you want some action
  /// to be performed in game code when this function is called.
  void BindExternalFunction4<T1, T2, T3, T4>(
      String funcName, Object Function(T1 t1, T2 t2, T3 t3, T4 t4) func,
      {bool lookaheadSafe = false}) {
    BindExternalFunctionGeneral(funcName, (List<Object> args) {
      assert(args.length == 4, 'External function expected four arguments');
      return func.call(TryCoerce<T1>(args[0]), TryCoerce<T2>(args[1]),
          TryCoerce<T3>(args[2]), TryCoerce<T4>(args[3]));
    }, lookaheadSafe: lookaheadSafe);
  }
}

class EvaluateFunctionTextOutput {
  final dynamic returned;
  final String output;

  const EvaluateFunctionTextOutput(this.returned, this.output);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is EvaluateFunctionTextOutput &&
          runtimeType == other.runtimeType &&
          returned == other.returned &&
          output == other.output;

  @override
  int get hashCode => returned.hashCode ^ output.hashCode;
}
