import 'dart:math';

import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:quiver/check.dart';

import 'call_stack.dart';
import 'choice.dart';
import 'container.dart';
import 'control_command.dart';
import 'extension.dart';
import 'flow.dart';
import 'glue.dart';
import 'ink_list.dart';
import 'json.dart';
import 'list_definition.dart';
import 'logger.dart';
import 'path.dart';
import 'pointer.dart';
import 'push_pop.dart';
import 'runtime_object.dart';
import 'runtime_void.dart';
import 'simple_json.dart';
import 'state_patch.dart';
import 'story.dart';
import 'story_exception.dart';
import 'tag.dart';
import 'value.dart';
import 'variable_state.dart';

/// All story state information is included in the StoryState class,
/// including global variables, read counts, the pointer to the current
/// point in the story, the call stack (for tunnels, functions, etc),
/// and a few other smaller bits and pieces. You can save the current
/// state using the json serialisation functions ToJson and LoadJson.
class StoryState {
  //
  // Backward compatible changes since v8:
  // v10: dynamic tags
  // v9:  multi-flows
  /// The current version of the state save file JSON-based format.
  static const int kInkSaveStateVersion =
      10; // new: multi-flows, but backward compatible
  static const int kMinCompatibleLoadVersion = 8;

  /// Callback for when a state is loaded
  Function? onDidLoadState;

  final _logger = Logger('StoryState');

  String get currentFlowName => _currentFlow.name;

  bool get currentFlowIsDefaultFlow => _currentFlow.name == kDefaultFlowName;

  /// Gets the visit/read count of a particular Container at the given path.
  /// For a knot or stitch, that path string will be in the form:
  ///
  ///     knot
  ///     knot.stitch
  ///
  /// returns The number of times the specific knot or stitch has
  /// been enountered by the ink engine.
  /// [pathString] The dot-separated path string of
  /// the specific knot or stitch.
  // public int VisitCountAtPathString(string pathString)
  int visitCountAtPathString(String pathString) {
    int? visitCountOut;
    if (_patch != null) {
      var container = story.contentAtPath(Path(pathString)).container;
      if (container == null) {
        throw Exception('Content at path not found: ' + pathString);
      }

      visitCountOut = _patch!.getVisitCount(container);
      if (visitCountOut != null) {
        return visitCountOut;
      }
    }
    return _visitCounts[pathString] ?? 0;
  }

  int visitCountForContainer(Container? container) {
    _logger.finest('visitCountForContainer');
    if (container == null) {
      throw ArgumentError.notNull('container');
    }
    if (!container.visitsShouldBeCounted) {
      story.error('Read count for target (' +
          container.name +
          ' - on ' +
          container.debugMetadata.toString() +
          ') unknown.');
      return 0;
    }

    int? count = 0;
    if (_patch != null) {
      _logger.finest('_patch is not null');
      count = _patch!.getVisitCount(container);
      if (count != null) {
        _logger.finest('count is not null, $count');
        return count;
      }
    }
    _logger.finest('container.path is ${container.path}');

    var containerPathStr = container.path.toString();
    _logger.finest('containerPathStr is $containerPathStr');

    count = _visitCounts[containerPathStr] ?? 0;
    _logger.finest('count is $count');
    return count;
  }

  /// https://github.com/inkle/ink/blob/01898140be43d29baac70c1cb6544bdb10164209/ink-engine-runtime/StoryState.cs#L101
  /// internal void IncrementVisitCountForContainer(Container container)
  void incrementVisitCountForContainer(Container container) {
    _logger.finest('IncrementVisitCountForContainer');
    if (_patch != null) {
      _logger.finest('patch is not null');
      var currCount = visitCountForContainer(container);
      currCount++;
      _logger.finest('currCount : $currCount');

      _patch!.setVisitCount(container, currCount);
      return;
    }

    var count = 0;
    var containerPathStr = container.path.toString();
    _logger.fine('containerPathStr : $containerPathStr');
    if (_visitCounts.containsKey(containerPathStr)) {
      count = _visitCounts[containerPathStr]!;
    }
    count++;
    _visitCounts[containerPathStr] = count;
    _logger.fine('_visitCounts : $count');
  }

  void recordTurnIndexVisitToContainer(Container container) {
    _logger.finest('RecordTurnIndexVisitToContainer');
    if (_patch != null) {
      _logger.finest('_patch.setTurnIndex : $currentTurnIndex');
      _patch!.setTurnIndex(container, currentTurnIndex);
      return;
    }

    var containerPathStr = container.path.toString();
    _logger.finest('containerPathStr : $containerPathStr');

    _turnIndices[containerPathStr] = currentTurnIndex;
  }

  int turnsSinceForContainer(Container container) {
    if (!container.turnIndexShouldBeCounted) {
      story.error('TURNS_SINCE() for target (' +
          container.name +
          ' - on ' +
          container.debugMetadata.toString() +
          ') unknown.');
    }

    var _i = _patch?.getTurnIndex(container);

    if (_i != null) {
      return currentTurnIndex - _i;
    }

    var containerPathStr = container.path.toString();
    if (_turnIndices.containsKey(containerPathStr)) {
      return currentTurnIndex - _turnIndices[containerPathStr]!;
    } else {
      return -1;
    }
  }

  int get callstackDepth => callStack.depth;

  // REMEMBER! REMEMBER! REMEMBER!
  // When adding state, update the Copy method and serialisation
  // REMEMBER! REMEMBER! REMEMBER!

  List<RuntimeObject> get outputStream => _currentFlow.outputStream;

  List<Choice> get currentChoices {
    // If we can continue generating text content rather than choices,
    // then we reflect the choice list as being empty, since choices
    // should always come at the end.
    if (canContinue) return <Choice>[];
    return _currentFlow.currentChoices;
  }

  List<Choice> get generatedChoices => _currentFlow.currentChoices;

  List<String>? _currentErrors;

  List<String>? get currentErrors => _currentErrors;

  List<String>? _currentWarnings;

  List<String>? get currentWarnings => _currentWarnings;

  late VariablesState _variablesState =
      VariablesState(callStack, story.listDefinitions);

  VariablesState get variablesState => _variablesState;

  CallStack get callStack => _currentFlow.callStack;
  late List<RuntimeObject> _evaluationStack;

  @override
  List<RuntimeObject> get evaluationStack => _evaluationStack;
  Pointer divertedPointer = Pointer.Null;

  int _currentTurnIndex = 0;

  int get currentTurnIndex => _currentTurnIndex;
  int storySeed = 0;

  int previousRandom = 0;
  bool didSafeExit = false;

  Story story;

  /// String representation of the location where the story currently is.
  String? get currentPathString {
    var pointer = currentPointer;
    if (pointer.isNull) {
      return null;
    }

    return pointer.path.toString();
  }

  Pointer get currentPointer => callStack.currentElement.currentPointer;

  set currentPointer(Pointer value) =>
      callStack.currentElement.currentPointer = value;

  Pointer get previousPointer =>
      callStack.currentThread.previousPointer.copyWith();

  set previousPointer(Pointer value) =>
      callStack.currentThread.previousPointer = value;

  bool get canContinue => !currentPointer.isNull && !hasError;

  bool get hasError => currentErrors != null && currentErrors!.isNotEmpty;

  bool get hasWarning => currentWarnings != null && currentWarnings!.isNotEmpty;

  String? get currentText {
    if (_outputStreamTextDirty) {
      _logger.finest('_outputStreamTextDirty');

      var sb = StringBuffer();
      bool inTag = false;
      for (var outputObj in outputStream) {
        var textContent = asOrNull<StringValue>(outputObj);
        if (!inTag && textContent != null) {
          sb.write(textContent.value);
        } else {
          var controlCommand = asOrNull<ControlCommand>(outputObj);
          if (controlCommand != null) {
            if (controlCommand.commandType == CommandType.BeginTag) {
              inTag = true;
            } else if (controlCommand.commandType == CommandType.EndTag) {
              inTag = false;
            }
          }
        }
      }

      _currentText = cleanOutputWhitespace(sb.toString());

      _outputStreamTextDirty = false;
    }
    _logger.finest('Current Text : \n $_currentText');
    return _currentText;
  }

  String? _currentText;

// Cleans inline whitespace in the following way:
//  - Removes all whitespace from the start and end of line (including just before a \n)
//  - Turns all consecutive space and tab runs into single spaces (HTML style)
  String cleanOutputWhitespace(String str) {
    var sb = StringBuffer();

    var currentWhitespaceStart = -1;
    var startOfLine = 0;

    for (var i = 0; i < str.length; i++) {
      var c = str[i];

      var isInlineWhitespace = c == ' ' || c == '\t';

      if (isInlineWhitespace && currentWhitespaceStart == -1) {
        currentWhitespaceStart = i;
      }

      if (!isInlineWhitespace) {
        if (c != '\n' &&
            currentWhitespaceStart > 0 &&
            currentWhitespaceStart != startOfLine) {
          sb.write(' ');
        }
        currentWhitespaceStart = -1;
      }

      if (c == '\n') {
        startOfLine = i + 1;
        _logger.finest('new line, startOfLine : $startOfLine');
      }

      if (!isInlineWhitespace) {
        sb.write(c);
      }
    }

    return sb.toString();
  }

  List<String> get currentTags {
    if (_outputStreamTagsDirty) {
      _currentTags = <String>[];
      bool inTag = false;
      final sb = StringBuffer();
      for (var outputObj in outputStream) {
        var controlCommand = asOrNull<ControlCommand>(outputObj);

        if (controlCommand != null) {
          if (controlCommand.commandType == CommandType.BeginTag) {
            if (inTag && sb.length > 0) {
              var txt = cleanOutputWhitespace(sb.toString());
              _currentTags.add(txt);
              sb.clear();
            }
            inTag = true;
          } else if (controlCommand.commandType == CommandType.EndTag) {
            if (sb.length > 0) {
              var txt = cleanOutputWhitespace(sb.toString());
              _currentTags.add(txt);
              sb.clear();
            }
            inTag = false;
          }
        } else if (inTag) {
          var strVal = asOrNull<StringValue>(outputObj);
          if (strVal != null) {
            sb.write(strVal.value);
          }
        } else {
          var tag = asOrNull<Tag>(outputObj);
          if (tag != null &&
              tag.text != null &&
              tag.text.isNotEmpty) {
            _currentTags.add(tag.text); // tag.text has whitespae already cleaned
          }
        }
      }
      if (sb.length > 0) {
        var txt = cleanOutputWhitespace(sb.toString());
        _currentTags.add(txt);
        sb.clear();
      }
      _outputStreamTagsDirty = false;
    }
    return _currentTags;
  }

  late List<String> _currentTags;

  bool get inExpressionEvaluation =>
      callStack.currentElement.inExpressionEvaluation;

  set inExpressionEvaluation(bool value) {
    callStack.currentElement.inExpressionEvaluation = value;
  }

  StoryState(this.story)
      : _currentFlow = Flow(kDefaultFlowName, story),
        _evaluationStack = <RuntimeObject>[],
        _visitCounts = <String, int>{},
        _turnIndices = <String, int>{},
        _currentTurnIndex = -1,
        previousRandom = 0 {
    outputStreamDirty();
    _variablesState = VariablesState(callStack, story.listDefinitions);

    // Seed the shuffle random numbers
    var timeSeed = DateTime.now().millisecond;

    storySeed = (Random(timeSeed)).nextInt(100);

    goToStart();
  }

  void goToStart() {
    callStack.currentElement.currentPointer =
        Pointer.startOf(story.mainContentContainer);
  }

  @internal
  void SwitchFlow_Internal(String? flowName) {
    if (flowName == null) {
      throw Exception('Must pass a non-null string to Story.SwitchFlow');
    }
    if (_namedFlows == null) {
      _namedFlows = <String, Flow>{};
      _namedFlows![kDefaultFlowName] = _currentFlow;
    }
    if (flowName == _currentFlow.name) {
      return;
    }
    Flow? flow = _namedFlows![flowName];
    if (flow == null) {
      flow = Flow(flowName, story);
      _namedFlows![flowName] = flow;
      _aliveFlowNamesDirty = true;
    }

    _currentFlow = flow;
    variablesState.callStack = _currentFlow.callStack;
    // Cause text to be regenerated from output stream if necessary
    outputStreamDirty();
  }

  @internal
  void SwitchToDefaultFlow_Internal() {
    if (_namedFlows == null) return;
    SwitchFlow_Internal(kDefaultFlowName);
  }

  @internal
  void RemoveFlow_Internal(String? flowName) {
    if (flowName == null) {
      throw Exception('Must pass a non-null string to Story.DestroyFlow');
    }
    if (flowName == kDefaultFlowName) {
      throw Exception('Cannot destroy default flow');
    }
    // If we're currently in the flow that's being removed, switch back to default
    if (_currentFlow.name == flowName) {
      SwitchToDefaultFlow_Internal();
    }
    _namedFlows?.remove(flowName);
    _aliveFlowNamesDirty = true;
  }

  /// Warning: Any Runtime.Object content referenced within the StoryState will
  /// be re-referenced rather than cloned. This is generally okay though since
  /// Runtime.Objects are treated as immutable after they've been set up.
  /// (e.g. we don't edit a Runtime.StringValue after it's been created an added.)
  /// I wonder if there's a sensible way to enforce that..??
  StoryState copyAndStartPatching() {
    var copy = StoryState(story);

    copy._patch = StatePatch(_patch);

    // Hijack the new default flow to become a copy of our current one
    // If the patch is applied, then this new flow will replace the old one in _namedFlows
    copy._currentFlow.name = _currentFlow.name;
    copy._currentFlow.callStack = CallStack.copy(_currentFlow.callStack);
    copy._currentFlow.currentChoices.addAll(_currentFlow.currentChoices);
    copy._currentFlow.outputStream.addAll(_currentFlow.outputStream);
    copy.outputStreamDirty();

    // The copy of the state has its own copy of the named flows dictionary,
    // except with the current flow replaced with the copy above
    // (Assuming we're in multi-flow mode at all. If we're not then
    // the above copy is simply the default flow copy and we're done)
    if (_namedFlows != null) {
      copy._namedFlows = <String, Flow>{};
      _namedFlows!.forEach((key, value) {
        copy._namedFlows![key] = value;
      });
      copy._namedFlows![_currentFlow.name] = copy._currentFlow;
      copy._aliveFlowNamesDirty = true;
    }

    if (hasError) {
      copy._currentErrors = List.of(currentErrors!);
    }
    if (hasWarning) {
      copy._currentWarnings = List.of(currentWarnings!);
    }

    // ref copy - exactly the same variables state!
    // we're expecting not to read it only while in patch mode
    // (though the callstack will be modified)
    copy._variablesState = variablesState;
    copy._variablesState.callStack = copy.callStack;
    copy._variablesState.patch = copy._patch;

    copy.evaluationStack.addAll(evaluationStack);

    if (!divertedPointer.isNull) {
      copy.divertedPointer = divertedPointer;
    }

    copy.previousPointer = previousPointer;

    // visit counts and turn indicies will be read only, not modified
    // while in patch mode
    copy._visitCounts = _visitCounts;
    copy._turnIndices = _turnIndices;

    copy._currentTurnIndex = currentTurnIndex;
    copy.storySeed = storySeed;
    copy.previousRandom = previousRandom;

    copy.didSafeExit = didSafeExit;

    return copy;
  }

  void restoreAfterPatch() {
    // VariablesState was being borrowed by the patched
    // state, so restore it with our own callstack.
    // _patch will be null normally, but if you're in the
    // middle of a save, it may contain a _patch for save purpsoes.
    variablesState.callStack = callStack;
    variablesState.patch = _patch; // usually null
  }

  void applyAnyPatch() {
    _logger.finest('applyAnyPatch()');
    if (_patch == null) return;

    variablesState.ApplyPatch();
    _patch!.visitCounts.forEach((key, value) {
      applyCountChanges(key, value, isVisit: true);
    });
    _patch!.turnIndices.forEach((key, value) {
      applyCountChanges(key, value, isVisit: false);
    });
    _patch = null;
  }

  void applyCountChanges(Container container, int newCount,
      {required bool isVisit}) {
    var counts = isVisit ? _visitCounts : _turnIndices;
    counts[container.path.toString()] = newCount;
  }

  void resetErrors() {
    _currentErrors = null;
    _currentWarnings = null;
  }

  void resetOutput([List<RuntimeObject>? objs]) {
    _logger.finest('=== resetOutput === ( objs : $objs )');
    outputStream.clear();
    if (objs != null) {
      outputStream.addAll(objs);
    }
    _logger.finest('_outputStream : ${outputStream}');
    outputStreamDirty();
  }

  // Push to output stream, but split out newlines in text for consistency
  // in dealing with them later.
  void PushToOutputStream(RuntimeObject obj) {
    _logger.finest('=== PushToOutputStream ===  ${obj.toString()}');
    var text = asOrNull<StringValue>(obj);
    if (text != null) {
      _logger.finest('PushToOutputStream: ${text.value}');
      var listText = trySplittingHeadTailWhitespace(text);
      _logger.finest('listText : $listText');

      if (listText != null) {
        listText.forEach((textObj) {
          pushToOutputStreamIndividual(textObj);
        });
        outputStreamDirty();
        return;
      }
    }

    pushToOutputStreamIndividual(obj);

    outputStreamDirty();
  }

  void PopFromOutputStream(int count) {
    outputStream.removeRange(outputStream.length - count, outputStream.length);
    outputStreamDirty();
  }

  // At both the start and the end of the string, split out the new lines like so:
  //
  //  "   \n  \n     \n  the string \n is awesome \n     \n     "
  //      ^-----------^                           ^-------^
  //
  // Excess newlines are converted into single newlines, and spaces discarded.
  // Outside spaces are significant and retained. "Interior" newlines within
  // the main string are ignored, since this is for the purpose of gluing only.
  //
  //  - If no splitting is necessary, null is returned.
  //  - A newline on its own is returned in a list for consistency.
  List<StringValue>? trySplittingHeadTailWhitespace(StringValue single) {
    var str = single.value;

    var headFirstNewlineIdx = -1;
    var headLastNewlineIdx = -1;
    for (var i = 0; i < str.length; ++i) {
      var c = str[i];
      if (c == '\n') {
        if (headFirstNewlineIdx == -1) {
          headFirstNewlineIdx = i;
        }
        headLastNewlineIdx = i;
      } else if (c == ' ' || c == '\t') {
        continue;
      } else {
        break;
      }
    }

    var tailLastNewlineIdx = -1;
    var tailFirstNewlineIdx = -1;
    for (var i = 0; i < str.length; ++i) {
      var c = str[i];
      if (c == '\n') {
        if (tailLastNewlineIdx == -1) {
          tailLastNewlineIdx = i;
        }
        tailFirstNewlineIdx = i;
      } else if (c == ' ' || c == '\t') {
        continue;
      } else {
        break;
      }
    }

    // No splitting to be done?
    if (headFirstNewlineIdx == -1 && tailLastNewlineIdx == -1) {
      return null;
    }

    var listTexts = <StringValue>[];
    var innerStrStart = 0;
    var innerStrEnd = str.length;

    if (headFirstNewlineIdx != -1) {
      if (headFirstNewlineIdx > 0) {
        var leadingSpaces = StringValue(str.substring(0, headFirstNewlineIdx));
        listTexts.add(leadingSpaces);
      }
      listTexts.add(StringValue('\n'));
      innerStrStart = headLastNewlineIdx + 1;
    }

    if (tailLastNewlineIdx != -1) {
      innerStrEnd = tailFirstNewlineIdx;
    }

    if (innerStrEnd > innerStrStart) {
      var innerStrText =
          str.substring(innerStrStart, innerStrEnd - innerStrStart);
      listTexts.add(StringValue(innerStrText));
    }

    if (tailLastNewlineIdx != -1 && tailFirstNewlineIdx > headLastNewlineIdx) {
      listTexts.add(StringValue('\n'));
      if (tailLastNewlineIdx < str.length - 1) {
        var numSpaces = (str.length - tailLastNewlineIdx) - 1;
        var trailingSpaces =
            StringValue(str.substring(tailLastNewlineIdx + 1, numSpaces));
        listTexts.add(trailingSpaces);
      }
    }

    return listTexts;
  }

  void pushToOutputStreamIndividual(RuntimeObject obj) {
    var glue = asOrNull<Glue>(obj);
    var text = asOrNull<StringValue>(obj);

    var includeInOutput = true;

    // New glue, so chomp away any whitespace from the end of the stream
    if (glue != null) {
      trimNewlinesFromOutputStream();
      includeInOutput = true;
    }

    // New text: do we really want to append it, if it's whitespace?
    // Two different reasons for whitespace to be thrown away:
    //   - Function start/end trimming
    //   - User defined glue: <>
    // We also need to know when to stop trimming, when there's non-whitespace.
    else if (text != null) {
      // Where does the current function call begin?
      var functionTrimIndex = -1;
      var currEl = callStack.currentElement;
      if (currEl.type == PushPopType.Function) {
        functionTrimIndex = currEl.functionStartInOutputStream;
      }

      // Do 2 things:
      //  - Find latest glue
      //  - Check whether we're in the middle of string evaluation
      // If we're in string eval within the current function, we
      // don't want to trim back further than the length of the current string.
      var glueTrimIndex = -1;
      for (var i = outputStream.length - 1; i >= 0; i--) {
        var o = outputStream[i];
        var c = asOrNull<ControlCommand>(o);
        var g = asOrNull<Glue>(o);

        // Find latest glue
        if (g != null) {
          _logger.finest('Find latest glue');
          glueTrimIndex = i;
          _logger.finest('glue found : $glueTrimIndex');
          break;
        }

        // Don't function-trim past the start of a string evaluation section
        else if (c != null && c.commandType == CommandType.BeginString) {
          _logger.finest(
              "Don't function-trim past the start of a string evaluation section");
          if (i >= functionTrimIndex) {
            functionTrimIndex = -1;
          }
          break;
        }
      }

      // Where is the most agressive (earliest) trim point?
      var trimIndex = -1;
      if (glueTrimIndex != -1 && functionTrimIndex != -1) {
        trimIndex = min(functionTrimIndex, glueTrimIndex);
      } else if (glueTrimIndex != -1) {
        trimIndex = glueTrimIndex;
      } else {
        trimIndex = functionTrimIndex;
      }
      _logger.finest(
          'Where is the most agressive (earliest) trim point? $trimIndex');

      // So, are we trimming then?
      if (trimIndex != -1) {
        // While trimming, we want to throw all newlines away,
        // whether due to glue or the start of a function
        if (text.isNewLine) {
          _logger.finest('While trimming, we want to throw all newlines away,'
              'whether due to glue or the start of a function');
          includeInOutput = false;
        }

        // Able to completely reset when normal text is pushed
        else if (text.isNonWhitespace) {
          _logger.finest('Able to completely reset when normal text is pushed');
          if (glueTrimIndex > -1) {
            removeExistingGlue();
          }

          // Tell all functions in callstack that we have seen proper text,
          // so trimming whitespace at the start is done.
          if (functionTrimIndex > -1) {
            var callstackElements = callStack.elements;
            for (var i = callstackElements.length - 1; i >= 0; i--) {
              var el = callstackElements[i];
              if (el.type == PushPopType.Function) {
                el.functionStartInOutputStream = -1;
              } else {
                break;
              }
            }
          }
        }
      }

      // De-duplicate newlines, and don't ever lead with a newline
      else if (text.isNewLine) {
        _logger.finest(
            "De-duplicate newlines, and don't ever lead with a newline");
        if (outputStreamEndsInNewline || !outputStreamContainsContent) {
          includeInOutput = false;
        }
      }
    }

    if (includeInOutput) {
      _logger.finest('includeInOutput, add ${obj}');
      outputStream.add(obj);
      outputStreamDirty();
    }
  }

  void trimNewlinesFromOutputStream() {
    var removeWhitespaceFrom = -1;

    // Work back from the end, and try to find the point where
    // we need to start removing content.
    //  - Simply work backwards to find the first newline in a string of whitespace
    // e.g. This is the content   \n   \n\n
    //                            ^---------^ whitespace to remove
    //                        ^--- first while loop stops here
    var i = outputStream.length - 1;
    while (i >= 0) {
      var obj = outputStream[i];
      var cmd = asOrNull<ControlCommand>(obj);
      var txt = asOrNull<StringValue>(obj);
      if (cmd != null || (txt != null && txt.isNonWhitespace)) {
        break;
      } else if (txt != null && txt.isNewLine) {
        removeWhitespaceFrom = i;
      }
      i--;
    }

    // Remove the whitespace
    if (removeWhitespaceFrom >= 0) {
      _logger.finest('Remove the whitespace');
      i = removeWhitespaceFrom;
      while (i < outputStream.length) {
        var text = asOrNull<StringValue>(outputStream[i]);
        if (text != null) {
          outputStream.removeAt(i);
        } else {
          i++;
        }
      }
    }

    outputStreamDirty();
  }

  bool get outputStreamEndsInNewline {
    if (outputStream.isNotEmpty) {
      for (var i = outputStream.length - 1; i >= 0; i--) {
        var obj = outputStream[i];
        if (obj is ControlCommand) {
          break;
        }
        var text = asOrNull<StringValue>(outputStream[i]);
        if (text != null) {
          if (text.isNewLine) {
            return true;
          } else if (text.isNonWhitespace) {
            break;
          }
        }
      }
    }

    return false;
  }

  // Only called when non-whitespace is appended
  void removeExistingGlue() {
    for (var i = outputStream.length - 1; i >= 0; i--) {
      var c = outputStream[i];
      if (c is Glue) {
        outputStream.removeAt(i);
      } else if (c is ControlCommand) {
        // e.g. BeginString
        break;
      }
    }

    outputStreamDirty();
  }

  bool get outputStreamContainsContent =>
      outputStream.any((content) => content is StringValue);

  bool get inStringEvaluation {
    _logger.fine('=== inStringEvaluation ===');
    for (var i = outputStream.length - 1; i >= 0; i--) {
      var cmd = asOrNull<ControlCommand>(outputStream[i]);
      if (cmd != null && cmd.commandType == CommandType.BeginString) {
        _logger.fine('inStringEvaluation : true');
        return true;
      }
    }
    _logger.fine('inStringEvaluation : false');
    return false;
  }

  void pushEvaluationStack(RuntimeObject? obj) {
    _logger.fine('=== pushEvaluationStack ===  $obj');
    // Include metadata about the origin List for list values when
    // they're used, so that lower level functions can make use
    // of the origin list to get related items, or make comparisons
    // with the integer values etc.
    var listValue = asOrNull<ListValue>(obj);
    if (listValue != null) {
      _logger.fine('ListValue');
      // Update origin when list is has something to indicate the list origin
      var rawList = listValue.value;
      if (rawList.originNames != null) {
        rawList.origins ??= <ListDefinition>[];
        rawList.origins!.clear();

        for (var n in rawList.originNames!) {
          var def = story.listDefinitions!.getDefinition(n);
          if (def != null && !rawList.origins!.contains(def)) {
            rawList.origins!.add(def);
          }
        }
      }
    }
    ArgumentError.checkNotNull(obj);

    evaluationStack.add(obj!);
  }

  RuntimeObject popEvaluationStack() {
    //var obj = evaluationStack[evaluationStack.length - 1];
    //evaluationStack.removeAt(evaluationStack.length - 1);
    //return obj;
    return evaluationStack.removeLast();
  }

  RuntimeObject peekEvaluationStack() {
    //return evaluationStack[evaluationStack.length - 1];
    return evaluationStack.last;
  }

  // public List<Runtime.Object> PopEvaluationStack(int numberOfObjects)
  List<RuntimeObject> PopEvaluationStack(int numberOfObjects) {
    if (numberOfObjects > evaluationStack.length) {
      throw Exception('trying to pop too many objects');
    }

    var popped =
        evaluationStack.sublist(evaluationStack.length - numberOfObjects);
    evaluationStack.removeRange(
        evaluationStack.length - numberOfObjects, evaluationStack.length);
    return popped;
  }

  /// <summary>
  /// Ends the current ink flow, unwrapping the callstack but without
  /// affecting any variables. Useful if the ink is (say) in the middle
  /// a nested tunnel, and you want it to reset so that you can divert
  /// elsewhere using ChoosePathString(). Otherwise, after finishing
  /// the content you diverted to, it would continue where it left off.
  /// Calling this is equivalent to calling -> END in ink.
  /// </summary>
  void ForceEnd() {
    callStack.reset();

    _currentFlow.currentChoices.clear();

    currentPointer = Pointer.Null;
    previousPointer = Pointer.Null;

    didSafeExit = true;
  }

// Add the end of a function call, trim any whitespace from the end.
  // We always trim the start and end of the text that a function produces.
  // The start whitespace is discard as it is generated, and the end
  // whitespace is trimmed in one go here when we pop the function.
  void trimWhitespaceFromFunctionEnd() {
    assert(callStack.currentElement.type == PushPopType.Function);

    var functionStartPoint =
        callStack.currentElement.functionStartInOutputStream;

    // If the start point has become -1, it means that some non-whitespace
    // text has been pushed, so it's safe to go as far back as we're able.
    if (functionStartPoint == -1) {
      functionStartPoint = 0;
    }

    // Trim whitespace from END of function call
    for (var i = outputStream.length - 1; i >= functionStartPoint; i--) {
      var obj = outputStream[i];

      var txt = asOrNull<StringValue>(obj);
      var cmd = asOrNull<ControlCommand>(obj);

      if (txt == null) continue;
      if (cmd != null) break;

      if (txt.isNewLine || txt.isInlineWhitespace) {
        outputStream.removeAt(i);
        outputStreamDirty();
      } else {
        break;
      }
    }
  }

  void popCallstack([PushPopType? popType]) {
    // Add the end of a function call, trim any whitespace from the end.
    if (callStack.currentElement.type == PushPopType.Function) {
      trimWhitespaceFromFunctionEnd();
    }

    callStack.pop(popType);
  }

  // Don't make public since the method need to be wrapped in Story for visit counting
  void SetChosenPath(Path path, bool incrementingTurnIndex) {
    _logger.finest(
        'SetChosenPath ( incrementingTurnIndex: $incrementingTurnIndex )');

    // Changing direction, assume we need to clear current set of choices
    _currentFlow.currentChoices.clear();

    var newPointer = story.pointerAtPath(path).copyWith();
    if (!newPointer.isNull && newPointer.index == -1) {
      newPointer.index = 0;
    }

    currentPointer = newPointer;

    if (incrementingTurnIndex) {
      _currentTurnIndex++;
    }
  }

  void startFunctionEvaluationFromGame(
      Container funcContainer, List<dynamic>? arguments) {
    callStack.push(PushPopType.FunctionEvaluationFromGame,
        externalEvaluationStackHeight: evaluationStack.length);
    callStack.currentElement.currentPointer = Pointer.startOf(funcContainer);

    passArgumentsToEvaluationStack(arguments);
  }

  void PassArgumentsToEvaluationStack(List<dynamic>? arguments) =>
      passArgumentsToEvaluationStack(arguments);

  void passArgumentsToEvaluationStack(List<dynamic>? arguments) {
    // Pass arguments onto the evaluation stack
    if (arguments != null) {
      for (var i = 0; i < arguments.length; i++) {
        if (!(arguments[i] is int ||
            arguments[i] is double ||
            arguments[i] is String ||
            arguments[i] is bool ||
            arguments[i] is InkList)) {
          throw Exception(
              'ink arguments when calling EvaluateFunction / ChoosePathStringWithParameters must be int, float, string, bool or InkList. Argument was ' +
                  (arguments[i] == null
                      ? 'null'
                      : arguments[i].runtimeType.toString()));
        }

        pushEvaluationStack(Value.create(arguments[i])!);
      }
    }
  }

  bool tryExitFunctionEvaluationFromGame() {
    if (callStack.currentElement.type ==
        PushPopType.FunctionEvaluationFromGame) {
      currentPointer = Pointer.Null;
      didSafeExit = true;
      return true;
    }

    return false;
  }

  dynamic completeFunctionEvaluationFromGame() {
    if (callStack.currentElement.type !=
        PushPopType.FunctionEvaluationFromGame) {
      throw StoryException(
          'Expected external function evaluation to be complete. Stack trace: ' +
              callStack.callStackTrace);
    }

    var originalEvaluationStackHeight =
        callStack.currentElement.evaluationStackHeightWhenPushed;

    // Do we have a returned value?
    // Potentially pop multiple values off the stack, in case we need
    // to clean up after ourselves (e.g. caller of EvaluateFunction may
    // have passed too many arguments, and we currently have no way to check for that)
    RuntimeObject? returnedObj;
    while (evaluationStack.length > originalEvaluationStackHeight) {
      var poppedObj = popEvaluationStack();
      returnedObj ??= poppedObj;
    }

    // Finally, pop the external function evaluation
    popCallstack(PushPopType.FunctionEvaluationFromGame);

    // What did we get back?
    if (returnedObj != null) {
      if (returnedObj is RuntimeVoid) return null;

      // Some kind of value, if not void
      var returnVal = returnedObj as Value;

      // DivertTargets get returned as the string of components
      // (rather than a Path, which isn't public)
      if (returnVal.valueType == ValueType.DivertTarget) {
        return returnVal.valueObject.toString();
      }

      // Other types can just have their exact object type:
      // int, float, string. VariablePointers get returned as strings.
      return returnVal.valueObject;
    }

    return null;
  }

  void addError(String message, bool isWarning) {
    if (!isWarning) {
      if (currentErrors == null) _currentErrors = <String>[];
      currentErrors!.add(message);
    } else {
      if (currentWarnings == null) _currentWarnings = <String>[];
      currentWarnings!.add(message);
    }
  }

  late Map<String, int> _visitCounts;
  late Map<String, int> _turnIndices;

  var _outputStreamTextDirty = true;
  var _outputStreamTagsDirty = true;

  StatePatch? _patch;

  late Flow _currentFlow;
  Map<String, Flow>? _namedFlows;
  static const kDefaultFlowName = 'DEFAULT_FLOW';
  bool _aliveFlowNamesDirty = true;

  void outputStreamDirty() {
    _outputStreamTextDirty = true;
    _outputStreamTagsDirty = true;
  }

  /// Exports the current state to json format, in order to save the game.
  String ToJson() {
    var writer = SimpleJsonWriter();
    WriteJson(writer);
    return writer.toString();
  }

  /// Exports the current state to json format, in order to save the game.
  Map<String, dynamic> toJson() {
    var writer = SimpleJsonWriter();
    WriteJson(writer);
    return writer.toJson();
  }

  void WriteJson(SimpleJsonWriter writer) {
    writer.WriteObjectStart();

    // Flows
    writer.WritePropertyStart("flows");
    writer.WriteObjectStart();

    // Multi-flow
    if (_namedFlows != null) {
      for (var namedFlow in _namedFlows!.entries) {
        writer.WriteProperty(
            namedFlow.key, (w) => namedFlow.value.WriteJson(w));
      }
    }

    // Single flow
    else {
      writer.WriteProperty(_currentFlow.name, (w) => _currentFlow.WriteJson(w));
    }

    writer.WriteObjectEnd();
    writer.WritePropertyEnd(); // end of flows

    writer.WriteProperty("currentFlowName", _currentFlow.name);

    writer.WriteProperty("variablesState", (w) => variablesState.WriteJson(w));

    writer.WriteProperty("evalStack",
        (w) => JsonSerialization.WriteListRuntimeObjs(w, evaluationStack));

    if (!divertedPointer.isNull) {
      writer.WriteProperty(
          "currentDivertTarget", divertedPointer.path!.componentsString);
    }

    writer.WriteProperty("visitCounts",
        (w) => JsonSerialization.WriteIntDictionary(w, _visitCounts));
    writer.WriteProperty("turnIndices",
        (w) => JsonSerialization.WriteIntDictionary(w, _turnIndices));

    writer.WriteProperty("turnIdx", currentTurnIndex);
    writer.WriteProperty("storySeed", storySeed);
    writer.WriteProperty("previousRandom", previousRandom);

    writer.WriteProperty("inkSaveVersion", kInkSaveStateVersion);

    // Not using this right now, but could do in future.
    writer.WriteProperty("inkFormatVersion", Story.inkVersionCurrent);

    writer.WriteObjectEnd();
  }

  @override
  void LoadJsonObj(Map<String, dynamic> json) {
    int? jSaveVersion = json['inkSaveVersion'];
    ArgumentError.checkNotNull(
        jSaveVersion, "ink save format incorrect, can't load.");
    checkArgument(jSaveVersion! >= kMinCompatibleLoadVersion,
        message:
            "Ink save format isn't compatible with the current version (saw '" +
                jSaveVersion.toString() +
                "', but minimum is $kMinCompatibleLoadVersion), so can't load.");

    // Flows: Always exists in latest format (even if there's just one default)
    // but this dictionary doesn't exist in prev format
    var flowsObj = json['flows'];
    if (flowsObj != null) {
      var flowsObjDict = flowsObj as Map<String, dynamic>;
      if (flowsObjDict.length == 1) {
        // Single default flow
        _namedFlows = null;
      } else if (_namedFlows == null) {
        // Multi-flow, already have a flows dict
        _namedFlows = {};
      } else {
        // Multi-flow, already have a flows dict
        _namedFlows!.clear();
      }

      // Load up each flow (there may only be one)
      flowsObjDict.forEach((name, value) {
        var flowObj = value as Map<String, dynamic>;

        // Load up this flow using JSON data
        var flow = Flow(name, story, flowObj);

        if (flowsObjDict.length == 1) {
          _currentFlow = Flow(name, story, flowObj);
        } else {
          _namedFlows![name] = flow;
        }
      });

      if (_namedFlows != null && _namedFlows!.length > 1) {
        var currFlowName = json['currentFlowName'] as String;
        // Adding a bang at the end, because we're trusting the save, as
        // done in upstream.  If the save is corrupted, the execution
        // is undefined.
        _currentFlow = _namedFlows![currFlowName]!;
      }
    } else {
      // Old format: individually load up callstack, output stream, choices in current/default flow
      _namedFlows = null;
      _currentFlow.name = kDefaultFlowName;
      _currentFlow.callStack.setJsonToken(json['callstackThreads'], story);
      _currentFlow.outputStream =
          JsonSerialization.JArrayToRuntimeObjList(json['outputStream']);
      _currentFlow.currentChoices =
          JsonSerialization.JArrayToRuntimeObjList<Choice>(
              json['currentChoices']);

      Map<String, dynamic> jChoiceThreadsObj = json['choiceThreads'] ?? {};
      _currentFlow.LoadFlowChoiceThreads(jChoiceThreadsObj, story);
    }

    outputStreamDirty();
    _aliveFlowNamesDirty = true;

    variablesState.SetJsonToken(json['variablesState'] as Map<String, dynamic>);
    variablesState.callStack = _currentFlow.callStack;

    _evaluationStack = JsonSerialization.JArrayToRuntimeObjList(
        json['evalStack'] as List<dynamic>);
    var currentDivertTargetPath = json['currentDivertTarget'];
    if (currentDivertTargetPath != null) {
      var divertPath = Path(currentDivertTargetPath.toString());
      divertedPointer = story.pointerAtPath(divertPath);
    }

    _visitCounts =
        Map<String, int>.from(json['visitCounts'] as Map<String, dynamic>);

    _turnIndices =
        Map<String, int>.from(json['turnIndices'] as Map<String, dynamic>);

    _currentTurnIndex = json['turnIdx'] as int;
    storySeed = json['storySeed'] as int;

    // Not optional, but bug in inkjs means it's actually missing in inkjs saves
    var previousRandomObj = json['previousRandom'];
    if (previousRandomObj != null) {
      previousRandom = previousRandomObj as int;
    } else {
      previousRandom = 0;
    }
  }

  @override
  int VisitCountAtPathString(String pathString) =>
      visitCountAtPathString(pathString);

  @override
  void ResetOutput([List<RuntimeObject>? objs]) => resetOutput(objs);

  @override
  void StartFunctionEvaluationFromGame(
          Container funcContainer, List<Object> arguments) =>
      startFunctionEvaluationFromGame(funcContainer, arguments);

  CompleteFunctionEvaluationFromGame() => completeFunctionEvaluationFromGame();

  @override
  void LoadJson(String json) {
    var jObject = SimpleJson.TextToDictionary(json);
    LoadJsonObj(jObject);
    onDidLoadState?.call();
  }

  List<String> get aliveFlowNames {
    if (_aliveFlowNamesDirty) {
      _aliveFlowNames = [];
    }
    if (_namedFlows != null) {
      for (var flowName in _namedFlows!.keys) {
        if (flowName != kDefaultFlowName) {
          _aliveFlowNames!.add(flowName);
        }
      }
    }
    _aliveFlowNamesDirty = false;
    return _aliveFlowNames!;
  }

  List<String>? _aliveFlowNames;
}
