// Order is significant for type coersion.
// If types aren't directly compatible for an operation,
// they're coerced to the same type, downward.
// Higher value types "infect" an operation.
// (This may not be the most sensible thing to do, but it's worked so far!)

import 'package:quiver/strings.dart';

import 'ink_list.dart';
import 'path.dart';
import 'runtime_object.dart';
import 'story_exception.dart';

enum ValueType {
  // Bool is new addition, keep enum values the same, with Int==0, Float==1 etc,
  // but for coersion rules, we want to keep bool with a lower value than Int
  // so that it converts in the right direction
  // -1
  Bool,
  // Used in coersion

  Int, // 0
  Float, // 1
  List, // 2
  String, // 3

  // Not used for coersion described above
  DivertTarget,
  VariablePointer,
}

extension ValueTypeX on ValueType {
  int get effectiveIndex {
    switch (this) {
      case ValueType.Bool:
        return -1;
      case ValueType.Int:
        return 0;
      case ValueType.Float:
        return 1;
      case ValueType.List:
        return 2;
      case ValueType.String:
        return 3;
      case ValueType.DivertTarget:
        return 4;
      case ValueType.VariablePointer:
        return 5;
    }
  }
}

abstract class Value<T> extends _Value {
  T value;

  Value(this.value);

  @override
  String toString() {
    return value.toString();
  }

  @override
  Value? cast(ValueType newType);

  @override
  dynamic get valueObject => value;

  static Value? create(dynamic val) {
    // TODO
    // Implicitly lose precision from any doubles we get passed in
    if (val is double) {
      var doub = val;
      //val = (float)doub as float;
    }

    if (val is bool) {
      return BoolValue(val);
    } else if (val is int) {
      return IntValue(val);
      //} else if (val is long) {
      //  return new IntValue (()val);
      //} else if (val is float) {
      //  return new FloatValue ((float)val);
    } else if (val is double) {
      return FloatValue(val);
    } else if (val is String) {
      return StringValue(val);
    } else if (val is Path) {
      return DivertTargetValue(val);
    } else if (val is InkList) {
      return ListValue(val);
    }

    return null;
  }
}

abstract class _Value extends RuntimeObject {
  ValueType get valueType;

  bool get isTruthy;

  _Value? cast(ValueType newType);

  _Value();

  dynamic get valueObject;

  StoryException badCaseException(ValueType targetType) {
    return StoryException(
        "Can't cast $valueObject from $valueType to $targetType");
  }
}

class BoolValue extends Value<bool> {
  BoolValue(bool value) : super(value);

  @override
  Value cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    if (newType == ValueType.Int) {
      return IntValue(this.value ? 1 : 0);
    }
    if (newType == ValueType.Float) {
      return FloatValue(value ? 1.0 : 0.0);
    }
    if (newType == ValueType.String) {
      return StringValue(value ? 'true' : 'false');
    }
    throw badCaseException(newType);
  }

  @override
  bool get isTruthy => value;

  @override
  ValueType get valueType => ValueType.Bool;
}

class IntValue extends Value<int> {
  IntValue([int value = 0]) : super(value);

  @override
  Value cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    if (newType == ValueType.Bool) {
      return BoolValue(value == 0 ? false : true);
    }
    if (newType == ValueType.Float) {
      // TODO
      // return FloatValue(this.value);
      return FloatValue(value.toDouble());
    }
    if (newType == ValueType.String) {
      return StringValue(value.toString());
    }
    throw badCaseException(newType);
  }

  @override
  bool get isTruthy => value != 0;

  @override
  ValueType get valueType => ValueType.Int;
}

// TODO dart has not float
class FloatValue extends Value<double> {
  FloatValue([double value = 0.0]) : super(value);

  @override
  Value cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    if (newType == ValueType.Bool) {
      return BoolValue(value == 0.0 ? false : true);
    }
    if (newType == ValueType.Int) {
      return IntValue(value.toInt());
    }
    if (newType == ValueType.String) {
      return StringValue(value.toString());
    }
    throw badCaseException(newType);
  }

  @override
  bool get isTruthy => value != 0.0;

  @override
  ValueType get valueType => ValueType.Float;
}

class StringValue extends Value<String> {
  StringValue([String value = '']) : super(value) {
    // Classify whitespace status
    _isNewLine = value == '\n';
    _isInlineWhitespace = true;
    for (int i = 0; i < value.length; i++) {
      var c = value[i];
      if (c != ' ' && c != '\t') {
        _isInlineWhitespace = false;
        break;
      }
    }
  }

  @override
  Value? cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    if (newType == ValueType.Int) {
      var parsedInt = int.tryParse(value);
      if (parsedInt != null) {
        return IntValue(parsedInt);
      } else {
        return null;
      }
    }
    if (newType == ValueType.Float) {
      var parsedFloat = double.tryParse(value);
      if (parsedFloat != null) {
        return FloatValue(parsedFloat);
      } else {
        return null;
      }
    }
    throw badCaseException(newType);
  }

  @override
  bool get isTruthy => isNotEmpty(value);

  @override
  ValueType get valueType => ValueType.String;

  bool _isNewLine = false;
  bool _isInlineWhitespace = false;

  bool get isNewLine => _isNewLine;

  bool get isInlineWhitespace => _isInlineWhitespace;

  bool get isNonWhitespace => !isNewLine && !isInlineWhitespace;
}

class DivertTargetValue extends Value<Path> {
  DivertTargetValue([Path? value]) : super(value ?? Path.NULL);

  @override
  Value cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    throw badCaseException(newType);
  }

  @override
  String toString() {
    return 'DivertTargetValue(${targetPath.toString()})';
  }

  @override
  bool get isTruthy => throw Exception(
      "Shouldn't be checking the truthiness of a divert target");

  @override
  ValueType get valueType => ValueType.DivertTarget;

  Path get targetPath => value == Path.NULL ? throw Exception('null') : value;

  set targetPath(value) => this.value = value;
}

// TODO: Think: Erm, I get that this contains a string, but should
// we really derive from Value<string>? That seems a bit misleading to me.
class VariablePointerValue extends Value<String?> {
  VariablePointerValue([String? variableName, this.contextIndex = -1])
      : super(variableName);

  String? get variableName => value;

  set variableName(value) => this.value = value;

  // Where the variable is located
  // -1 = default, unknown, yet to be determined
  // 0  = in global scope
  // 1+ = callstack element index + 1 (so that the first doesn't conflict with special global scope)
  int contextIndex;

  @override
  Value cast(ValueType newType) {
    if (newType == valueType) {
      return this;
    }
    throw badCaseException(newType);
  }

  @override
  String toString() {
    return 'VariablePointerValue($variableName)';
  }

  VariablePointerValue copy() {
    return VariablePointerValue(variableName, contextIndex);
  }

  @override
  bool get isTruthy => throw Exception(
      "Shouldn't be checking the truthiness of a variable pointer");

  @override
  ValueType get valueType => ValueType.VariablePointer;
}

class ListValue extends Value<InkList> {
  ListValue([InkList? value]) : super(value ?? InkList());

  ListValue.single(InkListItem singleItem, int singleValue) : super(InkList()) {
    value = InkList.fromEntry(MapEntry(singleItem, singleValue));
  }

  static void retainListOriginsForAssignment(
      RuntimeObject? oldValue, RuntimeObject? newValue) {
    // When assigning the empty list, try to retain any initial origin names
    if (oldValue is ListValue &&
        newValue is ListValue &&
        newValue.value.isEmpty) {
      newValue.value.setInitialOriginNames(oldValue.value.originNames);
    }
  }

  @override
  Value cast(ValueType newType) {
    if (newType == ValueType.Int) {
      var max = value.maxItem;
      if (max.key.isNull) {
        return IntValue(0);
      } else {
        return IntValue(max.value);
      }
    } else if (newType == ValueType.Float) {
      var max = value.maxItem;
      if (max.key.isNull) {
        return FloatValue(0.0);
      } else {
        return FloatValue(max.value.toDouble());
      }
    } else if (newType == ValueType.String) {
      var max = value.maxItem;
      if (max.key.isNull) {
        return StringValue('');
      } else {
        return StringValue(max.key.toString());
      }
    }
    if (newType == valueType) {
      return this;
    }
    throw badCaseException(newType);
  }

  @override
  bool get isTruthy => value.isNotEmpty;

  @override
  ValueType get valueType => ValueType.List;
}
