import 'dart:collection';

import 'call_stack.dart';
import 'extension.dart';
import 'json.dart';
import 'list_definitions_origin.dart';
import 'logger.dart';
import 'runtime_object.dart';
import 'simple_json.dart';
import 'state_patch.dart';
import 'story_exception.dart';
import 'value.dart';
import 'variable_assignment.dart';

typedef VariableChanged = void Function(
    String variableName, RuntimeObject newValue);

/// Encompasses all the global variables in an ink Story, and
/// allows binding of a VariableChanged event so that that game
/// code can be notified whenever the global variables change.
class VariablesState extends Iterable<String> {
  final _logger = Logger('VariablesState');
  CallStack callStack;
  VariableChanged? variableChangedEvent;
  StatePatch? patch;
  bool _batchObservingVariableChanges = false;

  VariablesState(this.callStack, ListDefinitionsOrigin? listDefinitions)
      : _listDefsOrigin = listDefinitions,
        _globalVariables = <String, RuntimeObject>{};

  HashSet<String>? _changedVariablesForBatchObs;
  ListDefinitionsOrigin? _listDefsOrigin;
  Map<String, RuntimeObject> _globalVariables = {};
  Map<String, RuntimeObject> _defaultGlobalVariables = {};

  bool get batchObservingVariableChanges => _batchObservingVariableChanges;

  set batchObservingVariableChanges(bool value) {
    _batchObservingVariableChanges = value;
    if (value) {
      _changedVariablesForBatchObs = HashSet<String>();
    }

    // Finished observing variables in a batch - now send
    // notifications for changed variables all in one go.
    else {
      if (_changedVariablesForBatchObs != null) {
        for (var variableName in _changedVariablesForBatchObs!) {
          var currentValue = _globalVariables[variableName];
          variableChangedEvent?.call(variableName, currentValue!);
        }
      }

      _changedVariablesForBatchObs = null;
    }
  }

  /// Get or set the value of a named global ink variable.
  /// The types available are the standard ink types. Certain
  /// types will be implicitly casted when setting.
  /// For example, doubles to floats, longs to ints, and bools
  /// to ints.
  dynamic operator [](String variableName) {
    RuntimeObject? varContents;

    if (patch != null && patch!.getGlobal(variableName) != null) {
      return (patch!.getGlobal(variableName) as Value).valueObject;
    }

    // Search main dictionary first.
    // If it's not found, it might be because the story content has changed,
    // and the original default value hasn't be instantiated.
    // Should really warn somehow, but it's difficult to see how...!

    varContents = (_globalVariables[variableName] ??
        _defaultGlobalVariables[variableName]);

    if (varContents != null) {
      return (varContents as Value).valueObject;
    } else {
      return null;
    }
  }

  /// set the value of a named global ink variable.
  /// The types available are the standard ink types. Certain
  /// types will be implicitly casted when setting.
  /// For example, doubles to floats, longs to ints, and bools
  /// to ints.
  void operator []=(String variableName, dynamic value) {
    if (!_defaultGlobalVariables.containsKey(variableName)) {
      throw StoryException('Cannot assign to a variable (' +
          variableName +
          ") that hasn't been declared in the story");
    }

    var val = Value.create(value);
    if (val == null) {
      if (value == null) {
        throw Exception('Cannot pass null to VariableState');
      } else {
        throw Exception(
            'Invalid value passed to VariableState: ' + value.toString());
      }
    }

    SetGlobal(variableName, val);
  }

  void SetGlobal(String variableName, RuntimeObject value) {
    RuntimeObject? oldValue;
    if (patch == null || patch!.getGlobal(variableName) != null) {
      oldValue = _globalVariables[variableName];
    }

    ListValue.retainListOriginsForAssignment(oldValue, value);

    if (patch != null) {
      patch!.setGlobal(variableName, value);
    } else {
      _globalVariables[variableName] = value;
    }

    if (variableChangedEvent != null && value != (oldValue)) {
      if (batchObservingVariableChanges) {
        if (patch != null) {
          patch!.addChangedVariable(variableName);
        } else if (_changedVariablesForBatchObs != null) {
          _changedVariablesForBatchObs!.add(variableName);
        }
      } else {
        variableChangedEvent!(variableName, value);
      }
    }
  }

  @override
  Iterator<String> get iterator => _globalVariables.keys.iterator;

  void variablesState(
      CallStack callStack, ListDefinitionsOrigin listDefsOrigin) {
    _globalVariables = <String, RuntimeObject>{};
    callStack = callStack;
    _listDefsOrigin = listDefsOrigin;
  }

  void ApplyPatch() {
    for (var namedVar in patch!.globals.entries) {
      _globalVariables[namedVar.key] = namedVar.value;
    }

    if (_changedVariablesForBatchObs != null) {
      for (var name in patch!.changedVariables) {
        _changedVariablesForBatchObs!.add(name);
      }
    }

    patch = null;
  }

  void SetJsonToken(Map<String, dynamic> jToken) {
    _globalVariables.clear();

    for (var varVal in _defaultGlobalVariables.entries) {
      var loadedToken = jToken[varVal.key];
      if (loadedToken != null) {
        _globalVariables[varVal.key] =
            JsonSerialization.JTokenToRuntimeObject(loadedToken)!;
      } else {
        _globalVariables[varVal.key] = varVal.value;
      }
    }
  }

  /// When saving out JSON state, we can skip saving global values that
  /// remain equal to the initial values that were declared in ink.
  /// This makes the save file (potentially) much smaller assuming that
  /// at least a portion of the globals haven't changed. However, it
  /// can also take marginally longer to save in the case that the
  /// majority HAVE changed, since it has to compare all globals.
  /// It may also be useful to turn this off for testing worst case
  /// save timing.
  static bool dontSaveDefaultValues = true;

  void WriteJson(SimpleJsonWriter writer) {
    writer.WriteObjectStart();
    for (var keyVal in _globalVariables.entries) {
      var name = keyVal.key;
      var val = keyVal.value;

      if (dontSaveDefaultValues) {
        // Don't write out values that are the same as the default global values
        RuntimeObject? defaultVal = _defaultGlobalVariables[name];
        if (defaultVal != null) {
          if ($runtimeObjectsEqual(val, defaultVal)) continue;
        }
      }

      writer.WritePropertyStart(name);
      JsonSerialization.WriteRuntimeObject(writer, val);
      writer.WritePropertyEnd();
    }
    writer.WriteObjectEnd();
  }

  bool $runtimeObjectsEqual(RuntimeObject? obj1, RuntimeObject obj2) {
    if (obj1.runtimeType != obj2.runtimeType) return false;

    // Perform equality on int/float/bool manually to avoid boxing
    var boolVal = asOrNull<BoolValue>(obj1);
    if (boolVal != null) {
      return boolVal.value == (obj2 as BoolValue?)?.value;
    }

    var intVal = asOrNull<IntValue>(obj1);
    if (intVal != null) {
      return intVal.value == (obj2 as IntValue).value;
    }

    var floatVal = asOrNull<FloatValue>(obj1);
    if (floatVal != null) {
      return floatVal.value == (obj2 as FloatValue).value;
    }

    // Other Value type (using proper Equals: list, string, divert path)
    var val1 = asOrNull<Value>(obj1);
    var val2 = obj2 as Value;
    if (val1 != null) {
      return val1.valueObject == val2.valueObject;
    }

    throw Exception(
        'FastRoughDefinitelyEquals: Unsupported runtime object type: ' +
            obj1.runtimeType.toString());
  }

  RuntimeObject? getVariableWithName(String? name, [int contextIndex = -1]) {
    var varValue = getRawVariableWithName(name, contextIndex);
    _logger.finest('varValue : $varValue');
    // Get value from pointer?
    var varPointer = asOrNull<VariablePointerValue>(varValue);
    if (varPointer != null) {
      varValue = valueAtVariablePointer(varPointer);
    }

    return varValue;
  }

  RuntimeObject? tryGetDefaultVariableValue(String name) {
    return _defaultGlobalVariables[name];
  }

  bool globalVariableExistsWithName(String name) {
    return _globalVariables.containsKey(name) ||
        _defaultGlobalVariables.containsKey(name);
  }

  RuntimeObject? getRawVariableWithName(String? name, int contextIndex) {
    RuntimeObject? varValue;
    _logger.finest(
        'getRawVariableWithName ( name : $name , contentIndex : $contextIndex )');

    _logger.finest(() => '_globalVariables : $_globalVariables');
    _logger.finest(() => '_defaultGlobalVariables : $_defaultGlobalVariables');

    // 0 context = global
    if (contextIndex == 0 || contextIndex == -1) {
      if (patch != null && patch!.getGlobal(name) != null) {
        _logger.finest('patch.get global');
        return patch!.getGlobal(name)!;
      }

      if (_globalVariables.containsKey(name)) {
        _logger.finest('_globalVariables');
        return _globalVariables[name]!;
      }

      // Getting variables can actually happen during globals set up since you can do
      //  VAR x = A_LIST_ITEM
      // So _defaultGlobalVariables may be null.
      // We need to do this check though in case a new global is added, so we need to
      // revert to the default globals dictionary since an initial value hasn't yet been set.
      if (_defaultGlobalVariables != null &&
          _defaultGlobalVariables.containsKey(name)) {
        return _defaultGlobalVariables[name]!;
      }
      if (_listDefsOrigin == null) {
        throw ArgumentError.notNull('VariablesState._listDefsOrigin');
      }

      var listItemValue = _listDefsOrigin!.findSingleItemListWithName(name);
      if (listItemValue != null) {
        return listItemValue;
      }
    }

    // Temporary
    varValue = callStack.getTemporaryVariableWithName(name,
        contextIndex: contextIndex);

    return varValue;
  }

  //public Object ValueAtVariablePointer(VariablePointerValue pointer)
  RuntimeObject? valueAtVariablePointer(VariablePointerValue pointer) {
    return getVariableWithName(pointer.variableName!, pointer.contextIndex);
  }

  void assign(VariableAssignment varAss, RuntimeObject value) {
    var name = varAss.variableName;
    var contextIndex = -1;

    // Are we assigning to a global variable?
    var _setGlobal = false;
    if (varAss.isNewDeclaration) {
      _setGlobal = varAss.isGlobal;
    } else {
      _setGlobal = globalVariableExistsWithName(name!);
    }

    // Constructing new variable pointer reference
    if (varAss.isNewDeclaration) {
      var varPointer = asOrNull<VariablePointerValue>(value);
      if (varPointer != null) {
        var fullyResolvedVariablePointer = _resolveVariablePointer(varPointer);
        value = fullyResolvedVariablePointer;
      }
    }

    // Assign to existing variable pointer?
    // Then assign to the variable that the pointer is pointing to by name.
    else {
      // De-reference variable reference to point to
      VariablePointerValue? existingPointer;
      do {
        existingPointer = asOrNull<VariablePointerValue>(
            getRawVariableWithName(name!, contextIndex));
        if (existingPointer != null) {
          name = existingPointer.variableName;
          contextIndex = existingPointer.contextIndex;
          _setGlobal = (contextIndex == 0);
        }
      } while (existingPointer != null);
    }

    if (_setGlobal) {
      SetGlobal(name!, value);
    } else {
      callStack.setTemporaryVariable(name!, value, varAss.isNewDeclaration,
          contextIndex: contextIndex);
    }
  }

  void snapshotDefaultGlobals() {
    _defaultGlobalVariables = Map<String, RuntimeObject>.from(_globalVariables);
  }

  //void RetainListOriginsForAssignment (Object oldValue, Object newValue)
  void _retainListOriginsForAssignment(
      RuntimeObject oldValue, RuntimeObject newValue) {
    if (oldValue is ListValue &&
        newValue is ListValue &&
        newValue.value.isEmpty) {
      newValue.value.setInitialOriginNames(oldValue.value.originNames);
    }
  }

  /// Given a variable pointer with just the name of the target known, resolve to a variable
  /// pointer that more specifically points to the exact instance: whether it's global,
  /// or the exact position of a temporary on the callstack.
  //VariablePointerValue ResolveVariablePointer(VariablePointerValue varPointer)
  VariablePointerValue _resolveVariablePointer(
      VariablePointerValue varPointer) {
    var contextIndex = varPointer.contextIndex;

    if (contextIndex == -1) {
      contextIndex = _getContextIndexOfVariableNamed(varPointer.variableName!);
    }

    var valueOfVariablePointedTo =
        getRawVariableWithName(varPointer.variableName, contextIndex);

    // Extra layer of indirection:
    // When accessing a pointer to a pointer (e.g. when calling nested or
    // recursive functions that take a variable references, ensure we don't create
    // a chain of indirection by just returning the final target.
    var doubleRedirectionPointer =
        asOrNull<VariablePointerValue>(valueOfVariablePointedTo);
    if (doubleRedirectionPointer != null) {
      return doubleRedirectionPointer;
    }

    // Make copy of the variable pointer so we're not using the value direct from
    // the  Temporary must be local to the current scope.
    else {
      return VariablePointerValue(varPointer.variableName, contextIndex);
    }
  }

  // 0  if named variable is global
  // 1+ if named variable is a temporary in a particular call stack element
  int _getContextIndexOfVariableNamed(String varName) {
    if (globalVariableExistsWithName(varName)) {
      return 0;
    }

    return callStack.currentElementIndex;
  }
}
