import 'package:ink_runtime/ink.dart';
import 'package:test/test.dart';

import 'test_utils.dart';

void main() {
  group('ConditionalSpec', () {
    setUp(() {
      logSetUp();
    });
    test('if true - evaluate the statements if the condition evaluates to true',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/iftrue.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'The value is 1.');
    });

    test(
        'if false -not evaluate the statement if the condition evaluates to false',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/iffalse.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'The value is 3.');
    });

    test(
        'if else - evaluate an else statement if it exists and no other condition evaluates'
        'to true', () {
      final text = <String>[];
      var json = getJsonString('conditional/ifelse.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'The value is 1.');
    });

    test(
        'if else ext - evaluate an extended else statement if it exists and no other condition'
        'evaluates to true', () {
      final text = <String>[];
      var json = getJsonString('conditional/ifelse-ext.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'The value is -1.');
    });

    test(
        'if else ext text1 - evaluate an extended else statement with text and divert at the end',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/ifelse-ext-text1.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'This is text 1.');

      expect(story.currentChoices.length, 1);
      story.chooseChoiceIndex(0);
      nextAll(story, text);
      expect(text.length, 2);

      expect(text[1], 'This is the end.');
    });

    test(
        'if else ext text2 - evaluate an extended else statement with text and divert at the end',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/ifelse-ext-text2.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'This is text 2.');

      expect(story.currentChoices.length, 1);
      story.chooseChoiceIndex(0);
      nextAll(story, text);
      expect(text.length, 2);

      expect(text[1], 'This is the end.');
    });

    test(
        'if else ext text3 - evaluate an extended else statement with text and divert at the end',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/ifelse-ext-text3.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);

      expect(text.length, 1);
      expect(text.first, 'This is text 3.');

      expect(story.currentChoices.length, 1);
      story.chooseChoiceIndex(0);
      nextAll(story, text);
      expect(text.length, 2);

      expect(text[1], 'This is the end.');
    });

    test(
        'cond text1 - work with conditional content which is not only logic (example 1)',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/condtext.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 3);
      expect(text[1],
          "I stared at Monsieur Fogg. \"But surely you are not serious?\" I demanded.");
    });

    test(
        'cond text2 - work with conditional content which is not only logic (example 2)',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/condtext.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      story.chooseChoiceIndex(1);

      text.clear();
      nextAll(story, text);
      expect(text.length, 2);
      expect(text[0],
          "I stared at Monsieur Fogg. \"But there must be a reason for this trip,\" I observed.");
    });

    test('cond opt1 - work with options as conditional content (example 1)',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/condopt.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(story.currentChoices.length, 1);
    });

    test('cond opt2 - work with options as conditional content (example 2)',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/condopt.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      story.chooseChoiceIndex(1);

      text.clear();
      nextAll(story, text);
      expect(story.currentChoices.length, 2);
    });

    test(
        'stopping - go through the alternatives and stick on last when the keyword is stopping',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/stopping.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], 'I entered the casino.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);

      expect(text.length, 1);
      expect(text[0], 'I entered the casino again.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], "Once more, I went inside.");
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], "Once more, I went inside.");
      story.chooseChoiceIndex(0);
    });

    test('cycle - show each in turn and then cycle when the keyword is cycle',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/cycle.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], 'I held my breath.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);

      expect(text.length, 1);
      expect(text[0], 'I waited impatiently.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], "I paused.");
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], "I held my breath.");
      story.chooseChoiceIndex(0);
    });

    test(
        'once - show each, once, in turn, until all have been shown when the keyword is once',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/once.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text[0], 'Would my luck hold?');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);

      print(text);

      expect(text.length, 1);
      expect(text[0], 'Could I win the hand?');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 0);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 0);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 0);
    });

    test('shuffle - show one at random when the keyword is shuffle', () {
      final text = <String>[];
      var json = getJsonString('conditional/shuffle.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
    });

    test('shuffle stopping', () {
      final text = <String>[];
      var json = getJsonString('conditional/shuffle_stopping.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'final');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'final');
    });

    test('shuffle once', () {
      final text = <String>[];
      var json = getJsonString('conditional/shuffle_once.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 0);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 0);
    });

    test('multiline - show multiple lines of texts from multiline list blocks',
        () {
      final text = <String>[];
      var json = getJsonString('conditional/multiline.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'At the table, I drew a card. Ace of Hearts.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 2);
      expect(text.first, 'I drew a card. 2 of Diamonds.');
      expect(text[1], "\"Should I hit you again,\" the croupier asks.");
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 2);
      expect(text.first, 'I drew a card. King of Spades.');
      expect(text[1], "\"You lose,\" he crowed.");
    });

    test('multiline divert - allow for embedded diverts', () {
      final text = <String>[];
      var json = getJsonString('conditional/multiline-divert.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'At the table, I drew a card. Ace of Hearts.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 2);
      expect(text.first, 'I drew a card. 2 of Diamonds.');
      expect(text[1], "\"Should I hit you again,\" the croupier asks.");
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 2);
      expect(text.first, 'I drew a card. King of Spades.');
      expect(text[1], "\"You lose,\" he crowed.");
    });

    test('multiline choice - allow for embedded choices', () {
      final text = <String>[];
      var json = getJsonString('conditional/multiline-choice.ink.json');
      final story = Story.fromJson(json);

      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'At the table, I drew a card. Ace of Hearts.');
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(story.currentChoices.length, 2);
      story.chooseChoiceIndex(0);

      text.clear();
      nextAll(story, text);
      expect(text.length, 1);
      expect(text.first, 'I left the table.');
    });
  });
}
