import 'dart:convert';

import 'package:collection/collection.dart';
import 'package:ink_runtime/ink.dart';
import 'package:test/test.dart';

import '../../../runtime_tests.dart';
import '../inkfiles/compiled/inkjs/tests.ink.json.dart' as $tests;

Function eq = const ListEquality().equals;

void main() {
  group('Integration', () {
    late Story story;

    final List<List<dynamic>> called = [];

    dynamic externalSpy(dynamic a, dynamic b, dynamic c) {
      called.add([a, b, c]);
      return a;
    }

    setUp(() async {
      story = await createStory(jsonEncode($tests.json));
      story.allowExternalFunctionFallbacks = true;
    });
    test('should load a file', () {
      expect(story.canContinue, isTrue);
    });

    test('should jump to a knot', () {
      story.ChoosePathString('knot');
      expect(story.canContinue, true);

      expect(story.Continue(), 'Knot content\n');
    });

    test('should get where the story currently is', () {
      story.ChoosePathString('knot');
      expect(story.state.currentPathString, 'knot.0');
      expect(story.canContinue, true);
      story.Continue();
      expect(story.state.currentPathString, null);
      expect(story.canContinue, false);
    });

    test('should jump to a stitch', () {
      story.ChoosePathString('knot.stitch');
      expect(story.canContinue, true);

      expect(story.Continue(), 'Stitch content\n');
    });

    test('should read variables from ink', () {
      expect(story.variablesState['stringvar'], 'Emilia');
      expect(story.variablesState['intvar'], 521);
      expect(story.variablesState['floatvar'], 52.1);
      expect(story.variablesState['divertvar'].toString(),
          'logic.logic_divert_dest');
    });
    test('should write variables to ink', () {
      expect(story.variablesState['stringvar'], 'Emilia');
      story.variablesState['stringvar'] = 'Jonas';
      expect(story.variablesState['stringvar'], 'Jonas');
    });

    /*
    test("should observe variables", ()  {
    story.ChoosePathString("integration.variable_observer");
        expect(story.variablesState["observedVar1"],1);
    expect(story.variablesState["observedVar2"],2);

    const spy1 = jasmine.createSpy("variable observer spy 1");
    const spy2 = jasmine.createSpy("variable observer spy 2");
    const commonSpy = jasmine.createSpy("variable observer spy common");
    story.ObserveVariable("observedVar1", spy1);
    story.ObserveVariable("observedVar2", spy2);
    story.ObserveVariable("observedVar1", commonSpy);
    story.ObserveVariable("observedVar2", commonSpy);

    expect(story.Continue(),"declared\n");

    expect(story.variablesState["observedVar1"],1);
    expect(story.variablesState["observedVar2"],2);
    expect(spy1).toHaveBeenCalledTimes(0);
    expect(spy2).toHaveBeenCalledTimes(0);
    expect(commonSpy).toHaveBeenCalledTimes(0);

    expect(story.Continue(),"mutated 1\n");

    expect(story.variablesState["observedVar1"],3);
    expect(story.variablesState["observedVar2"],2);
    expect(spy1).toHaveBeenCalledTimes(1);
    expect(spy1).toHaveBeenCalledWith("observedVar1", 3);
    expect(spy2).toHaveBeenCalledTimes(0);
    expect(commonSpy).toHaveBeenCalledTimes(1);
    expect(commonSpy).toHaveBeenCalledWith("observedVar1", 3);

    expect(story.Continue(),"mutated 2\n");

    expect(story.variablesState["observedVar1"],4);
    expect(story.variablesState["observedVar2"],5);

    expect(spy1).toHaveBeenCalledTimes(2);
    expect(spy1).toHaveBeenCalledWith("observedVar1", 4);
    expect(spy2).toHaveBeenCalledTimes(1);
    expect(spy2).toHaveBeenCalledWith("observedVar2", 5);
  });*/

    test('should increment the read count on each visit', () {
      expect(story.state.VisitCountAtPathString('integration.visit_count'), 0);

      for (var i = 0; i < 10; ++i) {
        story.ChoosePathString('integration.visit_count');
        expect(story.Continue(), 'visited\n');
        expect(story.canContinue, false);
        expect(story.state.VisitCountAtPathString('integration.visit_count'),
            i + 1);
        story.ChoosePathString('integration.variable_observer');
        story.Continue();
      }
    });

    test('should increment the read count when the callstack is reset', () {
      expect(story.state.VisitCountAtPathString('integration.visit_count'), 0);

      for (var i = 0; i < 10; ++i) {
        story.ChoosePathString('integration.visit_count');
        expect(story.Continue(), 'visited\n');
        expect(story.state.VisitCountAtPathString('integration.visit_count'),
            i + 1);
      }
    });

    test('should not increment the read count when the callstack is not reset',
        () {
      expect(story.state.VisitCountAtPathString('integration.visit_count'), 0);

      for (var i = 0; i < 10; ++i) {
        story.ChoosePathString('integration.visit_count',
            resetCallstack: false);
        expect(story.Continue(), 'visited\n');
        expect(
            story.state.VisitCountAtPathString('integration.visit_count'), 1);
      }
    });

    test('should call ink functions', () {
      expect(story.EvaluateFunction('fn_with_return').returned, 'returned');
      expect(story.EvaluateFunction('fn_without_return').returned, isNull);
      expect(story.EvaluateFunction('fn_print').returned, isNull);
      expect(story.EvaluateFunction('fn_calls_other').returned,
          'nested function called');
    });
    test('should call ink functions with params', () {
      expect(story.EvaluateFunction('fn_params', ['a', 'b']).returned, 'was a');
      expect(story.EvaluateFunction('fn_echo', ['string']).returned, 'string');
      expect(story.EvaluateFunction('fn_echo', [5]).returned, 5);
      expect(story.EvaluateFunction('fn_echo', [5.3]).returned, 5.3);
    });

    test('should return output and return value from ink function calls', () {
      expect(story.EvaluateFunction('fn_print', []),
          EvaluateFunctionTextOutput(null, 'function called\n'));

      expect(story.EvaluateFunction('fn_echo', ['string']),
          EvaluateFunctionTextOutput('string', 'string\n'));

      expect(story.EvaluateFunction('fn_echo', [5]),
          EvaluateFunctionTextOutput(5, '5\n'));

      expect(story.EvaluateFunction('fn_echo', [5.3]),
          EvaluateFunctionTextOutput(5.3, '5.3\n'));
    });

    test('should call external functions', () {
      story.allowExternalFunctionFallbacks = false;
      story.ChoosePathString('integration.external');

      story.BindExternalFunction3('fn_ext', externalSpy);
      story.BindExternalFunction1('gameInc', (a) => null);

      expect(story.continueMaximally(), '1\n1.1\na\na\n');

      expect(called.any((element) => eq(element, [1, 2, 3])), true);
      expect(called.any((element) => eq(element, [1.1, 2.2, 3.3])), true);
      expect(called.any((element) => eq(element, ['a', 'b', 'c'])), true);
      expect(called.any((element) => eq(element, ['a', 1, 2.2])), true);
    });

    test('should handle callstack changes', () {
      story.allowExternalFunctionFallbacks = false;

      /*
        const externalSpy = jasmine
            .createSpy("external function spy", (x)  {
        x++;
        x = parseInt(story.EvaluateFunction("inkInc", [x]));
    return x;
    })
        .and.callThrough();*/
      story.BindExternalFunction('fn_ext', () => null);
      story.BindExternalFunction1('gameInc', (Object x) {
        print('game inc : $x');
        var y = x as int;
        y = x + 1;
        y = story.EvaluateFunction('inkInc', [y]).returned;
        return y;
      });

      final result2 = story.EvaluateFunction('gameInc', [5]);
      expect(result2.returned, 6);

      final result = story.EvaluateFunction('topExternal', [5]);
      expect(result.output, 'In top external\n');
      expect(result.returned, 7);

      //expect(result.output, 'In top external\n');
    }, skip: true);

    test('should return a visit count', () {
      expect(story.state.VisitCountAtPathString('game_queries.turnssince'), 0);

      story.ChoosePathString('game_queries.turnssince');
      story.Continue();
      expect(story.state.VisitCountAtPathString('game_queries.turnssince'), 1);

      story.ChoosePathString('game_queries.turnssince_1');
      story.Continue();
      story.ChoosePathString('game_queries.turnssince');
      story.Continue();
      expect(story.state.VisitCountAtPathString('game_queries.turnssince'), 2);
    });

    group('Saving and Loading', () {
      test('should continue the story', () {
        story.ChoosePathString('saveload');
        expect(story.Continue(), 'a bit of content\n');
        final save = story.state.toJson();
        story.state.LoadJsonObj(save);
        expect(story.Continue(), 'the next bit\n');
      });

      test('should restore a choice point', () {
        story.ChoosePathString('saveload.choicepoint');
        story.Continue();
        expect(story.currentChoices.length, 2);
        expect(story.currentChoices[0].text, 'choice 1');
        expect(story.currentChoices[1].text, 'choice 2');

        final save = story.state.toJson();
        print(save);
        story.state.LoadJsonObj(save);
        final save2 = story.state.toJson();

        expect(save, save2);

        expect(story.currentChoices.length, 2);
        expect(story.currentChoices[0].text, 'choice 1');
        expect(story.currentChoices[1].text, 'choice 2');
      });
    });
  });
}
